/************************************************************************/
/* confxml.c - Configuration library Source				*/
/************************************************************************/
/*
 *  (c)2009,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <errno.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>

#include <netinfo/confxml.h>


struct	_Configuration {
	xmlDocPtr	xml_doc;
};


struct	_ConfigSearch {
	Configuration		*configuration;
	xmlXPathContextPtr	xpath_ctx;
	xmlXPathObjectPtr	xpath_obj;
	xmlNodeSetPtr		xml_nodeset;
	int			result_count;
	int			result_index;
};

/*
struct	_ConfigItem {
	xmlNodePtr	xml_node;
};
*/

/* app-wide structure */
static struct {
	ConfigChangeCallback	callback;
} configuration_block;



NETINFO_API int configuration_init (ConfigChangeCallback callback)
{
	LIBXML_TEST_VERSION;
	configuration_block.callback = callback;
	return ESUCCESS;
}


NETINFO_API int configuration_uninit (void)
{
	configuration_block.callback = NULL;
	xmlCleanupParser ();
	return ESUCCESS;
}


NETINFO_API int configuration_new (Configuration **cp)
{
	Configuration	*c = calloc (1, sizeof (Configuration));
	if (c == NULL)
		return ENOMEM;
	c->xml_doc = xmlNewDoc (BAD_CAST "1.0");
	if (c->xml_doc == NULL) {
		free (c);
		return ENOMEM;
	}
	*cp = c;
	return ESUCCESS;
}


NETINFO_API int configuration_free (Configuration **cp)
{
	Configuration	*c = *cp;
	if (c == NULL)
		return EINVAL;
	if (c->xml_doc != NULL)
		xmlFreeDoc (c->xml_doc);
	free (c);
	*cp = NULL;
	return ESUCCESS;
}



NETINFO_API int configuration_load_file (Configuration *c,
				char *filepath)
{
	int		rc;
	xmlDocPtr	doc = NULL;
	if (c == NULL || filepath == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (c,
						filepath,
						CONFIG_OP_LOAD_FILE);
	}
	doc = xmlReadFile (filepath, NULL, 0);
	if (doc == NULL)
		return EPARSE;
	if (c->xml_doc != NULL)
		xmlFreeDoc (c->xml_doc);
	c->xml_doc = doc;
	return ESUCCESS;
}


NETINFO_API int configuration_load_memory (Configuration *c,
				char *buffer,
				int bufferlen)
{
	int		rc;
	xmlDocPtr	doc = NULL;
	if (c == NULL || buffer == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (c,
						buffer,
						CONFIG_OP_LOAD_MEMORY);
	}
	doc = xmlReadMemory (buffer, bufferlen, "memory.xml", NULL, 0);
	if (doc == NULL)
		return EPARSE;
	if (c->xml_doc != NULL)
		xmlFreeDoc (c->xml_doc);
	c->xml_doc = doc;
	return ESUCCESS;
}


NETINFO_API int configuration_save_file (Configuration *c,
				char *filepath)
{
	int		rc;
	if (c == NULL || filepath == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (c,
						filepath,
						CONFIG_OP_SAVE_FILE);
	}
	rc = xmlSaveFile (filepath, c->xml_doc);
	if (rc == -1)
		return errno;
	return ESUCCESS;
}


NETINFO_API int configuration_save_memory (Configuration *c,
				char **bufferp,
				int *bufferlenp)
{
	int		rc;
	if (c == NULL || bufferp == NULL || bufferlenp == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (c,
						bufferp,
						CONFIG_OP_SAVE_MEMORY);
	}
	xmlDocDumpFormatMemory (c->xml_doc, (xmlChar **)bufferp, bufferlenp, 1);
	return ESUCCESS;
}



NETINFO_API int configuration_search_new (Configuration *c,
				ConfigSearch **csp,
				char *namespacelist)
{
	ConfigSearch	*cs = NULL;
	if (c == NULL || csp == NULL)
		return EINVAL;
	cs = calloc (1, sizeof (ConfigSearch));
	if (cs == NULL)
		return ENOMEM;
	cs->configuration = c;
	cs->xpath_ctx = xmlXPathNewContext (cs->configuration->xml_doc);
	if (cs->xpath_ctx == NULL)
		return ENOMEM;
	if (namespacelist != NULL) {
		/* TODO: add namespaces */
	}
	*csp = cs;
	return ESUCCESS;
}


NETINFO_API int configuration_search_begin (ConfigSearch *cs,
				char *nodepath)

{
	if (cs == NULL || nodepath == NULL)
		return EINVAL;

	/* Free preious object, if any */
	if (cs->xpath_obj != NULL)
		xmlXPathFreeObject (cs->xpath_obj);

	cs->xpath_obj = xmlXPathEvalExpression ((unsigned char *)nodepath,
						cs->xpath_ctx);
	if (cs->xpath_obj == NULL)
		return ENOMEM;
	cs->xml_nodeset = cs->xpath_obj->nodesetval;
	cs->result_count = cs->xml_nodeset->nodeNr;
	cs->result_index = 0;
	return ESUCCESS;
}


NETINFO_API int configuration_search_item_count (ConfigSearch *cs)
{
	if (cs == NULL)
		return 0;
	if (cs->xml_nodeset == NULL)
		return 0;
	return cs->result_count;
}


NETINFO_API int configuration_search_next_item (ConfigSearch *cs,
				ConfigItem **cip)
{
	if (cs == NULL || cip == NULL)
		return EINVAL;
	if (cs->xml_nodeset == NULL)
		return ENOENT;
	if (cs->result_index >= cs->result_count)
		return ENOENT;
	*cip = (void *)cs->xml_nodeset->nodeTab[cs->result_index];
	cs->result_index++;
	return ESUCCESS;
}


NETINFO_API int configuration_search_add_item (ConfigSearch *cs,
				ConfigItem *ci)
{
	int		rc;
	xmlNodePtr	node = NULL;
	if (cs->xml_nodeset == NULL)
		return EINVAL;
	if (cs->result_index >= cs->result_count)
		return ENOENT;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (cs->configuration,
						ci,
						CONFIG_OP_ITEM_ADD);
	}
	node = cs->xml_nodeset->nodeTab[cs->result_index];
	if (xmlAddSibling (node, (xmlNodePtr)ci) == NULL)
		return EINVAL;
	return ESUCCESS;
}


NETINFO_API int configuration_search_remove_item (ConfigSearch *cs,
				ConfigItem *ci)
{
	int		rc;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (cs->configuration,
						ci,
						CONFIG_OP_ITEM_REMOVE);
	}
	return ESUCCESS;
}


NETINFO_API int configuration_search_end (ConfigSearch *cs)
{
	if (cs == NULL)
		return EINVAL;
	if (cs->xpath_obj != NULL)
		xmlXPathFreeObject (cs->xpath_obj);
	cs->xpath_obj = NULL;
	return ESUCCESS;
}



NETINFO_API int configuration_search_free (ConfigSearch **csp)
{
	ConfigSearch	*cs = NULL;
	if (csp == NULL)
		return EINVAL;
	cs = *csp;
	if (cs->xpath_obj != NULL)
		xmlXPathFreeObject (cs->xpath_obj);
	if (cs->xpath_ctx != NULL)
		xmlXPathFreeContext (cs->xpath_ctx);
	free (cs);
	*csp = NULL;
	return ESUCCESS;
}


NETINFO_API int configuration_item_new (Configuration *c, ConfigItem **cip, char *name)
{
	xmlNodePtr	node = NULL;
	if (c == NULL || cip == NULL || name == NULL)
		return EINVAL;
	node = xmlNewNode (NULL, (unsigned char *)name);
	if (node == NULL)
		return ENOMEM;
	*cip = (ConfigItem *)node;
	return ESUCCESS;
}

NETINFO_API char *configuration_item_get_attribute (ConfigItem *ci,
					char *attribute)
{
	if (ci == NULL || attribute == NULL)
		return NULL;
	/* The caller MUST free the returned string */
	return (char *)xmlGetProp ((xmlNodePtr)ci,
				(unsigned char *)attribute);
}


NETINFO_API int configuration_item_add_attribute (ConfigItem *ci,
					char *attribute,
					char *value)
{
	int		rc;
	xmlAttrPtr	attr = NULL;
	if (ci == NULL || attribute == NULL || value == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (NULL,
						ci,
						CONFIG_OP_ITEM_ATTR_ADD);
	}
	attr = xmlNewProp ((xmlNodePtr)ci,
			(unsigned char *)attribute,
			(unsigned char *)value);
	if (attr == NULL)
		return ENOMEM;
	return ESUCCESS;
}


NETINFO_API int configuration_item_set_attribute (ConfigItem *ci,
					char *attribute,
					char *value)
{
	int		rc;
	xmlAttrPtr	attr = NULL;
	if (ci == NULL || attribute == NULL || value == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (NULL,
						ci,
						CONFIG_OP_ITEM_ATTR_CHANGE);
	}
	attr = xmlSetProp ((xmlNodePtr)ci,
			(unsigned char *)attribute,
			(unsigned char *)value);
	if (attr == NULL)
		return ENOMEM;
	return ESUCCESS;
}


NETINFO_API int configuration_item_remove_attribute (ConfigItem *ci,
					char *attribute)
{
	int		rc;
	xmlAttrPtr	attr = NULL;
	if (ci == NULL || attribute == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (NULL,
						ci,
						CONFIG_OP_ITEM_ATTR_REMOVE);
	}
	attr = xmlHasProp ((xmlNodePtr)ci, (unsigned char *)attribute);
	if (attr == NULL)
		return ENOENT;
	rc = xmlRemoveProp (attr);
	if (rc == -1)
		return errno;
	return ESUCCESS;
}


NETINFO_API char *configuration_item_get_content (ConfigItem *ci)
{
	if (ci == NULL)
		return NULL;
	/* The caller MUST free the returned string */
	return (char *)xmlNodeGetContent ((xmlNodePtr)ci);
}


NETINFO_API int configuration_item_set_content (ConfigItem *ci,
					char *content)
{
	int		rc;
	xmlChar		*ec = NULL;
	if (ci == NULL || content == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (NULL,
						ci,
						CONFIG_OP_ITEM_CONTENT_CHANGE);
	}
	ec = xmlEncodeSpecialChars (NULL, (unsigned char *)content);
	if (ec == NULL)
		return EINVAL;
	xmlNodeSetContent ((xmlNodePtr)ci, ec);
	xmlFree (ec);
	return ESUCCESS;
}


NETINFO_API int configuration_item_add_child (ConfigItem *ci,
					ConfigItem *cichild)
{
	int		rc;
	xmlNodePtr	node = NULL;
	if (ci == NULL || cichild == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (NULL,
						cichild,
						CONFIG_OP_ITEM_ADD);
	}
	node = (xmlNodePtr)ci;
	if (xmlAddChild (node, (xmlNodePtr)cichild) == NULL)
		return EINVAL;
	return ESUCCESS;
}


NETINFO_API ConfigItem *configuration_item_get_first_child (ConfigItem * ci)
{
	xmlNodePtr n;
	if (ci == NULL)
		return NULL;
	n = (xmlNodePtr)ci;
	n = n->children;
	while (n != NULL && n->type != XML_ELEMENT_NODE)
		n = n->next;
	return (ConfigItem *)n;
}


NETINFO_API ConfigItem *configuration_item_get_next_sibling (ConfigItem * ci)
{
	xmlNodePtr n = (xmlNodePtr)ci;
	if (ci == NULL)
		return NULL;
	while (n->next != NULL && n->next->type != XML_ELEMENT_NODE)
		n = n->next;
	return (ConfigItem *)n->next;
}


NETINFO_API int configuration_set_root_item (Configuration *c,
					ConfigItem *ci)
{
	int		rc;
	xmlNodePtr	node = NULL;
	xmlNodePtr	oldroot = NULL;
	if (c == NULL || c->xml_doc == NULL || ci == NULL)
		return EINVAL;
	if (configuration_block.callback != NULL) {
		rc = (configuration_block.callback) (NULL,
						ci,
						CONFIG_OP_ITEM_ADD);
	}
	node = (xmlNodePtr)ci;
	oldroot = xmlDocSetRootElement (c->xml_doc, node);
	if (oldroot != NULL) {
	}

	return ESUCCESS;
}




