/************************************************************************/
/* fsobject.c - Filesystem Object Library Source 			*/
/************************************************************************/
/*
 * (c) 2002-2003 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/



#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_POSIXACL
#include <sys/acl.h>
#endif
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "fsobject.h"

#ifdef WIN32
#define	lstat	stat
#endif

/************************************************************************/
/* Function Prototypes							*/
/************************************************************************/
FSObject *object_alloc(void);
uint32 get_object_type_posix(mode_t mode);

#ifdef HAVE_POSIXACL
FSObjectAcl *pacl_alloc(void);
FSObjectAcl *pacl_alloc_setup(uint32 tag, uint32 perm, uint32 id);
int pacl_free(FSObjectAcl **acl);
#endif

/************************************************************************/
/* Portability								*/
/************************************************************************/


/************************************************************************/
/* API Functions							*/
/************************************************************************/

int fsobject_stat(char *objectpath, FSObject **pfso) {
	FSObject *fso = NULL;
	int frc = 0;
	int rc;
	struct stat s;

	if((objectpath == NULL) || (pfso == NULL))
		return(-1);

	rc = lstat(objectpath, &s);
	if(rc != 0) {
		frc = rc;
		return(frc);
		}

	fso = object_alloc();
	if(fso == NULL) {
		frc = ENOMEM;
		return(frc);
		}
	fso->object_type = s.st_mode & 0xfe00;
	fso->object_path = strdup(objectpath);
	fso->object_type = get_object_type_posix(s.st_mode);
	fso->object_mode = s.st_mode & 0x01ff;
	fso->object_links = s.st_nlink;
	fso->object_size = (uint64)s.st_size;
#ifndef WIN32
	fso->object_blocksize = (uint64)s.st_blksize;
	fso->object_blocks = (uint64)s.st_blocks;
#else
	fso->object_blocksize = 0;
	fso->object_blocks = 0;
#endif
	fso->object_atime = (uint64)s.st_atime;
	fso->object_mtime = (uint64)s.st_mtime;
	fso->object_ctime = (uint64)s.st_ctime;
	rc = fsobject_get_acllist(objectpath, &fso->object_acllist);
	if(rc != 0 && rc != ENOSYS) {
		fs_object_free(&fso);
		frc = rc;
		}

	if(frc == 0) {
		*pfso = fso;
		}
	else {
		*pfso = NULL;
		}
	return(frc);
	}

#ifndef HAVE_POSIXACL
int fsobject_get_acllist(char *objectpath, FSObjectAcl **pfsoa) {
	int frc = ENOSYS;
	if(pfsoa != NULL)
		*pfsoa = NULL;
	return(frc);
	}

#else
int fsobject_get_acllist(char *objectpath, FSObjectAcl **pfsoa) {
	int aclcount;
	int count = 0;
	int frc = 0;
	int rc;
	int entryid = ACL_FIRST_ENTRY;

	acl_t acl = NULL;
	acl_entry_t acle;
	acl_permset_t permset;

	FSObjectAcl *fsoa = NULL;
	uint32 *qualifier = NULL;

	if((objectpath != NULL) && (pfsoa != NULL)) {
		acl = acl_get_file(objectpath, ACL_TYPE_ACCESS);
		if(acl != NULL) {
			aclcount = 0;
			do {
				rc = acl_get_entry(acl, entryid, &acle);
				if(rc == 1) {
					entryid = ACL_NEXT_ENTRY;
					aclcount++;
					}
				else if(rc == 0) {
					break;
					}
				else {
					frc = errno;
					*pfsoa = NULL;
					acl_free(acl);
					return(frc);
					}
				} while(1);
			fsoa = (FSObjectAcl *)calloc(count+1, sizeof(FSObjectAcl));
			if(fsoa != NULL) {
				entryid = ACL_FIRST_ENTRY;
				count = 0;
				while(count < aclcount) {
					rc = acl_get_entry(acl, entryid, &acle);
					if(rc == 1) {
						entryid = ACL_NEXT_ENTRY;
						acl_get_tag_type(acle, &fsoa[count].acl_tag);
						acl_get_permset(acle, &permset);
						fsoa[count].acl_perm = (uint32)permset;
						qualifier = (uint32 *)acl_get_qualifier(acle);
						fsoa[count].acl_id = (uint64)*qualifier;
						fsoa[count].acl_idname = NULL; /*TODO */
						acl_free(qualifier);
						}
					else if(rc == 0) {
						break;
						}
					else {
						frc = errno;
						free(fsoa);
						acl_free(acl);
						*pfsoa = NULL;
						return(frc);
						}
					count++;
					}
				}
			}
		}

	if(frc == 0) {
		*pfsoa = fsoa;
		}
	else {
		*pfsoa = NULL;
		}
	acl_free(acl);
	return(frc);
	}
#endif
		

/************************************************************************/
/* Internal Functions							*/
/************************************************************************/
#ifdef HAVE_POSIXACL
FSObjectAcl *pacl_alloc(void) {
	FSObjectAcl *fsoa = NULL;
	fsoa = (FSObjectAcl *)calloc(1, sizeof(FSObjectAcl));
	return(fsoa);
	}



FSObjectAcl *pacl_alloc_setup(uint32 tag, uint32 perm, uint32 id) {
	FSObjectAcl *fsoa = NULL;
	fsoa = pacl_alloc();
	if(fsoa != NULL) {
		fsoa->acl_tag = tag;
		fsoa->acl_perm = perm;
		fsoa->acl_id = id;
		fsoa->acl_idname = NULL;
		}
	return(fsoa);
	}


int pacl_free(FSObjectAcl **fsoa) {
	if(fsoa == NULL)
		return(1);
	if(*fsoa == NULL)
		return(2);
	free(*fsoa);
	*fsoa = NULL;
	return(0);
	}
#endif


FSObject *object_alloc(void) {
	FSObject *fso = NULL;
	fso = (FSObject *)calloc(1, sizeof(FSObject));
	return(fso);
	}



int fs_object_free(FSObject **fso) {
	FSObject *fsos;
	if(fso == NULL)
		return(1);
	if(*fso == NULL)
		return(2);
	fsos = *fso;
	if(fsos->object_path)
		free(fsos->object_path);
	free(fsos);
	*fso = NULL;
	return(0);
	}


uint32 get_object_type_posix(mode_t mode) {
#ifdef HAVE_POSIXACL
	if(S_ISSOCK(mode))
		return OBJECT_TYPE_SOCKET;
	if(S_ISLNK(mode))
		return OBJECT_TYPE_SYMLINK;
#endif
	if(S_ISREG(mode))
		return OBJECT_TYPE_FILE;
	if(S_ISBLK(mode))
		return OBJECT_TYPE_BLOCKDEV;
	if(S_ISDIR(mode))
		return OBJECT_TYPE_DIRECTORY;
	if(S_ISCHR(mode))
		return OBJECT_TYPE_CHARDEV;
	if(S_ISFIFO(mode))
		return OBJECT_TYPE_FIFO;
	return 0;
	}		

