/************************************************************************/
/* fstraverse.c - Filesystem Traversal Library Source 			*/
/************************************************************************/
/*
 * (c) 2002-2003 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

/* for dirname(), not a POSIX function */
#ifndef WIN32
#include <libgen.h>
#else
#include "dirent.h"
#endif


#include "fstraverse.h"
#include "fsobject.h"
#include "stack.h"



/************************************************************************/
/* Function Prototypes (internal)					*/
/************************************************************************/
int fs_trav_descend(FSTrav *fst);
int fs_trav_ascend(FSTrav *fst);

FSTrav *trav_alloc(void);
int trav_free(FSTrav **trav);
#ifdef WIN32
char *dirname (char *path);
#endif


/************************************************************************/
/* API Functions							*/
/************************************************************************/
FSTrav * fs_trav_open(char *path, uint32 flags) {
	FSTrav *fst = NULL;
	
	if(path == NULL)
		goto OPEN_ERROR;
	fst = trav_alloc();
	if(fst == NULL)
		goto OPEN_ERROR;
	if(getcwd(fst->workpath, MAX_PATH_LEN) == NULL)
		goto OPEN_ERROR;
	fst->impl_data = (void *)opendir(path);
	if(fst->impl_data == NULL)
		goto OPEN_ERROR;
	strcat(fst->currpath, path);
	fst->flags = flags & 0x0000ffff; /* bottom 16 bits = user flags */
	fst->stack = sstack_new();
	if(fst->stack == NULL)
		goto OPEN_ERROR;
	return(fst);
OPEN_ERROR:
	if(fst->impl_data)
		closedir((DIR *)fst->impl_data);
	if(fst)
		trav_free(&fst);
		
	return(NULL);
	}


int fs_trav_next_entry(FSTrav *fst, FSObject **fso)
{
	FSObject *o = NULL;
	char fullpath[MAX_PATH_LEN+1];
	int frc = 0;
	int rc;
	char *path;
	struct dirent *de = NULL;

	fullpath[0] = 0x00;
	if((fst != NULL) && (fso != NULL)) {
		if(fst->flags & FST_FLAG_LASTENTRYDIR) {
			/* Last entry was a dir... decend! */
			if (fst->flags & FST_FLAG_RECURSIVE) {
				rc = fs_trav_descend(fst);
				if(rc != 0) {
					if(rc == EACCES) {
						/* The directory entry was already */
						/* pasted to the user program, so */
						/* if we get an error we should */
						/* continue to provide entries from */
						/* the parent directory. */
						path = dirname(strdup(fst->currpath));
						strcpy(fst->currpath, path);
						free(path);
					} else {
						frc = rc;
						goto NEXT_ENTRY_ERROR;
					}
				}
			} else {
				path = dirname(strdup(fst->currpath));
				strcpy(fst->currpath, path);
				free(path);
			}
		}
READDIR:
		de = readdir((DIR *)fst->impl_data);
		if(de != NULL) {
			if( !(fst->flags & FST_FLAG_SHOWHIDDEN) && 
					de->d_name[0] == '.')
				goto READDIR;
			strcpy(fullpath, fst->currpath);
			strcat(fullpath, DIRDELIMITER);
			strcat(fullpath, de->d_name);

			rc = fsobject_stat(fullpath, &o);
			if(rc != 0) {
				frc = rc;
				goto NEXT_ENTRY_ERROR;
			}
			if(o->object_type == OBJECT_TYPE_DIRECTORY) {
				fst->flags |= FST_FLAG_LASTENTRYDIR;
				strcpy(fst->currpath, fullpath);
				frc = 0; /* Must set this (after ascend()) */
			} else {
				fst->flags -= (fst->flags & FST_FLAG_LASTENTRYDIR);
				frc = 0; /* Must set this (after ascend()) */
			}
		} else {
			frc = errno;
			if(sstack_items(fst->stack) > 0) {
				/* end of this directory, back up! */
				rc = fs_trav_ascend(fst);
				if(rc != 0) {
					frc = rc;
					goto NEXT_ENTRY_ERROR;
				}
				goto READDIR;
			} else {
				frc = EINVAL;
			}
		}
		if(frc == 0) {
			*fso = o;
		} else {
			*fso = NULL;
		}
	        return(frc);
	}

NEXT_ENTRY_ERROR:
	return(frc);
}


int fs_trav_descend(FSTrav *fst) {
	int rc;
	DIR *newdir;
	if(fst == NULL)
		return(EINVAL);
	newdir = (void *)opendir(fst->currpath);
	if(newdir == NULL) {
		rc = errno;
		return(rc);
		}
	sstack_push(fst->stack, fst->impl_data);
	fst->impl_data = (void *)newdir;
	return(0);
	}



int fs_trav_ascend(FSTrav *fst) {
	char *path;

	if(fst == NULL)
		return(EINVAL);
	closedir((DIR *)fst->impl_data);
	fst->impl_data = sstack_pop(fst->stack);
	path = dirname(strdup(fst->currpath));
	strcpy(fst->currpath, path);
	free(path);
	return(0);
	}



int fs_trav_close(FSTrav *fst) {
	if(fst == NULL)
		return(EINVAL);
	while (fst->impl_data != NULL) {
		closedir(fst->impl_data);
		fst->impl_data = sstack_pop(fst->stack);
		}
	trav_free (&fst);
	return(0);
	}


int fs_trav_set_callback(FSTrav *fst, FSTravCallback *cb) {
	return(0);
	}


int fs_trav_traverse(FSTrav *fst) {
	return(0);
	}


FSTrav *trav_alloc(void) {
	FSTrav *fst = NULL;
	fst = (FSTrav *)calloc(1, sizeof(FSTrav));
	if(fst == NULL)
		goto TRAVALLOC_ERROR;
	fst->currpath = (char *)calloc(1, MAX_PATH_LEN+1);
	if(fst->currpath == NULL)
		goto TRAVALLOC_ERROR;
	fst->workpath = (char *)calloc(1, MAX_PATH_LEN+1);
	if(fst->workpath == NULL)
		goto TRAVALLOC_ERROR;
	return(fst);
TRAVALLOC_ERROR:
	if(fst) {
		if(fst->currpath)
			free(fst->currpath);
		if(fst->workpath)
			free(fst->workpath);
		free(fst);
		}
	return(NULL);
	}



int trav_free(FSTrav **fst) {
	FSTrav *fsts;
	if(fst == NULL)
		return(1);
	if(*fst == NULL)
		return(2);
	fsts = *fst;
	if(fsts->currpath)
		free(fsts->currpath);
	if(fsts->workpath)
		free(fsts->workpath);
	if (fsts->stack)
		sstack_free (fsts->stack);
	free(fsts);
	*fst = NULL;
	return(0);
	}


#ifdef WIN32
char *dirname (char *path) {
    int i;
    if (path == NULL)
        return(NULL);
    i = strlen(path);
    i--;
    while (i>=0) {
        if (path[i] == '\\') {
            path [i] = 0;
            return(path);
        }
        i--;
    }
    path[0] = '.';
    path[1] = 0;
    return(path);    
}
#endif
