/************************************************************************/
/* ni_dir.c - Netinfo Directory monitoring plugin			*/
/************************************************************************/
/*
 *  (c)2005 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include <cutil/slist.h>

#include <netinfo/netinfo_portable.h>
#include <netinfo/netinfo.h>
#include <netinfo/nmodule.h>
#include <netinfo/conffile.h>

#include <fstraverse.h>
#include <fsobject.h>


#define	MODULE_NAME			"dir"
#define	MODULE_VERSION			"0.1.1"

#ifndef WIN32
#define DEFAULT_CONFIGFILE		"passive.conf"
#define DIR_DELIMITER			"/"
#else
#define DEFAULT_CONFIGFILE		"passive.conf"
#define DIR_DELIMITER			"\\"
#define sleep(x)			Sleep (1000 * x)
#endif


#define BUFFER_SIZE			256

#define POLL_INTERVAL_DEFAULT		60

#define CHECK_INTERVAL_DEFAULT		5

#define	DIR_FLAG_WARNMISSING		0x00000001


typedef struct	_DirData {
	char		*dirpath;
	char		*objectname;
	u_int32_t	flags;
	u_int32_t	checkinterval;
	u_int32_t	stepinterval;
	u_int32_t	warnentries;
	u_int32_t	critentries;
	u_int32_t	lastentries;
	u_int32_t	laststate;
	double		changepct;
} DirData;



typedef struct _ModuleData {
	ModuleInfo	*moduleinfo;
	u_int32_t	pollinterval;
	u_int32_t	curinterval;
	pthread_t	polltid;
	SList		*dirlist;
} ModuleData;


/* Global Variables */
ModuleData	module_data;


/* Function Prototypes */
void *poll_thread (void *arg);
int parse_configfile (char *filepath);


int netinfo_module_init (ModuleInfo *mi)
{
	int rc;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	memset (&module_data, 0, sizeof (ModuleData));
	module_data.moduleinfo = mi;
	module_data.moduleinfo->name = MODULE_NAME;
	module_data.pollinterval = POLL_INTERVAL_DEFAULT;
	module_data.curinterval = 0;

	rc = parse_configfile (SYSCONFDIR DIR_DELIMITER DEFAULT_CONFIGFILE);
	rc = pthread_create (&module_data.polltid,
				NULL,
				poll_thread,
				NULL);
	if (rc == 0) {
		vector_log_entry (mi,
				"Started poll thread (tid=%d) interval=%d\n",
				module_data.polltid,
				module_data.pollinterval);
	} else {
		vector_log_entry (mi,
				"Failed to start poll thread.\n");
		return NETINFO_RC_MODULEERROR;
	}
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_close (ModuleInfo *mi)
{
	int		rc;
	SList		*ditem = NULL;
	DirData		*dd = NULL;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	ditem = module_data.dirlist;
	while (ditem != NULL) {
		dd = (DirData *)slist_get_data (ditem);
		if (dd != NULL) {
			if (dd->dirpath != NULL)
				free (dd->dirpath);
			if (dd->objectname != NULL)
				free (dd->objectname);
			free (dd);
		}
		ditem = slist_get_next (ditem);
	}
	rc = slist_free (&module_data.dirlist);
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_execute_method (char *method,
					NetinfoHeader *nh,
					char **responsedata,
					unsigned long *responsedatalen)
{
	int rc;
	int len;
	int argdatalen;
	char argdata[NETINFO_HEADER_ARGDATA_SIZE];
	char buffer[BUFFER_SIZE];
	char *rd = NULL;
	if (method == NULL ||
			nh == NULL ||
			responsedata == NULL ||
			responsedatalen == NULL)
		return NETINFO_RC_INVALIDARG;
	*responsedata = NULL;
	*responsedatalen = 0;
	argdatalen = NETINFO_HEADER_ARGDATA_SIZE;
	rc = netinfo_header_get_argdata (nh,
				argdata,
				&argdatalen);
	if (!strcmp (method, "curinterval")) {
			len = sprintf (buffer,
					"%lu",
					(long)module_data.curinterval);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
	} else {
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	return NETINFO_RC_SUCCESS;
}


void *poll_thread (void *arg)
{
	int		rc;
	int		eventrc;
	int		len;
	u_int32_t	direntries;
	double		change;
	char		buffer[BUFFER_SIZE];
	struct stat	st;
	SList		*ditem = NULL;
	DirData		*dd = NULL;
        EventData       *event = NULL;
	EventChain	*eventchain = NULL;
	FSTrav		*fst = NULL;
	FSObject	*fso = NULL;
	ModuleData	*md = &module_data;

	sleep (2); /*FIXME:  module is loaded before netinfod can initialize SSL context */
	
        rc = eventdata_new (&event);
	eventdata_clear (event);
	md->curinterval = 0;
	direntries = 0;
	while (1) {
		ditem = md->dirlist;
		while (ditem != NULL) {
			dd = slist_get_data (ditem);
			dd->stepinterval++;
			if (dd->stepinterval == dd->checkinterval) {
				if (eventchain == NULL) {
					rc = vector_eventchain_open (module_data.moduleinfo,
								&eventchain);
				}
				direntries = 0;
				eventrc = 0;
				rc = stat (dd->dirpath, &st);
				if (rc != 0) {
					rc = errno;
					eventrc = 2;
					len = snprintf (buffer,
						BUFFER_SIZE,
						"%s: %s\n",
						dd->dirpath,
						strerror (rc));
				} else {
					fst = fs_trav_open (dd->dirpath, 0);
					if (fst != NULL) {
						while (1) {
							rc = fs_trav_next_entry (fst, &fso);
							if (rc != 0)
								break;
							direntries++;
							rc = fs_object_free (&fso);
						}
						rc = fs_trav_close (fst);
						if (dd->warnentries != 0 && 
							direntries >= dd->warnentries)
							eventrc = 1;
						if (dd->critentries != 0 &&
							direntries >= dd->critentries)
							eventrc = 2;
						if (dd->changepct != 0.0) {
							change = direntries - dd->lastentries;
							if (change < 0.0)
								change = change * -1.0;
							change = (change / dd->warnentries) * 100.0;
							if (change > dd->changepct)
								if (eventrc < 2)
								eventrc = 1;
						}
						len = snprintf (buffer,
							BUFFER_SIZE,
							"%s: entries=%lu\n",
							dd->dirpath,
							direntries);
					} else {
	vector_log_entry (module_data.moduleinfo,
		"trav_open failed for dir %s rc=%d(%s)\n", dd->dirpath);
						eventrc = 2;
						len = snprintf (buffer,
							BUFFER_SIZE,
							"%s: Failed to open directory for traversing.\n",
							dd->dirpath);
					}
				}
				if (eventrc != dd->laststate) {
					eventdata_clear (event);
					rc = eventdata_set (event,
								EVENT_FLAG_ONESHOT,
								eventrc,
								dd->objectname,
								buffer,
								len);
					rc = vector_eventchain_emit_event (eventchain,
								event);
				}
				dd->stepinterval = 0;
				dd->lastentries = direntries;
				dd->laststate = eventrc;
			}
			ditem = slist_get_next (ditem);
		}
		if (eventchain != NULL) {
			rc = vector_eventchain_close (&eventchain);
		}
		md->curinterval++;
		sleep (POLL_INTERVAL_DEFAULT);
	}
	eventdata_free (&event);
	pthread_exit (NULL);
	return NULL;
}


int parse_configfile (char *filepath)
{
	int		rc;
	int		i;
	char		buffer[BUFFER_SIZE];
	char		*directive = NULL;
	char		*attribs = NULL;
	char		*attrib = NULL;
	ConfigFile	*cf = NULL;
	DirData		*dd = NULL;
	if (filepath == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = configfile_open (&cf, filepath);
	if (rc != NETINFO_RC_SUCCESS)
		return rc;
	rc = configfile_search_begin (cf, "WatchDir");
	while (1) {
		rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		directive = strtok (buffer, "'\" \t");
		if (directive[0] == '#') {
			continue;
		}
		
		dd = calloc (1, sizeof (DirData));
		dd->dirpath = strdup (strtok (NULL, "'\"\t"));
		dd->checkinterval = CHECK_INTERVAL_DEFAULT;
		attribs = strtok (NULL, " \t\r\n");
		attrib = strtok (attribs, "= \t\r\n");
		while (attrib != NULL) {
			if (!strcmp (attrib, "objectname"))
				dd->objectname = strdup (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "checkinterval"))
				dd->checkinterval = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "warnentries"))
				dd->warnentries = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "critentries"))
				dd->critentries = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "changepct"))
				dd->changepct = atof (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "warnmissing")) {
				i = atol (strtok (NULL, ", \t\r\n"));
				if (i != 0)
					dd->flags |= DIR_FLAG_WARNMISSING;
			}
			attrib = strtok (NULL, "= \t\r\n");
		}
		vector_log_entry (module_data.moduleinfo,
				"Watching dir='%s' checkinterval=%d warnentries=%d critentries=%d changepct=%2.2f\n",
				dd->dirpath,
				dd->checkinterval,
				dd->warnentries,
				dd->critentries,
				dd->changepct);
		rc = slist_item_add (&module_data.dirlist, dd);
	}
	rc = configfile_search_end (cf);
	rc = configfile_close (&cf);
	return NETINFO_RC_SUCCESS;
}


