/************************************************************************/
/* ni_linux.c - Netinfo Linux monitoring plugin				*/
/************************************************************************/
/*
 *  (c)2005,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/

#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <pthread.h>
#include <sys/statvfs.h>
#include <float.h>

#include <cutil/cutil.h>

#include <netinfo/netinfo.h>
#include <netinfo/nmodule.h>
#include <netinfo/conffile.h>
#include <netinfo/event.h>

#include <average.h>

#define	MODULE_NAME			"linux"
#define MODULE_VERSION			"0.1.8"

#define PASSIVE_CONFIG_FILE		"/passive.conf"

#define RC_OKAY				0
#define RC_WARNING			1
#define RC_CRITICAL			2
#define RC_UNKNOWN			3


#define BUFFER_SIZE			256

#define POLL_FLAG_RELOADCONFIG		0x00000001

#define POLL_INTERVAL_DEFAULT		60
#define DEFAULT_AVERAGE_INTERVALS       5
#define DEFAULT_ALERT_AVG_INTERVALS	15


#define DEFAULT_DISK_WARNMAX_PERCENT    80
#define DEFAULT_DISK_CRITMAX_PERCENT    90
#define DEFAULT_DISK_OBJECTNAME         "MOUNT"

#define SYSTEM_CPU_OBJECT       	"SYSCPU"
#define	SYSTEM_CPU_MAXTOTAL_DEFAULT	75.00

#define SYSTEM_VM_OBJECT       		"SYSVM"
#define	SYSTEM_VM_SWAP_CRITPCT_DEFAULT	50.00
#define	SYSTEM_VM_FREE_CRITPCT_DEFAULT	10.00

#define SYSTEM_LOAD_OBJECT       	"SYSLOAD"

#define	COUNTER_TYPE_GAUGE		0x00000001
#define COUNTER_TYPE_COUNTER		0x00000002

#define	DEVNAME_SIZE			32


typedef struct	_CheckInterval	CheckInterval;
struct	_CheckInterval {
	int		seconds;	/* seconds per interval */
	int		required;	/* # of required intervals */
	int		current;	/* current interval # */
};


typedef struct _DiskData {
        char                    *objectname;
        char                    *path;
        double                  warnmax;
        double                  critmax;
	u_int32_t		lastrc;
} DiskData;


typedef struct _CpuStats {
	u_int64_t	u;
	u_int64_t	n;
	u_int64_t	s;
	u_int64_t	i;
	u_int64_t	w;
	u_int64_t	x;
	u_int64_t	y;
} CpuStats;


typedef struct _CpuPctStats {
	double		u;
	double		n;
	double		s;
	double		i;
	double		w;
	double		x;
	double		y;
} CpuPctStats;


typedef struct _SystemCpuData SystemCpuData;
struct _SystemCpuData {
	double		maxuser;
	double		maxsystem;
	double		maxwait;
	double		maxx;
	double		maxy;
	double		maxtotal;
	u_int32_t	alertavgintervals;
	u_int32_t	alertavgcurinterval;
	u_int32_t	currentstate;
	u_int32_t	curstatsid;
	u_int32_t	oldstatsid;
	CpuStats	cpustats[2];
	Average		*alertavgu;
	Average		*alertavgn;
	Average		*alertavgs;
	Average		*alertavgi;
	Average		*alertavgw;
	Average		*alertavgx;
	Average		*alertavgy;
	u_int32_t	lastrc;
};


typedef struct _SystemVMData	SystemVMData;
struct _SystemVMData {
	double		memfreeminpctwarn;
	double		memfreeminpctcrit;
	double		swapinusemaxpctwarn;
	double		swapinusemaxpctcrit;
	u_int32_t	swappginsecwarn;
	u_int32_t	swappginseccrit;
	u_int32_t	swappgoutsecwarn;
	u_int32_t	swappgoutseccrit;
	u_int32_t	lastrc;
};

typedef struct _VMStats	VMStats;
struct _VMStats {
	u_int64_t	memtotal;
	u_int64_t	memfree;
	u_int64_t	meminuse;
	u_int64_t	memshared;
	u_int64_t	membuffers;
	u_int64_t	memcached;
	u_int64_t	swaptotal;
	u_int64_t	swapfree;
	u_int64_t	swapinuse;

	u_int64_t	memactive;
	u_int64_t	meminactive;
	u_int64_t	hightotal;
	u_int64_t	highfree;
	u_int64_t	highinuse;
	u_int64_t	lowtotal;
	u_int64_t	lowfree;
	u_int64_t	lowinuse;
	u_int64_t	bptotal;
	u_int64_t	bpfree;
	u_int64_t	bpinuse;
};


typedef struct _SystemLoadData	SystemLoadData;
struct _SystemLoadData {
	u_int32_t	cpucount;
	double		warningload;
	double		criticalload;
	double		priorload[3];
	u_int32_t	lastrc;
};


typedef struct	_Counter	Counter;
struct _Counter {
	u_int32_t	flags;
	u_int32_t	prevtime;
	double		prevval;
	u_int32_t	currtime;
	double		currval;
};


typedef struct _InterfaceStats	InterfaceStats;
struct	_InterfaceStats {
	char		*devname;
	char		*objectname;
	u_int32_t	lastupdate;
	Counter		rxbytes;
	Counter		rxpkts;
	Counter		rxerrors;
	Counter		rxdrops;
	Counter		rxfifo;
	Counter		rxframe;
	Counter		rxcomp;
	Counter		rxmcast;
	Counter		txbytes;
	Counter		txpkts;
	Counter		txerrors;
	Counter		txdrops;
	Counter		txfifo;
	Counter		txcolls;
	Counter		txcarr;
	Counter		txcomp;
	double		*samrxbytes;
	double		*samrxpkts;
	double		*samrxerrors;
	double		*samrxdrops;
	double		*samrxfifo;
	double		*samrxframe;
	double		*samrxcomp;
	double		*samrxmcast;
	double		*samtxbytes;
	double		*samtxpkts;
	double		*samtxerrors;
	double		*samtxdrops;
	double		*samtxfifo;
	double		*samtxcolls;
	double		*samtxcarr;
	double		*samtxcomp;
};


// device per-collection structure
struct diskstat_t {
	uint32_t	reads;
	uint32_t	readsmerged;
	uint32_t	readsectors;
	uint32_t	readms;
	uint32_t	writes;
	uint32_t	writesmerged;
	uint32_t	writesectors;
	uint32_t	writems;
	uint32_t	iosoutstanding;
	uint32_t	ioswaitms;
	uint32_t	iocompletionms;
};


// per-device structure
struct diskstatinfo_t {
	uint32_t		major;
	uint32_t		minor;
	char			name[DEVNAME_SIZE];
	int			current; // either 0 or 1
	struct diskstat_t	ds[2];
};



typedef struct _ModuleData {
	ModuleInfo	*moduleinfo;
	FILE		*meminfo_fd;
	FILE		*stat_fd;
	int		pollinterval;
	u_int32_t	avgintervals;
	u_int32_t	curinterval;
	pthread_t	polltid;
	CheckInterval	cpu_perf_interval;	/* the interval for CPU perf */
	CheckInterval	cpu_alert_interval;	/* the interval for CPU alert */
	CpuStats	cpulast;
	CpuPctStats	*cpusampleavg;
/*	CpuPctStats	*cpusamplealert;
	u_int32_t	cpualertinterval; */
	SystemCpuData	cpudata;
	SystemVMData	vmdata;
	SystemLoadData	loaddata;
	SList		*interfacelist;
	SList		*disklist;
	EventChain	*eventchain;
	EventData	*event;
	pthread_mutex_t loopmutex;
	pthread_cond_t  loopcondition;

	// device statistics variables
	AVLTree			*dev_tree;
	SList			*dev_list;
} ModuleData;


ModuleData	module_data;

void *poll_thread (void *arg);
u_int32_t sum_ticks (CpuStats *cs);
double calc_percent (u_int64_t part, u_int64_t total);
int file_find_line_with (FILE *fd,
			char *searchstr,
			char *buffer,
			int *bufferlen);
/* int system_counters_status (void); */
void counter_init (Counter *counter, u_int32_t type);
void counter_clear (Counter *counter);
void counter_push_value (Counter *counter, double value);
double counter_get_value (Counter *counter);
double counter_value_push_and_get (Counter *counter, double value);
InterfaceStats *interfacestats_new (int samples);
int interfacestats_free (InterfaceStats *is);
int interfaces_initialize (void);
int interfaces_collect_stats (void);
InterfaceStats	*interfacelist_get_interface (SList *list, char *ifname);
int interface_stats_report (void);
double samples_average (double *samplelist, int samples);
int passive_config_parse (char *filepath);
int disks_scan(void);
int disks_check_status (SList *disklist);
int disks_free (void);
int vm_collect_statistics (void);
int load_initialize (void);
int load_collect_statistics (void);
int cpu_collect_statistics (void);
int diskstatinfo_compare (struct diskstatinfo_t *dsi1, struct diskstatinfo_t *dsi2);
int diskstatinfo_compare_name (char *devname, struct diskstatinfo_t *dsi2);
int diskstatinfo_calculate_stats (struct diskstatinfo_t *dsi, struct diskstat_t *ds);
int diskstats_collect (void);


int netinfo_module_init (ModuleInfo *mi)
{
	int	rc;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	memset (&module_data, 0, sizeof (ModuleData));
	module_data.moduleinfo = mi;
	module_data.moduleinfo->name = MODULE_NAME;
	module_data.avgintervals = DEFAULT_AVERAGE_INTERVALS;

	module_data.cpu_perf_interval.required = DEFAULT_AVERAGE_INTERVALS;
	module_data.cpu_alert_interval.required = DEFAULT_ALERT_AVG_INTERVALS;

	module_data.cpudata.maxtotal = SYSTEM_CPU_MAXTOTAL_DEFAULT;
	module_data.cpusampleavg = calloc (module_data.avgintervals,
					sizeof (CpuPctStats));
	if (module_data.cpusampleavg == NULL)
		return NETINFO_RC_MEMORYERROR;

	module_data.dev_list = NULL;
	rc = avltree_new (&module_data.dev_tree, diskstatinfo_compare, NULL);

	rc = eventdata_new (&module_data.event);
	if (rc != NETINFO_RC_SUCCESS) {
		free (module_data.cpusampleavg);
		return rc;
	}
	module_data.cpudata.alertavgintervals = DEFAULT_ALERT_AVG_INTERVALS;
	module_data.cpudata.oldstatsid = 1;
	module_data.cpudata.alertavgu = average_new (AVERAGE_TYPE_DOUBLE,
						module_data.cpudata.alertavgintervals);
	module_data.cpudata.alertavgn = average_new (AVERAGE_TYPE_DOUBLE,
						module_data.cpudata.alertavgintervals);
	module_data.cpudata.alertavgs = average_new (AVERAGE_TYPE_DOUBLE,
						module_data.cpudata.alertavgintervals);
	module_data.cpudata.alertavgi = average_new (AVERAGE_TYPE_DOUBLE,
						module_data.cpudata.alertavgintervals);
	module_data.cpudata.alertavgw = average_new (AVERAGE_TYPE_DOUBLE,
						module_data.cpudata.alertavgintervals);
	module_data.cpudata.alertavgx = average_new (AVERAGE_TYPE_DOUBLE,
						module_data.cpudata.alertavgintervals);
	module_data.cpudata.alertavgy = average_new (AVERAGE_TYPE_DOUBLE,
						module_data.cpudata.alertavgintervals);
	module_data.vmdata.swapinusemaxpctcrit = SYSTEM_VM_SWAP_CRITPCT_DEFAULT;
	module_data.meminfo_fd = fopen ("/proc/meminfo", "r");
	if (module_data.meminfo_fd == NULL) {
		eventdata_free (&module_data.event);
		free (module_data.cpusampleavg);
		return NETINFO_RC_MODULEERROR;
	}
	module_data.stat_fd = fopen ("/proc/stat", "r");
	if (module_data.stat_fd == NULL) {
		fclose (module_data.meminfo_fd);
		eventdata_free (&module_data.event);
		free (module_data.cpusampleavg);
		return NETINFO_RC_MODULEERROR;
	}
	module_data.pollinterval = POLL_INTERVAL_DEFAULT;
	rc = interfaces_initialize ();
	rc = disks_scan ();
	rc = passive_config_parse (SYSCONFDIR PASSIVE_CONFIG_FILE);
	rc = pthread_mutex_init (&module_data.loopmutex, NULL);
	if (rc != 0) {
		vector_log_entry (mi,
				"Failed to initialize loopmutex rc=%d\n",
				rc);
		eventdata_free (&module_data.event);
		free (module_data.cpusampleavg);
		return NETINFO_RC_MODULEERROR;
	}
	rc = pthread_cond_init (&module_data.loopcondition, NULL);
	if (rc != 0) {
		vector_log_entry (mi,
				"Failed to initialize loopcondition rc=%d\n",
				rc);
		pthread_mutex_destroy (&module_data.loopmutex);
		eventdata_free (&module_data.event);
		free (module_data.cpusampleavg);
		return NETINFO_RC_MODULEERROR;
	}
	rc = pthread_create (&module_data.polltid,
				NULL,
				poll_thread,
				NULL);
	if (rc == 0) {
		vector_log_entry (mi,
				"Started poll thread (tid=%d) interval=%d\n",
				module_data.polltid,
				module_data.pollinterval);
	} else {
		vector_log_entry (mi,
				"Failed to start poll thread.\n");
		fclose (module_data.stat_fd);
		fclose (module_data.meminfo_fd);
		eventdata_free (&module_data.event);
		free (module_data.cpusampleavg);
		return NETINFO_RC_MODULEERROR;
	}
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_close (ModuleInfo *mi)
{
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	pthread_cond_signal (&module_data.loopcondition);
	if (module_data.meminfo_fd != NULL)
		fclose (module_data.meminfo_fd);
	if (module_data.stat_fd != NULL)
		fclose (module_data.stat_fd);
	if (module_data.cpusampleavg != NULL)
		free (module_data.cpusampleavg);
	if (module_data.event != NULL)
		eventdata_free (&module_data.event);
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_execute_method (char *method,
					NetinfoHeader *nh,
					char **responsedata,
					unsigned long *responsedatalen)
{
	int rc;
	int len;
	int argdatalen;
	u_int64_t	lvalue;
	char argdata[NETINFO_HEADER_ARGDATA_SIZE];
	char buffer[BUFFER_SIZE];
	char *rd = NULL;
	if (method == NULL ||
			nh == NULL ||
			responsedata == NULL ||
			responsedatalen == NULL)
		return NETINFO_RC_INVALIDARG;
	*responsedata = NULL;
	*responsedatalen = 0;
	argdatalen = NETINFO_HEADER_ARGDATA_SIZE;
	rc = netinfo_header_get_argdata (nh,
				argdata,
				&argdatalen);
	if (!strcmp (method, "memfree")) {
		len = BUFFER_SIZE;
		rc = file_find_line_with (module_data.meminfo_fd,
					"MemFree:",
					buffer,
					&len);
		if (len != 0) {
			rc = sscanf (buffer, "MemFree: %llu", &lvalue);
			if (rc != 1)
				lvalue = 0;
			len = sprintf (buffer, "%llu kB", (unsigned long long)lvalue);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
		}
	} else if (!strcmp (method, "memtotal")) {
		len = BUFFER_SIZE;
		rc = file_find_line_with (module_data.meminfo_fd,
					"MemTotal:",
					buffer,
					&len);
		if (len != 0) {
			rc = sscanf (buffer, "MemTotal: %llu", &lvalue);
			if (rc != 1)
				lvalue = 0;
			len = sprintf (buffer, "%llu kB", (unsigned long long)lvalue);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
		}
	} else if (!strcmp (method, "swapfree")) {
		len = BUFFER_SIZE;
		rc = file_find_line_with (module_data.meminfo_fd,
					"SwapFree:",
					buffer,
					&len);
		if (len != 0) {
			rc = sscanf (buffer, "SwapFree: %llu", &lvalue);
			if (rc != 1)
				lvalue = 0;
			len = sprintf (buffer, "%llu kB", (unsigned long long)lvalue);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
		}
	} else if (!strcmp (method, "swaptotal")) {
		len = BUFFER_SIZE;
		rc = file_find_line_with (module_data.meminfo_fd,
					"SwapTotal:",
					buffer,
					&len);
		if (len != 0) {
			rc = sscanf (buffer, "SwapTotal: %llu", &lvalue);
			if (rc != 1)
				lvalue = 0;
			len = sprintf (buffer, "%llu kB", (unsigned long long)lvalue);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
		}
	} else if (!strcmp (method, "runqueue")) {
		len = BUFFER_SIZE;
		rc = file_find_line_with (module_data.stat_fd,
					"procs_running",
					buffer,
					&len);
		if (len != 0) {
			rc = sscanf (buffer, "procs_running %llu", &lvalue);
			if (rc != 1)
				lvalue = 0;
			len = sprintf (buffer, "%llu", (unsigned long long)lvalue);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
		}
	} else if (!strcmp (method, "blockedqueue")) {
		len = BUFFER_SIZE;
		rc = file_find_line_with (module_data.stat_fd,
					"procs_blocked",
					buffer,
					&len);
		if (len != 0) {
			rc = sscanf (buffer, "procs_blocked %llu", &lvalue);
			if (rc != 1)
				lvalue = 0;
			len = sprintf (buffer, "%llu", (unsigned long long)lvalue);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
		}
	} else if (!strcmp (method, "cpuuser")) {
		len = sprintf (buffer, "%3.2f",
			module_data.cpusampleavg[module_data.curinterval].u);
		rd = calloc (1, len);
		memcpy (rd, buffer, len);
		*responsedata = rd;
		*responsedatalen = len;
		rc = netinfo_header_set_status (nh,
				NETINFO_STATUS_RESPONSE);
		rc = netinfo_header_set_flags (nh,
				NETINFO_FLAG_DATAPRESENT);
		rc = netinfo_header_set_rc (nh,
				NETINFO_RC_SUCCESS);
	} else if (!strcmp (method, "cpunice")) {
		len = sprintf (buffer, "%3.2f",
			module_data.cpusampleavg[module_data.curinterval].n);
		rd = calloc (1, len);
		memcpy (rd, buffer, len);
		*responsedata = rd;
		*responsedatalen = len;
		rc = netinfo_header_set_status (nh,
				NETINFO_STATUS_RESPONSE);
		rc = netinfo_header_set_flags (nh,
				NETINFO_FLAG_DATAPRESENT);
		rc = netinfo_header_set_rc (nh,
				NETINFO_RC_SUCCESS);
	} else if (!strcmp (method, "cpusystem")) {
		len = sprintf (buffer, "%3.2f",
			module_data.cpusampleavg[module_data.curinterval].s);
		rd = calloc (1, len);
		memcpy (rd, buffer, len);
		*responsedata = rd;
		*responsedatalen = len;
		rc = netinfo_header_set_status (nh,
				NETINFO_STATUS_RESPONSE);
		rc = netinfo_header_set_flags (nh,
				NETINFO_FLAG_DATAPRESENT);
		rc = netinfo_header_set_rc (nh,
				NETINFO_RC_SUCCESS);
	} else if (!strcmp (method, "cpuidle")) {
		len = sprintf (buffer, "%3.2f",
			module_data.cpusampleavg[module_data.curinterval].i);
		rd = calloc (1, len);
		memcpy (rd, buffer, len);
		*responsedata = rd;
		*responsedatalen = len;
		rc = netinfo_header_set_status (nh,
				NETINFO_STATUS_RESPONSE);
		rc = netinfo_header_set_flags (nh,
				NETINFO_FLAG_DATAPRESENT);
		rc = netinfo_header_set_rc (nh,
				NETINFO_RC_SUCCESS);
	} else if (!strcmp (method, "cpuwait")) {
		len = sprintf (buffer, "%3.2f",
			module_data.cpusampleavg[module_data.curinterval].w);
		rd = calloc (1, len);
		memcpy (rd, buffer, len);
		*responsedata = rd;
		*responsedatalen = len;
		rc = netinfo_header_set_status (nh,
				NETINFO_STATUS_RESPONSE);
		rc = netinfo_header_set_flags (nh,
				NETINFO_FLAG_DATAPRESENT);
		rc = netinfo_header_set_rc (nh,
				NETINFO_RC_SUCCESS);
	} else {
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	return NETINFO_RC_SUCCESS;
}


void *poll_thread (void *arg)
{
	int		rc;
	/* int		intvl; */
	struct timespec ts;
	struct timeval  tv;
	ModuleData	*md = &module_data;
	sleep (2); /*FIXME:  module is loaded before netinfod can initialize SSL context */
	rc = vector_eventchain_open (module_data.moduleinfo,
				&module_data.eventchain);

	// initial disk stats collection
	rc = diskstats_collect ();

	md->curinterval = 0;
	/* intvl = 0; */
	while (1) {
		md->curinterval++;
		/* intvl++;*/
/*		md->cpualertinterval++; */
		rc = interfaces_collect_stats ();
		rc = cpu_collect_statistics ();
		if (md->curinterval == md->avgintervals) {
			/* rc = eventdata_new (&module_data.event); */
/*			rc = system_counters_status (); */
			rc = interface_stats_report ();
			/* re-scan for new disks */
			rc = disks_scan ();
			rc = disks_check_status (module_data.disklist);
			rc = vm_collect_statistics();
			rc = diskstats_collect ();
			/* rc = eventdata_free (&module_data.event); */
			md->curinterval = 0;
			/* intvl = 0; */
		}
		/* sleep (POLL_INTERVAL_DEFAULT); */
		rc = gettimeofday (&tv, NULL);
		ts.tv_sec = tv.tv_sec + md->pollinterval;
		ts.tv_nsec = tv.tv_usec * 1000;
		pthread_mutex_lock (&md->loopmutex);
		rc = pthread_cond_timedwait (&md->loopcondition,
				&md->loopmutex,
				&ts);
		pthread_mutex_unlock (&md->loopmutex);
		if (rc != ETIMEDOUT)
			break;
	}
	rc = vector_eventchain_close (&module_data.eventchain);
	pthread_exit (NULL);
	return NULL;
}


u_int32_t sum_ticks (CpuStats *cs)
{
	if (cs == NULL)
		return 0;
	return (cs->u + cs->n + cs->s + cs->i +	cs->w + cs->x + cs->y);
}


double calc_percent (u_int64_t part, u_int64_t total)
{
	return ( ((double)part / (double)total) * 100.0);
}


int file_find_line_with (FILE *fd,
			char *searchstr,
			char *buffer,
			int *bufferlen)
{
	int rc;
	if (fd == NULL || searchstr == NULL || buffer == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = fseek (fd, 0L, SEEK_SET);
	while (!feof(fd)) {
		fgets (buffer, *bufferlen, fd);
		if (strstr (buffer, searchstr) != NULL) {
			*bufferlen = strlen (buffer);
			return NETINFO_RC_SUCCESS;
		}
	}
	memset (buffer, 0, *bufferlen);
	*bufferlen = 0;
	return NETINFO_RC_SUCCESS;
}


void counter_init (Counter *counter, u_int32_t type)
{
	if (counter == NULL)
		return;
	counter->flags |= type;
	counter->prevval = 0.0;
	counter->currval = 0.0;
	return;
}


void counter_clear (Counter *counter)
{
	if (counter == NULL)
		return;
	counter->prevval = 0.0;
	counter->currval = 0.0;
	return;
}


void counter_value_push (Counter *counter, double value)
{
	if (counter == NULL)
		return;
	counter->prevval = counter->currval;
	counter->prevtime = counter->currtime;
	counter->currval = value;
	counter->currtime = time(NULL);
	return;
}


double counter_value_get (Counter *counter)
{
	double value;
	if (counter == NULL)
		return 0.0;
	value = counter->currval - counter->prevval;
	if (counter->flags & COUNTER_TYPE_COUNTER) {
		if (value < 0.0)
			value = (DBL_MAX - counter->prevval) + counter->currval;
		return (value / (counter->currtime - counter->prevtime));
	}
	return value;
}


double counter_value_push_and_get (Counter *counter, double value)
{
	if (counter == NULL)
		return 0.0;
	counter->prevval = counter->currval;
	counter->prevtime = counter->currtime;
	counter->currval = value;
	counter->currtime = time(NULL);
	value = counter->currval - counter->prevval;
	if (counter->flags & COUNTER_TYPE_COUNTER) {
		return (value / (counter->currtime - counter->prevtime));
	}
	return value;
}


InterfaceStats *interfacestats_new (int samples)
{
	InterfaceStats *is = NULL;
	is = calloc (1, sizeof (InterfaceStats));
	if (is == NULL)
		return NULL;
	counter_init (&is->rxbytes, COUNTER_TYPE_COUNTER);
	counter_init (&is->rxpkts, COUNTER_TYPE_COUNTER);
	counter_init (&is->rxerrors, COUNTER_TYPE_COUNTER);
	counter_init (&is->rxdrops, COUNTER_TYPE_COUNTER);
	counter_init (&is->rxfifo, COUNTER_TYPE_COUNTER);
	counter_init (&is->rxframe, COUNTER_TYPE_COUNTER);
	counter_init (&is->rxcomp, COUNTER_TYPE_COUNTER);
	counter_init (&is->rxmcast, COUNTER_TYPE_COUNTER);
	counter_init (&is->txbytes, COUNTER_TYPE_COUNTER);
	counter_init (&is->txpkts, COUNTER_TYPE_COUNTER);
	counter_init (&is->txerrors, COUNTER_TYPE_COUNTER);
	counter_init (&is->txdrops, COUNTER_TYPE_COUNTER);
	counter_init (&is->txfifo, COUNTER_TYPE_COUNTER);
	counter_init (&is->txcolls, COUNTER_TYPE_COUNTER);
	counter_init (&is->txcarr, COUNTER_TYPE_COUNTER);
	counter_init (&is->txcomp, COUNTER_TYPE_COUNTER);
	is->samrxbytes = calloc (samples, sizeof (double));
	is->samrxpkts = calloc (samples, sizeof (double));
	is->samrxerrors = calloc (samples, sizeof (double));
	is->samrxdrops = calloc (samples, sizeof (double));
	is->samrxfifo = calloc (samples, sizeof (double));
	is->samrxframe = calloc (samples, sizeof (double));
	is->samrxcomp = calloc (samples, sizeof (double));
	is->samrxmcast = calloc (samples, sizeof (double));
	is->samtxbytes = calloc (samples, sizeof (double));
	is->samtxpkts = calloc (samples, sizeof (double));
	is->samtxerrors = calloc (samples, sizeof (double));
	is->samtxdrops = calloc (samples, sizeof (double));
	is->samtxfifo = calloc (samples, sizeof (double));
	is->samtxcolls = calloc (samples, sizeof (double));
	is->samtxcarr = calloc (samples, sizeof (double));
	is->samtxcomp = calloc (samples, sizeof (double));
	if (is->samrxbytes == NULL ||
	    is->samrxpkts == NULL ||
	    is->samrxerrors == NULL ||
	    is->samrxdrops == NULL ||
	    is->samrxfifo == NULL ||
	    is->samrxframe == NULL ||
	    is->samrxcomp == NULL ||
	    is->samrxmcast == NULL) {
		free (is);
		return NULL;
	}
	if (is->samtxbytes == NULL ||
	    is->samtxpkts == NULL ||
	    is->samtxerrors == NULL ||
	    is->samtxdrops == NULL ||
	    is->samtxfifo == NULL ||
	    is->samtxcolls == NULL ||
	    is->samtxcarr == NULL ||
	    is->samtxcomp == NULL) {
		free (is);
		return NULL;
	}
	return is;
}


int interfacestats_free (InterfaceStats *is)
{
	if (is == NULL)
		return 1;
	free (is->samrxbytes);
	free (is->samrxpkts);
	free (is->samrxerrors);
	free (is->samrxdrops);
	free (is->samrxfifo);
	free (is->samrxframe);
	free (is->samrxcomp);
	free (is->samrxmcast);
	free (is->samtxbytes);
	free (is->samtxpkts);
	free (is->samtxerrors);
	free (is->samtxdrops);
	free (is->samtxfifo);
	free (is->samtxcolls);
	free (is->samtxcarr);
	free (is->samtxcomp);
	free (is->devname);
	free (is);
	return 0;
}


int interfaces_initialize (void)
{
	int		rc;
	char		ifname[16];
	char		buffer[BUFFER_SIZE];
	char		*p = NULL;
	FILE		*fd = NULL;
	InterfaceStats	*is = NULL;
	module_data.interfacelist = NULL;
	fd = fopen ("/proc/net/dev", "r");
	if (fd == NULL) 
		return NETINFO_RC_MODULEERROR;
	while (!feof (fd)) {
		ifname[0] = 0;
		if (fgets (buffer, BUFFER_SIZE, fd) == NULL)
			break;
		if (strstr (buffer, ":") == NULL)
			continue;
		rc = sscanf (buffer, "%16s: %*s", ifname);
		if (rc == 1) {
			p = strstr (ifname, ":");
			if (p != NULL)
				*p = 0;

			/*
			 * AVG_INTERVALS + 1: There needs to be enough space for
			 * one extra sample, because of the loop setup. This needs
			 * to be fixed.
			*/
			is = interfacestats_new (DEFAULT_AVERAGE_INTERVALS + 1);

			if (is != NULL) {
				is->devname = strdup (ifname);
				slist_item_add (&module_data.interfacelist, is);
				vector_log_entry (module_data.moduleinfo,
					"Added interface '%s'\n",
					ifname);
			} else {
				vector_log_entry (module_data.moduleinfo,
					"Failed to alloc memory for '%s'\n",
					ifname);
			}
		}
	}
	fclose (fd);
	return NETINFO_RC_SUCCESS;
}
	


int interfaces_collect_stats (void)
{
	int		rc;
	time_t		timestamp;
	double		value;
	double		rxbytes;
	double		rxpkts;
	double		rxerrors;
	double		rxdrops;
	double		rxfifo;
	double		rxframe;
	double		rxcomp;
	double		rxmcast;
	double		txbytes;
	double		txpkts;
	double		txerrors;
	double		txdrops;
	double		txfifo;
	double		txcolls;
	double		txcarr;
	double		txcomp;
	char		ifname[16];
	char		buffer[BUFFER_SIZE];
	char		*p = NULL;
	FILE		*fd = NULL;
	InterfaceStats	*is = NULL;
	SList		*litem = NULL;
	SList		*rilist = NULL;

	timestamp = time(NULL);
	fd = fopen ("/proc/net/dev", "r");
	if (fd == NULL) 
		return NETINFO_RC_MODULEERROR;
	while (!feof (fd)) {
		if (fgets (buffer, BUFFER_SIZE, fd) == NULL)
			break;
		ifname[0] = 0;
		if (strstr (buffer, ":") == NULL)
			continue;
		rc = sscanf (buffer, "%16s: %*s", ifname);
		p = strstr (ifname, ":");
		if (p != NULL)
			*p = 0;
		if (rc == 1) {
			is = interfacelist_get_interface (
					module_data.interfacelist,
					ifname);
			if (is == NULL) {
				/*
				* AVG_INTERVALS + 1: There needs to be enough space for
				* one extra sample, because of the loop setup. This needs
				* to be fixed.
				*/
				is = interfacestats_new (DEFAULT_AVERAGE_INTERVALS + 1);
				if (is != NULL) {
					is->devname = strdup (ifname);
					slist_item_add (&module_data.interfacelist, is);
					vector_log_entry (module_data.moduleinfo,
						"Added interface '%s'\n",
						ifname);
				} else {
					vector_log_entry (module_data.moduleinfo,
						"Failed to alloc memory for '%s'\n",
						ifname);
				}
			}
			if (is != NULL) {
				p = strstr (buffer, ":");
				p++;
				rc = sscanf (p, 
					" %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
					&rxbytes,
					&rxpkts,
					&rxerrors,
					&rxdrops,
					&rxfifo,
					&rxframe,
					&rxcomp,
					&rxmcast,
					&txbytes,
					&txpkts,
					&txerrors,
					&txdrops,
					&txfifo,
					&txcolls,
					&txcarr,
					&txcomp);
				if (rc != 16)
					continue;
				value = counter_value_push_and_get (&is->rxbytes,
								rxbytes);
				is->samrxbytes[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->rxpkts,
								rxpkts);
				is->samrxpkts[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->rxerrors,
								rxerrors);
				is->samrxerrors[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->rxdrops,
								rxdrops);
				is->samrxdrops[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->rxfifo,
								rxfifo);
				is->samrxfifo[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->rxframe,
								rxframe);
				is->samrxframe[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->rxcomp,
								rxcomp);
				is->samrxcomp[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->rxmcast,
								rxmcast);
				is->samrxmcast[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->txbytes,
								txbytes);
				is->samtxbytes[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->txpkts,
								txpkts);
				is->samtxpkts[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->txerrors,
								txerrors);
				is->samtxerrors[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->txdrops,
								txdrops);
				is->samtxdrops[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->txfifo,
								txfifo);
				is->samtxfifo[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->txcolls,
								txcolls);
				is->samtxcolls[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->txcarr,
								txcarr);
				is->samtxcarr[module_data.curinterval] = value;
				value = counter_value_push_and_get (&is->txcomp,
								txcomp);
				is->samtxcomp[module_data.curinterval] = value;
				is->lastupdate = timestamp;
			}
		}
	}
	fclose (fd);

	/* build list of interfaces that no longer exist */
	litem = module_data.interfacelist;
	while (litem != NULL) {
		is = (InterfaceStats *)slist_get_data (litem);
		if (is != NULL) {
			if (is->lastupdate < timestamp) {
				rc = slist_item_add (&rilist, is);
			}
		}
		litem = slist_get_next (litem);
	}

	/* Remove interfaces that no longer exist */
	if (rilist != NULL) {
		litem = rilist;
		while (litem != NULL) {
			is = (InterfaceStats *)slist_get_data (litem);
			rc = slist_item_remove (&module_data.interfacelist, is);
			vector_log_entry (module_data.moduleinfo,
					"Removed interface '%s'\n",
					is->devname);
			rc = interfacestats_free (is);
			litem = slist_get_next (litem);
		}
		rc = slist_free (&rilist);
	}

	return NETINFO_RC_SUCCESS;
}


InterfaceStats	*interfacelist_get_interface (SList *list, char *ifname)
{
	SList		*litem = NULL;
	InterfaceStats	*is = NULL;
	if (list == NULL || ifname == NULL)
		return NULL;
	litem = module_data.interfacelist;
	while (litem != NULL) {
		is = slist_get_data (litem);
		if (!strcmp (is->devname, ifname))
			return is;
		litem = slist_get_next (litem);
	}
	return NULL;
}


int interface_stats_report (void)
{
	int		rc;
	int		len;
	double		rxbytes;
	double		rxpkts;
	double		rxerrors;
	double		rxdrops;
	double		rxfifo;
	double		rxframe;
	double		rxcomp;
	double		rxmcast;
	double		txbytes;
	double		txpkts;
	double		txerrors;
	double		txdrops;
	double		txfifo;
	double		txcolls;
	double		txcarr;
	double		txcomp;
	char            buffer[BUFFER_SIZE];
	SList		*litem = NULL;
	InterfaceStats	*is = NULL;

	litem = module_data.interfacelist;
	while (litem != NULL) {
		is = slist_get_data (litem);
		if (is == NULL) {
			litem = slist_get_next (litem);
			continue;
		}
		rxbytes = samples_average (is->samrxbytes,
					module_data.avgintervals);
		rxpkts = samples_average (is->samrxpkts,
					module_data.avgintervals);
		rxerrors = samples_average (is->samrxerrors,
					module_data.avgintervals);
		rxdrops = samples_average (is->samrxdrops,
					module_data.avgintervals);
		rxfifo = samples_average (is->samrxfifo,
					module_data.avgintervals);
		rxframe = samples_average (is->samrxframe,
					module_data.avgintervals);
		rxcomp = samples_average (is->samrxcomp,
					module_data.avgintervals);
		rxmcast = samples_average (is->samrxmcast,
					module_data.avgintervals);
		txbytes = samples_average (is->samtxbytes,
					module_data.avgintervals);
		txpkts = samples_average (is->samtxpkts,
					module_data.avgintervals);
		txerrors = samples_average (is->samtxerrors,
					module_data.avgintervals);
		txdrops = samples_average (is->samtxdrops,
					module_data.avgintervals);
		txfifo = samples_average (is->samtxfifo,
					module_data.avgintervals);
		txcolls = samples_average (is->samtxcolls,
					module_data.avgintervals);
		txcarr = samples_average (is->samtxcarr,
					module_data.avgintervals);
		txcomp = samples_average (is->samtxcomp,
					module_data.avgintervals);
		memset (buffer, 0, BUFFER_SIZE);
		len = snprintf (buffer,
				BUFFER_SIZE,
				"%s: %f rxbytes/sec %f txbytes/sec|interface=%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f;%3.3f\n",
				is->devname,
				rxbytes,
				txbytes,
				rxbytes,
				rxpkts,
				rxerrors,
				rxdrops,
				rxfifo,
				rxframe,
				rxcomp,
				rxmcast,
				txbytes,
				txpkts,
				txerrors,
				txdrops,
				txfifo,
				txcolls,
				txcarr,
				txcomp);
		eventdata_clear (module_data.event);
		rc = eventdata_set (module_data.event,
				0,
				0,
				is->devname,
				buffer,
				len);

		rc = vector_eventchain_emit_event (module_data.eventchain,
						module_data.event);
		litem = slist_get_next (litem);
	}
	return NETINFO_RC_SUCCESS;
}

double samples_average (double *samplelist, int samples)
{
	int i;
	double value;
	if (samplelist == NULL || samples == 0.0)
		return 0.0;
	value = 0.0;
	for (i=0 ; i<samples ; i++)
		value += samplelist[i];
	return (value / (double)samples);
}


int passive_config_parse (char *filepath)
{
	int		rc;
	char		buffer[BUFFER_SIZE];
	char		*directive = NULL;
	char		*attribs = NULL;
	char		*attrib = NULL;
	ConfigFile	*cf = NULL;
	SList		*litem = NULL;
	DiskData	*dd = NULL;

	if (filepath == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = configfile_open (&cf, filepath);
	if (rc != NETINFO_RC_SUCCESS)
		return rc;

	rc = configfile_search_begin (cf, "WatchDisk");
	while (1) {
		rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		directive = strtok (buffer, "'\" \t");
		if (directive[0] == '#') {
			continue;
		}

		directive = strtok (NULL, "'\"\t");
		litem = module_data.disklist;
		while (litem != NULL) {
			dd = slist_get_data (litem);
			if (!strcmp (dd->path, directive)) {
				vector_log_entry (module_data.moduleinfo,
					"Altering defaults for %s\n",
					dd->path);
				attribs = strtok (NULL, " \t\r\n");
				attrib = strtok (attribs, "= \t\r\n");
				while (attrib != NULL) {
					if (!strcmp (attrib, "objectname")) {
						if (dd->objectname)
							free (dd->objectname);
						dd->objectname = 
							strdup (strtok (NULL,
								", \t\r\n"));
						vector_log_entry (module_data.moduleinfo,
							"%s: objectname=%s\n",
							dd->path,
							dd->objectname);
					} else if (!strcmp (attrib, "warnmax")) {
						dd->warnmax = 
							atof (strtok (NULL,
								", \t\r\n"));
						vector_log_entry (module_data.moduleinfo,
							"%s: warnmax=%f\n",
							dd->path,
							dd->warnmax);
					} else if (!strcmp (attrib, "critmax")) {
						dd->critmax = 
							atof (strtok (NULL,
								", \t\r\n"));
						vector_log_entry (module_data.moduleinfo,
							"%s: critmax=%f\n",
							dd->path,
							dd->critmax);
					}
				    attrib = strtok (NULL, "= \t\r\n");
				}
				vector_log_entry (module_data.moduleinfo, "\n");
			}
			litem = slist_get_next (litem);
		}
	}
	rc = configfile_search_end (cf);

	rc = configfile_search_begin (cf, "WatchCounter");
	while (1) {
		rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		directive = strtok (buffer, "'\" \t");
		if (directive[0] == '#') {
			continue;
		}
		directive = strtok (NULL, "'\"\t");
		if (!strcmp (directive, "SystemCpu")) {
			vector_log_entry (module_data.moduleinfo,
				"Altering defaults for SystemCpu:\n");
			attribs = strtok (NULL, " \t\r\n");
			attrib = strtok (attribs, "= \t\r\n");
			while (attrib != NULL) {
				if (!strcmp (attrib, "maxuser")) {
					module_data.cpudata.maxuser = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   maxuser=%3.2f\n",
						module_data.cpudata.maxuser);
				} else if (!strcmp (attrib, "maxsystem")) {
					module_data.cpudata.maxsystem = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   maxsystem=%3.2f\n",
						module_data.cpudata.maxsystem);
				} else if (!strcmp (attrib, "maxwait")) {
					module_data.cpudata.maxwait = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   maxwait=%3.2f\n",
						module_data.cpudata.maxwait);
				} else if (!strcmp (attrib, "maxtotal")) {
					module_data.cpudata.maxtotal = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   maxtotal=%3.2f\n",
						module_data.cpudata.maxtotal);
				} else if (!strcmp (attrib, "alertavgintervals")) {
					module_data.cpudata.alertavgintervals = 
						atoi (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"  alertavgintervals=%d\n",
						module_data.cpudata.alertavgintervals);
				}
				attrib = strtok (NULL, "= \t\r\n");
			}
		} else if (!strcmp (directive, "SystemVM")) {
			vector_log_entry (module_data.moduleinfo,
				"Altering defaults for SystemVM:\n");
			attribs = strtok (NULL, " \t\r\n");
			attrib = strtok (attribs, "= \t\r\n");
			while (attrib != NULL) {
				if (!strcmp (attrib, "memfreepctwarn")) {
					module_data.vmdata.memfreeminpctwarn = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   memfreepctwarn=%0.2f\n",
						module_data.vmdata.memfreeminpctwarn);
				} else if (!strcmp (attrib, "memfreepctcrit")) {
					module_data.vmdata.memfreeminpctcrit = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   memfreepctcrit=%0.2f\n",
						module_data.vmdata.memfreeminpctcrit);
				}
				if (!strcmp (attrib, "swapinusepctwarn")) {
					module_data.vmdata.swapinusemaxpctwarn = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   swapinusepctwarn=%0.2f\n",
						module_data.vmdata.swapinusemaxpctwarn);
				} else if (!strcmp (attrib, "swapinusepctcrit")) {
					module_data.vmdata.swapinusemaxpctcrit = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   swapinusepctcrit=%0.2f\n",
						module_data.vmdata.swapinusemaxpctcrit);
				}
				attrib = strtok (NULL, "= \t\r\n");
			}
		} else if (!strcmp (directive, "SystemLoad")) {
			vector_log_entry (module_data.moduleinfo,
				"Altering defaults for SystemLoad:\n");
			attribs = strtok (NULL, " \t\r\n");
			attrib = strtok (attribs, "= \t\r\n");
			while (attrib != NULL) {
				if (!strcmp (attrib, "warningload")) {
					module_data.loaddata.warningload = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   warningload=%0.2f\n",
						module_data.loaddata.warningload);
				} else if (!strcmp (attrib, "criticalload")) {
					module_data.loaddata.criticalload = 
						atof (strtok (NULL,
							", \t\r\n"));
					vector_log_entry (module_data.moduleinfo,
						"   criticalload=%0.2f\n",
						module_data.loaddata.criticalload);
				}
				attrib = strtok (NULL, "= \t\r\n");
			}
		}
	}
	rc = configfile_search_end (cf);

	rc = configfile_close (&cf);
	return NETINFO_RC_SUCCESS;
}


int disks_scan(void)
{
	int		rc;
/*
	u_int64_t	stotal;
	u_int64_t	sfree;
	u_int64_t	ftotal;
	u_int64_t	ffree;
*/
	struct		statvfs st;
	char		buffer[BUFFER_SIZE];
	char		device[BUFFER_SIZE];
	char		mpoint[BUFFER_SIZE];
	char		fstype[BUFFER_SIZE];
	FILE		*fp = NULL;
	DiskData	*dd = NULL;
	SList		*litem = NULL;

	fp = fopen ("/proc/mounts", "r");
	if (fp == NULL)
		return 1;
	while (!feof (fp)) {
		if (fgets (buffer, BUFFER_SIZE, fp) == NULL)
			break;
		rc = sscanf (buffer,
				"%s %s %s %*s %*s %*s",
				device,
				mpoint,
				fstype);
		if (rc != 3)
			continue;
		if (!strcmp (fstype, "rootfs") ||
		    !strcmp (fstype, "nfs") ||
		    !strcmp (fstype, "tmpfs") ||
		    !strcmp (fstype, "ramfs") ||
		    !strcmp (device, "none"))
			continue;
		rc = statvfs (mpoint, &st);
		if (rc != 0)
			continue;
/*
		stotal = st.f_frsize * (u_int64_t)st.f_blocks;
		sfree = st.f_frsize * (u_int64_t)st.f_bavail;
		ftotal = st.f_files;
		ffree = st.f_favail;
*/
		if (st.f_blocks > 0) {
			litem = module_data.disklist;
			while (litem != NULL) {
				dd = (DiskData*)slist_get_data (litem);
				if (!strcmp (mpoint, dd->path))
					break;
				litem = slist_get_next (litem);
			}
			if (litem == NULL) {
				dd = calloc (1, sizeof (DiskData));
				dd->path = strdup (mpoint);
				dd->objectname = strdup (DEFAULT_DISK_OBJECTNAME);
				dd->warnmax = DEFAULT_DISK_WARNMAX_PERCENT;
				dd->critmax = DEFAULT_DISK_CRITMAX_PERCENT;
				rc = slist_item_add (&module_data.disklist, dd);
				vector_log_entry (module_data.moduleinfo,
					"Added filesystem '%s'\n",
					mpoint);
			}
		}
	}
	fclose (fp);
	return NETINFO_RC_SUCCESS;
}


int disks_check_status (SList *disklist)
{
	int			rc;
	int			len;
	u_int64_t		stotal;
	u_int64_t		sfree;
	u_int64_t		sused;
	struct			statvfs st;
	double			percentused;
	u_int32_t		eventrc;
	char			buffer[BUFFER_SIZE];
	char			objname[BUFFER_SIZE];
	DiskData		*dd = NULL;
	SList			*litem = NULL;

	if (disklist == NULL)
		return NETINFO_RC_INVALIDARG;
	litem = disklist;
	while (litem != NULL) {
		dd = (DiskData*)slist_get_data (litem);
		rc = statvfs (dd->path, &st);
		if (rc != 0)
			continue;
		stotal = st.f_frsize * (u_int64_t)st.f_blocks;
		sfree = st.f_frsize * (u_int64_t)st.f_bavail;
		sused = stotal - sfree;
		percentused = ((double)(sused) / stotal) * 100.0;
		stotal /= 1024;
		sfree /= 1024;
		sused /= 1024;
		eventrc = 0;
		if (percentused >= dd->warnmax)
			eventrc = 1;
		if (percentused >= dd->critmax)
			eventrc = 2;
		snprintf (objname,
			BUFFER_SIZE,
			"%s:%s",
			dd->path,
			dd->objectname);
		len = snprintf (buffer,
				BUFFER_SIZE,
				"Filesystem %s is at %3.2f%%|filesystem=%3.2f;%3.2f;%3.2f;%llu;%llu",
				dd->path,
				percentused,
				percentused,
				dd->warnmax,
				dd->critmax,
				(unsigned long long)sused,
				(unsigned long long)stotal);
		eventdata_clear (module_data.event);
		eventdata_set (module_data.event,
				0,
				eventrc,
				objname,
				buffer,
				len);
		rc = vector_eventchain_emit_event (module_data.eventchain,
						module_data.event);
		if (eventrc != dd->lastrc) {
			eventdata_clear (module_data.event);
			eventdata_set (module_data.event,
					EVENT_FLAG_ONESHOT,
					eventrc,
					objname,
					buffer,
					len);
			rc = vector_eventchain_emit_event (module_data.eventchain,
							module_data.event);
		}
		dd->lastrc = eventrc;
		litem = slist_get_next (litem);
	}
	return NETINFO_RC_SUCCESS;
}



int vm_collect_statistics (void)
{
	int		rc;
	int		len;
	int		eventrc;
	double		mempct;
	double		swappct;
	char		buffer[BUFFER_SIZE];
	char		vmmsg[BUFFER_SIZE];
	VMStats		vms; 
	FILE		*fd = NULL;
	SystemVMData	*vmd = &module_data.vmdata;

	memset (&vms, 0, sizeof (VMStats));
	fd = fopen ("/proc/meminfo", "r");
	if (fd == NULL)
		return errno;

	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "MemTotal:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"MemTotal: %llu",
				(unsigned long long *)&vms.memtotal);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "MemFree:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"MemFree: %llu",
				(unsigned long long *)&vms.memfree);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "Buffers:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"Buffers: %llu",
				(unsigned long long *)&vms.membuffers);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "Cached:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"Cached: %llu",
				(unsigned long long *)&vms.memcached);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "SwapTotal:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"SwapTotal: %llu",
				(unsigned long long *)&vms.swaptotal);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "SwapFree:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"SwapFree: %llu",
				(unsigned long long *)&vms.swapfree);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "Active:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"Active: %llu",
				(unsigned long long *)&vms.memactive);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "HighTotal:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"HighTotal: %llu",
				(unsigned long long *)&vms.hightotal);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "HighFree:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"HighFree: %llu",
				(unsigned long long *)&vms.highfree);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "LowTotal:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"LowTotal: %llu",
				(unsigned long long *)&vms.lowtotal);
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (fd, "LowFree:", buffer, &len);
	if (len != 0) {
		rc = sscanf (buffer,
				"LowFree: %llu",
				(unsigned long long *)&vms.lowfree);
	}
	rc = fclose(fd);
	vms.meminuse = vms.memtotal - vms.memfree;
	vms.swapinuse = vms.swaptotal - vms.swapfree;
	vms.highinuse = vms.hightotal - vms.highfree;
	vms.lowinuse = vms.lowtotal - vms.lowfree;
	vms.bpinuse = vms.bptotal - vms.bpfree;
	vms.memtotal *= 1024;
	vms.memfree *= 1024;
	vms.meminuse *= 1024;
	vms.membuffers *= 1024;
	vms.memcached *= 1024;
	vms.swaptotal *= 1024;
	vms.swapfree *= 1024;
	vms.swapinuse *= 1024;
	vms.memactive *= 1024;
	vms.meminactive *= 1024;
	vms.hightotal *= 1024;
	vms.highfree *= 1024;
	vms.highinuse *= 1024;
	vms.lowtotal *= 1024;
	vms.lowfree *= 1024;
	vms.lowinuse *= 1024;
	vms.bptotal *= 1024;
	vms.bpfree *= 1024;
	vms.bpinuse *= 1024;

	eventrc = 0;
	mempct = (double)vms.memfree / (double)vms.memtotal * 100.0;
	swappct = (double)vms.swapinuse / (double)vms.swaptotal * 100.0;
	if (vmd->memfreeminpctwarn != 0.0 && mempct <= vmd->memfreeminpctwarn) {
		eventrc = RC_WARNING;
		len = snprintf (vmmsg,
				BUFFER_SIZE,
				"Free memory is at %0.2f%%",
				mempct);
	}
	if (vmd->swapinusemaxpctwarn != 0.0 && swappct >= vmd->swapinusemaxpctwarn) {
		eventrc = RC_WARNING;
		len = snprintf (vmmsg,
				BUFFER_SIZE,
				"Swap usage is at %0.2f%%",
				swappct);
	}
	if (vmd->memfreeminpctcrit != 0.0 && mempct <= vmd->memfreeminpctcrit) {
		eventrc = RC_CRITICAL;
		len = snprintf (vmmsg,
				BUFFER_SIZE,
				"Free memory is at %0.2f%%",
				mempct);
	}
	if (vmd->swapinusemaxpctcrit != 0.0 && swappct >= vmd->swapinusemaxpctcrit) {
		eventrc = RC_CRITICAL;
		len = snprintf (vmmsg,
				BUFFER_SIZE,
				"Swap usage is at %0.2f%%",
				swappct);
	}
	if (eventrc == 0) {
		len = snprintf (vmmsg,
				BUFFER_SIZE,
				"No issues to report.");
	}

	len = snprintf (buffer,
			BUFFER_SIZE,
			"SYSVM: %s|sysvm=%llu;%llu;%llu;%llu;%llu;%llu;%llu;%llu;%llu;%llu;%llu",
			vmmsg,
			(unsigned long long)vms.memtotal,
			(unsigned long long)vms.memfree,
			(unsigned long long)vms.membuffers,
			(unsigned long long)vms.memcached,
			(unsigned long long)vms.swaptotal,
			(unsigned long long)vms.swapfree,
			(unsigned long long)vms.memactive,
			(unsigned long long)vms.lowtotal,
			(unsigned long long)vms.lowfree,
			(unsigned long long)vms.hightotal,
			(unsigned long long)vms.highfree);
	eventdata_clear (module_data.event);
	eventdata_set (module_data.event,
			0,
			eventrc,
			SYSTEM_VM_OBJECT,
			buffer,
			len);
	rc = vector_eventchain_emit_event (module_data.eventchain,
					module_data.event);
	if (eventrc != vmd->lastrc) {
		eventdata_clear (module_data.event);
		eventdata_set (module_data.event,
				EVENT_FLAG_ONESHOT,
				eventrc,
				SYSTEM_VM_OBJECT,
				buffer,
				len);
		rc = vector_eventchain_emit_event (module_data.eventchain,
						module_data.event);
	}
	vmd->lastrc = eventrc;
	return rc;
}


int load_initialize (void)
{
	int		rc;
	int		cpus;
	char		buffer[BUFFER_SIZE];
	FILE		*fd = NULL;
	fd = fopen ("/proc/cpuinfo", "r");
	if (fd == NULL)
		return errno;
	cpus = 0;
	while (!feof(fd)) {
		fgets (buffer, BUFFER_SIZE, fd);
		if (strncmp (buffer, "processor", 9) == 0)
			cpus++;
	}
	rc = fclose (fd);
	module_data.loaddata.cpucount = cpus;
	module_data.loaddata.warningload = (double)cpus * 0.75;
	module_data.loaddata.criticalload = (double)cpus * 1.25;
	return 0;
}


int load_collect_statistics (void)
{
	int		rc;
	int		len;
	int		eventrc;
	double		load1;
	double		load5;
	double		load15;
	char		buffer[BUFFER_SIZE];
	FILE		*fd = NULL;
	fd = fopen ("/proc/loadavg", "r");
	if (fd == NULL)
		return errno;
	fgets (buffer, BUFFER_SIZE, fd);
	fclose (fd);
	rc = sscanf (buffer,
			"%lf %lf %lf ",
			&load1,
			&load5,
			&load15);
	eventrc = 0;
	if (rc == 3) {
		if (load15 >= module_data.loaddata.warningload)
			eventrc = RC_WARNING;
		if (load15 >= module_data.loaddata.criticalload)
			eventrc = RC_CRITICAL;
		module_data.loaddata.priorload[0] = load1;
		module_data.loaddata.priorload[1] = load5;
		module_data.loaddata.priorload[2] = load15;
	}
	len = snprintf (buffer,
			BUFFER_SIZE,
			"System load averages: %0.2f, %0.2f, %0.2f|loadavg=;%0.2f;%0.2f;%0.2f;%0.2f;%0.2f",
			load1,
			load5,
			load15,
			load1,
			load5,
			load15,
			module_data.loaddata.warningload,
			module_data.loaddata.criticalload);
	eventdata_clear (module_data.event);
	eventdata_set (module_data.event,
			0,
			eventrc,
			SYSTEM_LOAD_OBJECT,
			buffer,
			len);
	rc = vector_eventchain_emit_event (module_data.eventchain,
					module_data.event);
	if (eventrc != module_data.loaddata.lastrc) {
		eventdata_clear (module_data.event);
		eventdata_set (module_data.event,
				EVENT_FLAG_ONESHOT,
				eventrc,
				SYSTEM_LOAD_OBJECT,
				buffer,
				len);
		rc = vector_eventchain_emit_event (module_data.eventchain,
						module_data.event);
	}
	module_data.loaddata.lastrc = eventrc;
	return rc;
}


int cpu_collect_statistics (void)
{
	int		cl;
	int		ol;
	int		rc;
	int		len;
	int		eventrc;
	u_int32_t	ttotal;
	double		u;
	double		n;
	double		s;
	double		i;
	double		w;
	double		x;
	double		y;
	double		total;
	char		buffer[BUFFER_SIZE];
	CpuStats	cpu;
	FILE		*pollfd = NULL;
	ModuleData	*md = &module_data;

	/* update interval counters and check to see if we met an interval */
	/* if not, return */
	md->cpu_perf_interval.current++;
	md->cpu_alert_interval.current++;
	if (md->cpu_perf_interval.current != md->cpu_perf_interval.required && 
	    md->cpu_alert_interval.current != md->cpu_alert_interval.required)
		return NETINFO_RC_SUCCESS;

	cl = md->cpudata.curstatsid;
	ol = md->cpudata.oldstatsid;
	pollfd = fopen ("/proc/stat", "r");
	if (pollfd == NULL) {
		vector_log_entry (md->moduleinfo,
				"Failed to open /proc/stat (rc=%d).\n", errno);
		return NETINFO_RC_INVALIDARG;
	}
	len = BUFFER_SIZE;
	rc = file_find_line_with (pollfd,
				"cpu ",
				buffer,
				&len);
	if (len == 0) {
		vector_log_entry (md->moduleinfo,
				"cpu line not found.\n");
		fclose (pollfd);
		return NETINFO_RC_INVALIDARG;
	}
	rc = sscanf (buffer,
			"cpu %Lu %Lu %Lu %Lu %Lu %Lu %Lu",
			(unsigned long long *)&md->cpudata.cpustats[cl].u,
			(unsigned long long *)&md->cpudata.cpustats[cl].n,
			(unsigned long long *)&md->cpudata.cpustats[cl].s,
			(unsigned long long *)&md->cpudata.cpustats[cl].i,
			(unsigned long long *)&md->cpudata.cpustats[cl].w,
			(unsigned long long *)&md->cpudata.cpustats[cl].x,
			(unsigned long long *)&md->cpudata.cpustats[cl].y);
	fclose (pollfd);
	/* get the 'ticks' during this interval */
	cpu.u = md->cpudata.cpustats[cl].u - md->cpudata.cpustats[ol].u;
	cpu.n = md->cpudata.cpustats[cl].n - md->cpudata.cpustats[ol].n;
	cpu.s = md->cpudata.cpustats[cl].s - md->cpudata.cpustats[ol].s;
	cpu.i = md->cpudata.cpustats[cl].i - md->cpudata.cpustats[ol].i;
	cpu.w = md->cpudata.cpustats[cl].w - md->cpudata.cpustats[ol].w;
	cpu.x = md->cpudata.cpustats[cl].x - md->cpudata.cpustats[ol].x;
	cpu.y = md->cpudata.cpustats[cl].y - md->cpudata.cpustats[ol].y;
	ttotal = sum_ticks (&cpu);
	/* generate interval percentages */
	u = calc_percent (cpu.u, ttotal);
	n = calc_percent (cpu.n, ttotal);
	s = calc_percent (cpu.s, ttotal);
	i = calc_percent (cpu.i, ttotal);
	w = calc_percent (cpu.w, ttotal);
	x = calc_percent (cpu.x, ttotal);
	y = calc_percent (cpu.y, ttotal);
	/* update alert averages */
	rc = average_push_value (md->cpudata.alertavgu, &u);
	rc = average_push_value (md->cpudata.alertavgn, &n);
	rc = average_push_value (md->cpudata.alertavgs, &s);
	rc = average_push_value (md->cpudata.alertavgi, &i);
	rc = average_push_value (md->cpudata.alertavgw, &w);
	rc = average_push_value (md->cpudata.alertavgx, &x);
	rc = average_push_value (md->cpudata.alertavgy, &y);
	md->cpudata.alertavgcurinterval++;
	if (cl == 0) {
		md->cpudata.curstatsid = 1;
		md->cpudata.oldstatsid = 0;
	} else {
		md->cpudata.curstatsid = 0;
		md->cpudata.oldstatsid = 1;
	}

        eventrc = RC_OKAY;
	/* if this is an alert interval, check/report state */
	if (md->cpu_alert_interval.current == md->cpu_alert_interval.required) {
		rc = average_get_average (md->cpudata.alertavgu, &u);
		rc = average_get_average (md->cpudata.alertavgn, &n);
		rc = average_get_average (md->cpudata.alertavgs, &s);
		rc = average_get_average (md->cpudata.alertavgi, &i);
		rc = average_get_average (md->cpudata.alertavgw, &w);
		rc = average_get_average (md->cpudata.alertavgx, &x);
		rc = average_get_average (md->cpudata.alertavgy, &y);
		total = u + n + s + w + x + y;
		if (md->cpudata.maxuser > 0.0 && u >= md->cpudata.maxuser)
			eventrc = RC_CRITICAL;
		if (md->cpudata.maxsystem > 0.0 && s >= md->cpudata.maxsystem)
			eventrc = RC_CRITICAL;
		if (md->cpudata.maxwait > 0.0 && w >= md->cpudata.maxwait)
			eventrc = RC_CRITICAL;
		if (md->cpudata.maxtotal > 0.0 && total >= md->cpudata.maxtotal)
			eventrc = RC_CRITICAL;
		md->cpudata.currentstate = eventrc;
		md->cpu_alert_interval.current = 0;
	}
	if (md->cpu_perf_interval.current == md->cpu_perf_interval.required) {
		md->cpu_perf_interval.current = 0;
		/* just a perf interval, use current state (don't flap) */
		eventrc = md->cpudata.currentstate;
	}
	len = snprintf (buffer,
		BUFFER_SIZE,
                "System CPU: user=%3.2f%% system=%3.2f%% iowait=%3.2f%% idle=%3.2f%% nice=%3.2f%%|syscpu=%3.2f;%3.2f;%3.2f;%3.2f;%3.2f",
		u,
		s,
		w,
		i,
		n,
		u,
		s,
		i,
		w,
		n);
	eventdata_clear (module_data.event);
	rc = eventdata_set (module_data.event,
			0,
			eventrc,
			SYSTEM_CPU_OBJECT,
			buffer,
			len);
	rc = vector_eventchain_emit_event (module_data.eventchain,
					module_data.event);
	if (eventrc != md->cpudata.lastrc) {
		eventdata_clear (module_data.event);
		rc = eventdata_set (module_data.event,
				0,
				eventrc,
				SYSTEM_CPU_OBJECT,
				buffer,
				len);
		rc = vector_eventchain_emit_event (module_data.eventchain,
						module_data.event);
	}
	md->cpudata.lastrc = eventrc;
	return NETINFO_RC_SUCCESS;
}


// comparison function for AVLTree
int diskstatinfo_compare (struct diskstatinfo_t *dsi1, struct diskstatinfo_t *dsi2)
{
	return strcmp(dsi1->name, dsi2->name);
}


// comparison function for AVLTree (devname)
int diskstatinfo_compare_name (char *devname, struct diskstatinfo_t *dsi2)
{
	return strcmp(devname, dsi2->name);
}


// Create statistics from a diskstatinfo_t structure
int diskstatinfo_calculate_stats (struct diskstatinfo_t *dsi, struct diskstat_t *ds)
{
	int	c;
	int	p;

	// which slot are the current and previous stats in
	if (dsi->current == 0) {
		c = 0;
		p = 1;
	} else {
		c = 1;
		p = 0;
	}

	// calculate statistics
	ds->reads = dsi->ds[c].reads - dsi->ds[p].reads;
	ds->readsmerged = dsi->ds[c].readsmerged - dsi->ds[p].readsmerged;
	ds->readsectors = dsi->ds[c].readsectors - dsi->ds[p].readsectors;
	ds->readms = dsi->ds[c].readms - dsi->ds[p].readms;
	ds->writes = dsi->ds[c].writes - dsi->ds[p].writes;
	ds->writesmerged = dsi->ds[c].writesmerged - dsi->ds[p].writesmerged;
	ds->writesectors = dsi->ds[c].writesectors - dsi->ds[p].writesectors;
	ds->writems = dsi->ds[c].writems - dsi->ds[p].writems;
	ds->iosoutstanding = dsi->ds[c].iosoutstanding - dsi->ds[p].iosoutstanding;
	ds->ioswaitms = dsi->ds[c].ioswaitms - dsi->ds[p].ioswaitms;
	ds->iocompletionms = dsi->ds[c].iocompletionms - dsi->ds[p].iocompletionms;

	return 0;
}


// Collect disk stattistics from the /proc filesystem
int diskstats_collect (void)
{
	int			c;
	int			rc;
	int			len;
	int			interval;
	uint32_t		major;
	uint32_t		minor;
	char			devname[DEVNAME_SIZE];
	char			buffer[BUFFER_SIZE];
	FILE			*fd;
	struct diskstat_t	ds;
	struct diskstatinfo_t	*dsi;

	interval = module_data.pollinterval * module_data.curinterval;

	fd = fopen ("/proc/diskstats", "r");
	if (fd == NULL) {
		rc = errno;
		return rc;
	}
	while (!feof (fd)) {
		rc = fscanf (fd, "%u %u %s %u %u %u %u %u %u %u %u %u %u %u",
			&major,
			&minor,
			devname,
			&ds.reads,
			&ds.readsmerged,
			&ds.readsectors,
			&ds.readms,
			&ds.writes,
			&ds.writesmerged,
			&ds.writesectors,
			&ds.writems,
			&ds.iosoutstanding,
			&ds.ioswaitms,
			&ds.iocompletionms);
		if (rc != 14)
			continue;

		// skip some devices
		if (strstr(devname, "ram"))
			continue;
		if (strstr(devname, "loop"))
			continue;

		// locate the device
		dsi = avltree_find_by_compare (module_data.dev_tree,
						diskstatinfo_compare_name,
						devname);
		
		// Add a newly discovered device
		if (dsi == NULL) {
			dsi = calloc (1, sizeof (struct diskstatinfo_t));
			dsi->major = major;
			dsi->minor = minor;
			strcpy (dsi->name, devname);
			dsi->current = 1;
			avltree_insert (module_data.dev_tree, dsi);
			slist_item_add (&module_data.dev_list, dsi);
			vector_log_entry (module_data.moduleinfo,
				"Added device '%s'\n",
				devname);
		}

		// populate the dsi.ds structure
		if (dsi->current == 0) {
			c = 1;
			dsi->current = 1;
		} else {
			c = 0;
			dsi->current = 0;
		}
		dsi->ds[c].reads = ds.reads;
		dsi->ds[c].readsmerged = ds.readsmerged;
		dsi->ds[c].readsectors = ds.readsectors;
		dsi->ds[c].readms = ds.readms;
		dsi->ds[c].writes = ds.writes;
		dsi->ds[c].writesmerged = ds.writesmerged;
		dsi->ds[c].writesectors = ds.writesectors;
		dsi->ds[c].writems = ds.writems;
		dsi->ds[c].iosoutstanding = ds.iosoutstanding;
		dsi->ds[c].ioswaitms = ds.ioswaitms;
		dsi->ds[c].iocompletionms = ds.iocompletionms;

		// calculate interval statistics and emit performance events
		diskstatinfo_calculate_stats (dsi, &ds);
		len = snprintf (buffer,
			BUFFER_SIZE,
	                "I/O stats: readsec/s=%0.2f writesec/s=%0.2f waitms=%0.2f pendingios=%u|diskstat=%0.2f;%0.2f;%0.2f;%0.2f;%0.2f;%0.2f;%0.2f;%0.2f;%u;%0.2f;%0.2f",
			(float)ds.readsectors / interval,
			(float)ds.writesectors / interval,
			(float)ds.ioswaitms / interval,
			ds.iosoutstanding,
			(float)ds.reads / interval,
			(float)ds.readsmerged / interval,
			(float)ds.readsectors / interval,
			(float)ds.readms / interval,
			(float)ds.writes / interval,
			(float)ds.writesmerged / interval,
			(float)ds.writesectors / interval,
			(float)ds.writems / interval,
			ds.iosoutstanding,
			(float)ds.ioswaitms / interval,
			(float)ds.iocompletionms / interval);
		eventdata_clear (module_data.event);
		rc = eventdata_set (module_data.event,
				0,
				0,
				dsi->name,
				buffer,
				len);
		rc = vector_eventchain_emit_event (module_data.eventchain,
						module_data.event);
	}
	rc = fclose (fd);
	return 0;
}



