/************************************************************************/
/* ni_logpcre.c - Netinfo Log monitoring (with PCRE) plugin		*/
/************************************************************************/
/*
 *  (c)2008 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/

#define _FILE_OFFSET_BITS       64
#define _LARGEFILE64_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#ifndef WIN32
#include <pthread.h>
#endif

#include <cutil/slist.h>

#include <netinfo/netinfo.h>
#include <netinfo/nmodule.h>
#include <netinfo/conffile.h>

#include <pcre.h>
#include <pcrelog.h>


#define	MODULE_NAME			"logpcre"
#define MODULE_VERSION                  "0.1.2"

#define DEFAULT_CONFIGFILE		"passive.conf"
#define DEFAULT_RULEFILE		"logpcre.rules"

#ifndef WIN32
#define DIR_DELIMITER			"/"
#else
#define DIR_DELIMITER			"\\"
#define sleep(x)    Sleep (1000 * x)
/*
#define pthread_t   DWORD
int pthread_create (pthread_t *thread,
			void *attr,
			void *(*start_routine)(void *),
			void *arg);
*/
#endif


#define BUFFER_SIZE			512
#define FBUFFER_SIZE			8192

#define POLL_INTERVAL_DEFAULT		60

#define CHECK_INTERVAL_DEFAULT		5

#define	LOG_FLAG_WARNMISSING		0x00000001
#define	LOG_FLAG_REPORTSIZE		0x00000010
#define	LOG_FLAG_REPORTMODIFY		0x00000020
#define	LOG_FLAG_FIRSTCHECK		0x10000000

#define REASON_NONE			0
#define REASON_SIZE			1
#define REASON_GROWTH			2
#define REASON_MTIME			3


typedef struct	_LogData {
	InputSource	*inputsrc;
	MatchGroup	*matchgroup;
	char		*rulefile;
	char		*prematch;
	char		*logpath;
	char		*objectname;
	char		*errormatch;
	char		*matchobjectname;
	u_int64_t	lastsize;
	u_int32_t	flags;
	u_int32_t	checkinterval;
	u_int32_t	stepinterval;
	u_int32_t	warnlag;	
	u_int32_t	critlag;
	u_int64_t	warnsize;
	u_int64_t	critsize;
	u_int64_t	warngrowth;
	u_int64_t	critgrowth;
} LogData;



typedef struct _ModuleData {
	ModuleInfo	*moduleinfo;
	FILE		*meminfo_fd;
	FILE		*stat_fd;
	u_int32_t	pollinterval;
	u_int32_t	curinterval;
	pthread_t	polltid;
	SList		*loglist;
	EventChain	*eventchain;
	EventData	*event;
	pthread_mutex_t	loopmutex;
	pthread_cond_t	loopcondition;
} ModuleData;


/* Global Variables */
ModuleData	module_data;
const char	*SizeString[] = {
	"B",
	"KB",
	"MB",
	"GB",
	"TB",
	"PB",
	"EB",
	NULL
};
#define SIZESTRING_ORDER_MAX	6

/* Function Prototypes */
void *poll_thread (void *arg);
int parse_configfile (char *filepath);
int parse_log_growth (LogData *ld);
int size_to_string (u_int64_t *sizep, const char **sizestr);
int size_to_scale (u_int64_t *sizep);
int match_callback (MatchRule *rule, MatchData *matchdata);


int netinfo_module_init (ModuleInfo *mi)
{
	int rc;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	memset (&module_data, 0, sizeof (ModuleData));
	module_data.moduleinfo = mi;
	module_data.moduleinfo->name = MODULE_NAME;
	module_data.pollinterval = POLL_INTERVAL_DEFAULT;
	module_data.curinterval = 0;
	rc = pthread_mutex_init (&module_data.loopmutex, NULL);
	if (rc != 0) {
		vector_log_entry (mi,
				"Failed to initialize loopmutex rc=%d\n",
				rc);
		eventdata_free (&module_data.event);
		return NETINFO_RC_MODULEERROR;
	}
	rc = pthread_cond_init (&module_data.loopcondition, NULL);
	if (rc != 0) {
		vector_log_entry (mi,
				"Failed to initialize loopcondition rc=%d\n",
				rc);
		pthread_mutex_destroy (&module_data.loopmutex);
		eventdata_free (&module_data.event);
		return NETINFO_RC_MODULEERROR;
	}
	rc = eventdata_new (&module_data.event);
	rc = vector_eventchain_open (module_data.moduleinfo,
					&module_data.eventchain);

	rc = parse_configfile (SYSCONFDIR DIR_DELIMITER DEFAULT_CONFIGFILE);
	rc = pthread_create (&module_data.polltid,
				NULL,
				poll_thread,
				NULL);
	if (rc == 0) {
		vector_log_entry (mi,
				"Started poll thread (tid=%d) interval=%d\n",
				module_data.polltid,
				module_data.pollinterval);
	} else {
		vector_log_entry (mi,
				"Failed to start poll thread.\n");
		pthread_mutex_destroy (&module_data.loopmutex);
		pthread_cond_destroy (&module_data.loopcondition);
		return NETINFO_RC_MODULEERROR;
	}
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_close (ModuleInfo *mi)
{
	int		rc;
	SList		*litem = NULL;
	LogData		*ld = NULL;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	pthread_cond_signal (&module_data.loopcondition);
	litem = module_data.loglist;
	while (litem != NULL) {
		ld = (LogData *)slist_get_data (litem);
		if (ld != NULL) {
			if (ld->logpath != NULL)
				free (ld->logpath);
			if (ld->objectname != NULL)
				free (ld->objectname);
			if (ld->matchobjectname != NULL)
				free (ld->matchobjectname);
			free (ld);
		}
		litem = slist_get_next (litem);
	}
	rc = slist_free (&module_data.loglist);
	rc = eventdata_free (&module_data.event);
	rc = vector_eventchain_close (&module_data.eventchain);
	pthread_mutex_destroy (&module_data.loopmutex);
	pthread_cond_destroy (&module_data.loopcondition);
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_execute_method (char *method,
					NetinfoHeader *nh,
					char **responsedata,
					unsigned long *responsedatalen)
{
	int rc;
	int len;
	int argdatalen;
	char argdata[NETINFO_HEADER_ARGDATA_SIZE];
	char buffer[BUFFER_SIZE];
	char *rd = NULL;
	if (method == NULL ||
			nh == NULL ||
			responsedata == NULL ||
			responsedatalen == NULL)
		return NETINFO_RC_INVALIDARG;
	*responsedata = NULL;
	*responsedatalen = 0;
	argdatalen = NETINFO_HEADER_ARGDATA_SIZE;
	rc = netinfo_header_get_argdata (nh,
				argdata,
				&argdatalen);
	if (!strcmp (method, "curinterval")) {
			len = sprintf (buffer,
					"%lu",
					(long)module_data.curinterval);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
	} else {
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	return NETINFO_RC_SUCCESS;
}


void *poll_thread (void *arg)
{
	int		rc;
	int		eventrc;
	int		len;
	int		rcreason;
	int		lscale;
	int		gscale;
	u_int32_t	growth;
	u_int64_t	logsize;
	u_int64_t	growthsize;
	char		str[BUFFER_SIZE];
	char		buffer[BUFFER_SIZE];
	time_t		t;
	struct stat	st;
	struct timespec	ts;
	struct timeval	tv;
	struct tm	*tms = NULL;
	SList		*litem = NULL;
	LogData		*ld = NULL;
/*	EventChain	*eventchain = NULL; */
	ModuleData	*md = &module_data;

/*	sleep (2); */ /*FIXME:  module is loaded before netinfod can initialize SSL context */
	
	md->curinterval = 0;
	while (1) {
		litem = md->loglist;
		while (litem != NULL) {
			ld = slist_get_data (litem);
			ld->stepinterval++;
/*
			vector_log_entry (module_data.moduleinfo,
				"Checking log %s\n", ld->logpath);
*/
			if (ld->stepinterval == ld->checkinterval) {
/*
				vector_log_entry (module_data.moduleinfo,
					"Processing log %s\n", ld->logpath);
*/
				eventrc = 0;
				rcreason = REASON_NONE;
				t = time (NULL);
				rc = stat (ld->logpath, &st);
				if (rc == 0) {
					if (st.st_size < ld->lastsize) {
						/* log rotated/truncated */
						ld->lastsize = 0;
					}
					growth = st.st_size - ld->lastsize;
					if (ld->warnsize != 0 && 
					    st.st_size >= ld->warnsize) {
						eventrc = 1;
						rcreason = REASON_SIZE;
					}
					if (ld->warnlag != 0 &&
					    (t - st.st_mtime) >= ld->warnlag) {
						eventrc = 1;
						rcreason = REASON_MTIME;
					}
					if (ld->warngrowth != 0 &&
					    growth >= ld->warngrowth) {
						eventrc = 1;
						rcreason = REASON_GROWTH;
					}
					if (ld->critsize != 0 &&
					    st.st_size >= ld->critsize) {
						eventrc = 2;
						rcreason = REASON_SIZE;
					}
					if (ld->critlag != 0 &&
					    (t - st.st_mtime) >= ld->critlag) {
						eventrc = 2;
						rcreason = REASON_MTIME;
					}
					if (ld->critgrowth != 0 &&
					    growth >= ld->critgrowth) {
						eventrc = 2;
						rcreason = REASON_GROWTH;
					}
					tms = localtime (&st.st_mtime);
					len = snprintf (str,
						BUFFER_SIZE,
						"%04d-%02d-%02d %02d:%02d:%02d",
						tms->tm_year+1900,
						tms->tm_mon+1,
						tms->tm_mday,
						tms->tm_hour,
						tms->tm_min,
						tms->tm_sec);
					logsize = (u_int64_t)st.st_size;
					growthsize = growth;
					lscale = size_to_scale (&logsize);
					gscale = size_to_scale (&growthsize);
					if (rcreason == REASON_SIZE) {
						len = snprintf (buffer,
							BUFFER_SIZE,
							"%s: size=%llu%s|fsize=%lu;%llu;%llu\n",
							ld->logpath,
							logsize,
							SizeString[lscale],
							st.st_size,
							ld->warnsize,
							ld->critsize);
					} else if (rcreason == REASON_MTIME) {
						len = snprintf (buffer,
							BUFFER_SIZE,
							"%s: mtime='%s'|fsize=%lu;%llu;%llu\n",
							ld->logpath,
							str,
							st.st_size,
							ld->warnsize,
							ld->critsize);
					} else if (rcreason == REASON_GROWTH) {
						len = snprintf (buffer,
							BUFFER_SIZE,
							"%s: growth=%llu%s|fsize=%lu;%llu;%llu\n",
							ld->logpath,
							growthsize,
							SizeString[gscale],
							st.st_size,
							ld->warnsize,
							ld->critsize);
					} else {
						len = snprintf (buffer,
							BUFFER_SIZE,
							"%s: Okay|fsize=%lu;%llu;%llu",
							ld->logpath,
							st.st_size,
							ld->warnsize,
							ld->critsize);
					}
					
					/* BEGIN: Log watch code */
					if (growth > 0) {
						if (ld->flags & LOG_FLAG_FIRSTCHECK)
							ld->flags -= ld->flags & LOG_FLAG_FIRSTCHECK;
						else
							rc = parse_log_growth (ld);
					}
					/* END: Log watch code */
					ld->lastsize = st.st_size;
				} else {
					ld->lastsize = 0;
					eventrc = 1;
					len = snprintf (buffer,
						BUFFER_SIZE,
						"%s: %s\n",
						ld->logpath,
						strerror (errno));
				}
				if (rcreason != REASON_NONE) {
					eventdata_clear (md->event);
					rc = eventdata_set (md->event,
								EVENT_FLAG_ONESHOT,
								eventrc,
								ld->objectname,
								buffer,
								len);
					rc = vector_emit_event (module_data.moduleinfo,
								md->event);
				}
				ld->lastsize = st.st_size;
				ld->stepinterval = 0;
			}
			litem = slist_get_next (litem);
		}
		md->curinterval++;
		/* sleep (POLL_INTERVAL_DEFAULT); */
/* old: use select for loop timing
		tv.tv_sec = md->pollinterval;
		tv.tv_usec = 0;
		rc = select (0, NULL, NULL, NULL, &tv);
*/
		rc = gettimeofday (&tv, NULL);
		ts.tv_sec = tv.tv_sec + md->pollinterval;
		ts.tv_nsec = tv.tv_usec * 1000;
		pthread_mutex_lock (&md->loopmutex);
		rc = pthread_cond_timedwait (&md->loopcondition,
						&md->loopmutex,
						&ts);
		pthread_mutex_unlock (&md->loopmutex);
		if (rc != ETIMEDOUT)
			break;
	}
	pthread_exit (NULL);
	return NULL;
}


int parse_configfile (char *filepath)
{
	int		rc;
	int		i;
	char		buffer[BUFFER_SIZE];
	char		*directive = NULL;
	char		*attribs = NULL;
	char		*attrib = NULL;
	ConfigFile	*cf = NULL;
	LogData		*ld = NULL;
	if (filepath == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = configfile_open (&cf, filepath);
	if (rc != NETINFO_RC_SUCCESS)
		return rc;
	rc = configfile_search_begin (cf, "WatchLogPcre");
	while (1) {
		rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		directive = strtok (buffer, "'\" \t");
		if (directive[0] == '#') {
			continue;
		}
		
		ld = calloc (1, sizeof (LogData));
		ld->logpath = strdup (strtok (NULL, "'\"\t"));
		ld->checkinterval = CHECK_INTERVAL_DEFAULT;
		ld->flags |= LOG_FLAG_FIRSTCHECK;
		attribs = strtok (NULL, " \t\r\n");
		attrib = strtok (attribs, "= \t\r\n");
		while (attrib != NULL) {
			if (!strcmp (attrib, "objectname"))
				ld->objectname = strdup (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "checkinterval"))
				ld->checkinterval = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "warnmodifylag"))
				ld->warnlag = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "critmodifylag"))
				ld->critlag = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "warnsize"))
				ld->warnsize = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "critsize"))
				ld->critsize = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "warngrowth"))
				ld->warngrowth = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "critgrowth"))
				ld->critgrowth = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "errormatch"))
				ld->errormatch =  strdup (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "matchobjectname"))
				ld->matchobjectname = strdup (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "warnmissing")) {
				i = atol (strtok (NULL, ", \t\r\n"));
				if (i != 0)
					ld->flags |= LOG_FLAG_WARNMISSING;
			} else if (!strcmp (attrib, "reportsize")) {
				i = atol (strtok (NULL, ", \t\r\n"));
				if (i != 0)
					ld->flags |= LOG_FLAG_REPORTSIZE;
			} else if (!strcmp (attrib, "reportmodify")) {
				i = atol (strtok (NULL, ", \t\r\n"));
				if (i != 0)
					ld->flags |= LOG_FLAG_REPORTMODIFY;
			} else if (!strcmp (attrib, "rulefile")) {
				if (ld->rulefile != NULL)
					free (ld->rulefile);
				ld->rulefile = strdup (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "prematch")) {
				ld->prematch = strdup (strtok (NULL,
							", \t\r\n"));
			}
			attrib = strtok (NULL, "= \t\r\n");
		}

		if (ld->rulefile == NULL)
			ld->rulefile = strdup (SYSCONFDIR DIR_DELIMITER DEFAULT_RULEFILE);
		rc = matchgroup_new (&ld->matchgroup,
					ld->objectname,
					ld->prematch);
		if (rc != 0) {
			vector_log_entry (module_data.moduleinfo,
				"matchgroup_new(%s); failed rc=%d\n",
				ld->objectname,
				rc);
			free (ld->logpath);
			free (ld);
			continue;
		}
		rc = matchgroup_set_action_callback (ld->matchgroup,
					match_callback, ld);
		rc = matchgroup_load_rule_file (ld->matchgroup,
					ld->rulefile,
					0);
		if (rc != 0) {
			vector_log_entry (module_data.moduleinfo,
				"matchgroup_load_rule_file(%s); failed rc=%d\n",
				ld->objectname,
				rc);
			free (ld->logpath);
			free (ld);
			continue;
		}
		rc = inputsource_new (&ld->inputsrc,
				ld->objectname,
				ld->logpath,
				FBUFFER_SIZE);
		if (rc != 0) {
			vector_log_entry (module_data.moduleinfo,
				"inputsource_new(%s); failed rc=%d\n",
				ld->objectname,
				rc);
			free (ld->logpath);
			free (ld);
			continue;
		}
		rc = inputsource_set_matchgroup (ld->inputsrc, ld->matchgroup);
		rc = inputsource_open (ld->inputsrc);
		if (rc != 0) {
			vector_log_entry (module_data.moduleinfo,
				"inputsource_open(%s); failed rc=%d\n",
				ld->objectname,
				rc);
			free (ld->logpath);
			free (ld);
			continue;
		}
		vector_log_entry (module_data.moduleinfo,
			"Watching log '%s' prematch='%s' rulefile='%s'\n",
			ld->logpath,
			ld->prematch,
			ld->rulefile);
		rc = slist_item_add (&module_data.loglist, ld);
	}
	rc = configfile_search_end (cf);
	rc = configfile_close (&cf);
	return NETINFO_RC_SUCCESS;
}


int parse_log_growth (LogData *ld)
{
	int		rc;
	rc = inputsource_process (ld->inputsrc);
	if (rc == EROTATED) {
	}
	return rc;
}


int size_to_scale (u_int64_t *sizep)
{
	int s;
	u_int64_t size;
	if (sizep == NULL)
		return 0;
	size = (u_int64_t)*sizep;
	s = 0;
	while (size > 1024) {
		size /= 1024;
		s++;
	}
	if (s < 0 || s > SIZESTRING_ORDER_MAX)
		s = 0;
	*sizep = size;
	return s;
}


int match_callback (MatchRule *rule, MatchData *matchdata)
{
	int		rc;
	int		erc;
	unsigned int	eflags;
	int		len;
	char		buffer[BUFFER_SIZE];
	char		*objectname = NULL;
	char		*value = NULL;
	LogData		*ld = NULL;
	if (rule == NULL || matchdata == NULL)
		return EINVAL;
	rc = matchrule_get_argument (rule, &ld);
	eventdata_clear (module_data.event);
	if (ld->matchobjectname != NULL)
		objectname = ld->matchobjectname;
	else
		objectname = ld->objectname;
	len = BUFFER_SIZE;
	rc = matchrule_construct_message (rule, matchdata, buffer, &len);
	value = matchrule_get_attribute_value (rule, "eventobjectname");
	if (value != NULL)
		objectname = value;
	erc = 0;
	value = matchrule_get_attribute_value (rule, "eventrc");
	if (value != NULL)
		erc = atoi (value);
	eflags = EVENT_FLAG_ONESHOT;
	value = matchrule_get_attribute_value (rule, "eventflags");
	if (value != NULL)
		eflags = atol (value);
	rc = eventdata_set (module_data.event,
				eflags,
				erc,
				objectname,
				buffer,
				len);
	rc = vector_emit_event (module_data.moduleinfo,
				module_data.event);
	return 0;
}


