/************************************************************************/
/* ni_netcomm.c - Netinfo Network Comm monitoring plugin			*/
/************************************************************************/
/*
 *  (c)2006 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/



#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include <cutil/slist.h>

#include <netinfo/netinfo_portable.h>
#include <netinfo/netinfo.h>
#include <netinfo/nmodule.h>
#include <netinfo/conffile.h>


#define	MODULE_NAME			"netcomm"
#define MODULE_VERSION                  "0.1.1"

#define DEFAULT_CONFIGFILE		"passive.conf"

#ifndef WIN32
#define INVALID_SOCKET			-1
#define GET_ERRNO			errno
#else
#define DIR_DELIMITER			"\\"
#define sleep(x)    			Sleep (1000 * x)
#define read(x,y,z)			_read(x,y,z)
#define close(s)			closesocket(s)
#define GET_ERRNO			WSAGetLastError()

#endif


#define RC_OKAY		0
#define RC_WARNING	1
#define RC_CRITICAL	2
#define RC_UNKNOWN	3


#define BUFFER_SIZE			256

#define POLL_INTERVAL_DEFAULT		60

#define CHECK_INTERVAL_DEFAULT		5

#define FLAG_READ_BANNER		1


typedef struct	_NetCommData {
	char		*checkhost;
	char		*objectname;
	u_int32_t	checkport;
	u_int32_t	flags;
	u_int32_t	checkinterval;
	u_int32_t	stepinterval;
	double		warntime;
	double		crittime;
	double		lasttime;
	u_int32_t	laststate;
} NetCommData;



typedef struct _ModuleData {
	ModuleInfo	*moduleinfo;
	u_int32_t	pollinterval;
	u_int32_t	curinterval;
	pthread_t	polltid;
	SList		*nclist;
} ModuleData;


/* Global Variables */
ModuleData	module_data;


/* Function Prototypes */
void *poll_thread (void *arg);
int parse_configfile (char *filepath);
int check_connection (NetCommData *nc,
		char *message,
		int messagelen,
		double *connecttime);


int netinfo_module_init (ModuleInfo *mi)
{
	int rc;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	memset (&module_data, 0, sizeof (ModuleData));
	module_data.moduleinfo = mi;
	module_data.moduleinfo->name = MODULE_NAME;
	module_data.pollinterval = POLL_INTERVAL_DEFAULT;
	module_data.curinterval = 0;

	rc = parse_configfile (SYSCONFDIR DIR_DELIMITER DEFAULT_CONFIGFILE);
	rc = pthread_create (&module_data.polltid,
				NULL,
				poll_thread,
				NULL);
	if (rc == 0) {
		vector_log_entry (mi,
				"Started poll thread (tid=%d) interval=%d\n",
				module_data.polltid,
				module_data.pollinterval);
	} else {
		vector_log_entry (mi,
				"Failed to start poll thread.\n");
		return NETINFO_RC_MODULEERROR;
	}
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_close (ModuleInfo *mi)
{
	int		rc;
	SList		*nitem = NULL;
	NetCommData		*nc = NULL;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	nitem = module_data.nclist;
	while (nitem != NULL) {
		nc = (NetCommData *)slist_get_data (nitem);
		if (nc != NULL) {
			if (nc->checkhost != NULL)
				free (nc->checkhost);
			if (nc->objectname != NULL)
				free (nc->objectname);
			free (nc);
		}
		nitem = slist_get_next (nitem);
	}
	rc = slist_free (&module_data.nclist);
	return NETINFO_RC_SUCCESS;
}

int netinfo_module_execute_method (char *method,
					NetinfoHeader *nh,
					char **responsedata,
					unsigned long *responsedatalen)
{
	int rc;
	int len;
	int argdatalen;
	char argdata[NETINFO_HEADER_ARGDATA_SIZE];
	char buffer[BUFFER_SIZE];
	char *rd = NULL;
	if (method == NULL ||
			nh == NULL ||
			responsedata == NULL ||
			responsedatalen == NULL)
		return NETINFO_RC_INVALIDARG;
	*responsedata = NULL;
	*responsedatalen = 0;
	argdatalen = NETINFO_HEADER_ARGDATA_SIZE;
	rc = netinfo_header_get_argdata (nh,
				argdata,
				&argdatalen);
	if (!strcmp (method, "curinterval")) {
			len = sprintf (buffer,
					"%lu",
					(long)module_data.curinterval);
			rd = calloc (1, len);
			memcpy (rd, buffer, len);
			*responsedata = rd;
			*responsedatalen = len;
			rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
			rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
			rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
	} else {
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	return NETINFO_RC_SUCCESS;
}


void *poll_thread (void *arg)
{
	int		rc;
	int		eventrc;
	int		len;
	double		connecttime;
	char		buffer[BUFFER_SIZE];
	char		banner[BUFFER_SIZE];
	SList		*nitem = NULL;
	NetCommData	*nc = NULL;
	EventData       *event = NULL;
	EventChain	*eventchain = NULL;
	ModuleData	*md = &module_data;

	sleep (2); /*FIXME:  module is loaded before netinfod can initialize SSL context */
	
        rc = eventdata_new (&event);
	eventdata_clear (event);
	md->curinterval = 0;
	while (1) {
		nitem = md->nclist;
		while (nitem != NULL) {
			nc = slist_get_data (nitem);
			nc->stepinterval++;
			if (nc->stepinterval == nc->checkinterval) {
				if (eventchain == NULL) {
					rc = vector_eventchain_open (module_data.moduleinfo,
								&eventchain);
				}
				eventrc = 0;
				rc = check_connection (nc,
							banner,
							BUFFER_SIZE,
							&connecttime);
				if (rc != 0) {
					eventrc = rc;
					len = snprintf (buffer,
						BUFFER_SIZE,
						"%s|latency=0.0;%2.3f;%2.3f\n",
						banner,
						nc->warntime,
						nc->crittime);
				} else {
					if (nc->warntime != 0 && 
						connecttime >= nc->warntime)
						eventrc = 1;
					if (nc->crittime != 0 && 
						connecttime >= nc->crittime)
						eventrc = 2;
					len = snprintf (buffer,
						BUFFER_SIZE,
						"%s|latency=%2.3f;%2.3f;%2.3f\n",
						banner,
						connecttime,
						nc->warntime,
						nc->crittime);
				}
				rc = eventdata_set (event,
							0,
							eventrc,
							nc->objectname,
							buffer,
							len);
				rc = vector_eventchain_emit_event (eventchain,
							event);
				if (eventrc != nc->laststate) {
					rc = eventdata_set (event,
								EVENT_FLAG_ONESHOT,
								eventrc,
								nc->objectname,
								buffer,
								len);
					rc = vector_eventchain_emit_event (eventchain,
								event);
				}
				nc->stepinterval = 0;
				nc->lasttime = connecttime;
				nc->laststate = eventrc;
			}
			nitem = slist_get_next (nitem);
		}
		if (eventchain != NULL) {
			rc = vector_eventchain_close (&eventchain);
		}
		md->curinterval++;
		sleep (POLL_INTERVAL_DEFAULT);
	}
	eventdata_free (&event);
	pthread_exit (NULL);
	return NULL;
}


int parse_configfile (char *filepath)
{
	int		rc;
	int		i;
	char		buffer[BUFFER_SIZE];
	char		*directive = NULL;
	char		*attribs = NULL;
	char		*attrib = NULL;
	ConfigFile	*cf = NULL;
	NetCommData	*nc = NULL;
	if (filepath == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = configfile_open (&cf, filepath);
	if (rc != NETINFO_RC_SUCCESS)
		return rc;
	rc = configfile_search_begin (cf, "CheckNetComm");
	while (1) {
		rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		directive = strtok (buffer, "'\" \t");
		if (directive[0] == '#') {
			continue;
		}
		
		nc = calloc (1, sizeof (NetCommData));
		nc->checkhost = strdup (strtok (NULL, "'\"\t"));
		nc->checkinterval = CHECK_INTERVAL_DEFAULT;
		nc->checkport = 21;
		attribs = strtok (NULL, " \t\r\n");
		attrib = strtok (attribs, "= \t\r\n");
		while (attrib != NULL) {
			if (!strcmp (attrib, "objectname"))
				nc->objectname = strdup (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "checkport"))
				nc->checkport = atoi (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "checkinterval"))
				nc->checkinterval = atol (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "warntime"))
				nc->warntime = atof (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "crittime"))
				nc->crittime = atof (strtok (NULL,
								", \t\r\n"));
			else if (!strcmp (attrib, "readbanner")) {
				i = atol (strtok (NULL, ", \t\r\n"));
				if (i != 0)
					nc->flags |= FLAG_READ_BANNER;
			}
			attrib = strtok (NULL, "= \t\r\n");
		}
		vector_log_entry (module_data.moduleinfo,
				"Checking host='%s' port=%d checkinterval=%d warntime=%2.3f crittime=%2.3f\n",
				nc->checkhost,
				nc->checkport,
				nc->checkinterval,
				nc->warntime,
				nc->crittime);
		rc = slist_item_add (&module_data.nclist, nc);
	}
	rc = configfile_search_end (cf);
	rc = configfile_close (&cf);
	return NETINFO_RC_SUCCESS;
}




int check_connection (NetCommData *nc,
		char *message,
		int messagelen,
		double *connecttime)
{
	int rc;
	int s;
	struct hostent *he;
	struct sockaddr_in sa;
	double sus;
	double mus;
#ifndef WIN32
	struct timeval  stime;
	struct timeval  mtime;
#else
	WSADATA wsaData;
#endif
	char buffer[BUFFER_SIZE];

	*connecttime = 0.0;
	if (nc == NULL || connecttime == NULL)
		return EINVAL;


#ifdef WIN32
	WSAStartup(MAKEWORD(2,0), &wsaData);
#endif

	/* Get the IP address for the hostname */
	he = gethostbyname (nc->checkhost);
	if (he == NULL || he->h_addr_list[0] == NULL) {
		snprintf (message,
			messagelen,
			"Error finding IP for '%s'. Check this.\n",
			nc->checkhost);
		return RC_UNKNOWN;
	}
	memcpy (&sa.sin_addr, he->h_addr_list[0], he->h_length);

	sa.sin_port = htons (nc->checkport);
	sa.sin_family = AF_INET;

	/* Create a socket */
#ifndef WIN32
	rc = gettimeofday (&stime, NULL);
#else
    sus = (double)GetTickCount() / 1000.0;
#endif
	s = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (s == INVALID_SOCKET) {
		rc = GET_ERRNO;
		snprintf (message,
			messagelen,
			"Connection to %s:%d socket() rc=%d",
			nc->checkhost,
			nc->checkport,
			rc);
#ifdef WIN32
		WSACleanup();
#endif
		return RC_UNKNOWN;
	}

	/* Connect to the remote host/port */
	rc = connect (s, (struct sockaddr *)&sa, sizeof(struct sockaddr_in));
	if (rc == -1) {
		rc = GET_ERRNO;
		snprintf (message,
			messagelen,
			"Connection to %s:%d connect() rc=%d",
			nc->checkhost,
			nc->checkport,
			rc);
#ifdef WIN32
		WSACleanup();
#endif
		return RC_CRITICAL;
	}

	/* Read the response */
	if (nc->flags & FLAG_READ_BANNER) {
		errno = 0;
		rc = recv(s, buffer, BUFFER_SIZE, 0);
		if (rc == -1) {
			rc = GET_ERRNO;
			snprintf (message,
				messagelen,
				"Connection to %s:%d read() rc=%d",
				nc->checkhost,
				nc->checkport,
				rc);
			rc = close (s);
#ifdef WIN32
			WSACleanup();
#endif
			return RC_CRITICAL;
		} else if (rc == 0) {
			rc = GET_ERRNO;
			snprintf (message,
				messagelen,
				"Connection to %s:%d read() returned 0 bytes",
				nc->checkhost,
				nc->checkport);
			rc = close (s);
#ifdef WIN32
			WSACleanup();
#endif
			return RC_WARNING;
		}
	}
#ifndef WIN32
	rc = gettimeofday (&mtime, NULL);
#else
    mus = (double)GetTickCount() / 1000.0;
#endif
	rc = close (s);
#ifndef WIN32
	sus = (double)(stime.tv_sec) + ((double)stime.tv_usec / 1000000);
	mus = (double)(mtime.tv_sec) + ((double)mtime.tv_usec / 1000000);
#endif
	*connecttime = mus - sus;
	snprintf (message,
		messagelen,
		"Connection to %s:%d took %2.6f seconds",
		nc->checkhost,
		nc->checkport,
		*connecttime);

#ifdef WIN32
	WSACleanup();
#endif
	return RC_OKAY;
}
