/************************************************************************/
/* ni_process.c - Netinfo process monitor plugin			*/
/************************************************************************/
/*
 *  (c)2005 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>

#include <cutil/slist.h>

#include <netinfo/netinfo_portable.h>
#include <netinfo/netinfo.h>
#include <netinfo/nmodule.h>
#include <netinfo/conffile.h>

#ifndef WIN32
#include <signal.h>
#include <pwd.h>
#else
#include <windows.h>
#include <lmcons.h>
#include <wtsapi32.h>
#include <psapi.h>

/* Some cross-gcc environments don't have a complete wtsapi32.h */
#ifndef WTS_PROCESS_INFO
typedef struct _WTS_PROCESS_INFO	WTS_PROCESS_INFO;
struct _WTS_PROCESS_INFO {
	DWORD SessionId;
	DWORD ProcessId;
	LPTSTR pProcessName;
	PSID pUserSid;
};
typedef WTS_PROCESS_INFO*	PWTS_PROCESS_INFO;
#define PWTS_PROCESS_INFOA		WTS_PROCESS_INFO
#define	WTS_CURRENT_SERVER_HANDLE	0
#endif

BOOL WINAPI WTSEnumerateProcessesA(HANDLE, DWORD, DWORD, PWTS_PROCESS_INFO *, DWORD *);
void WINAPI WTSFreeMemory(PVOID); 

#define WTSEnumerateProcesses		WTSEnumerateProcessesA
/*
#define	WTSFreeMemory			WTSFreeMemoryA
#define	GetProcessMemoryInfo		GetProcessMemoryInfoA
*/
#endif


#define MODULE_NAME                     "process"
#define	MODULE_VERSION			"0.2.0"

#ifndef WIN32
#define DEFAULT_CONFIGFILE              "/passive.conf"
#else
#define DEFAULT_CONFIGFILE              "\\passive.conf"
#endif


#define POLL_INTERVAL_DEFAULT_SECS	60
#ifndef WIN32
#define SINGLE_INTERVAL         	1
#else
#define SINGLE_INTERVAL         	1000
#endif

#define	BUFFER_SIZE			512


#define	PROCNAME_SIZE			32
#define	OWNERNAME_SIZE			32
#define FILEPATH_SIZE			256
#define STATDATA_SIZE			1024
#define	PROCARRAY_SIZE			2048


#define	RC_OKAY				0
#define	RC_WARNING			1
#define	RC_CRITICAL			2
#define	RC_UNKNOWN			3


#define SYNCPIPE_READ			0
#define SYNCPIPE_WRITE			1


#define RULE_FLAG_DEBUGINFO		1
#define RULE_FLAG_TERMINATE		2


typedef struct _ProcessStat	ProcessStat;
struct _ProcessStat {
	int		state;
	int		cpualertintervals;
/*
	int		cpuwarncount;
	int		cpucritcount;
*/
	int		res1;
	int		res2;
};


typedef struct _ProcessInfo ProcessInfo;
struct _ProcessInfo {
	ProcessStat	pi_ps;
	int		pi_pid;
	char		pi_procname[PROCNAME_SIZE];
	char		pi_ownername[OWNERNAME_SIZE];
	int		pi_uid;
	char		pi_state;
	int		pi_ppid;
	int		pi_pgrp;
	int		pi_session;
	int		pi_tty_nr;
	int		pi_tpgid;
	unsigned long	pi_flags;
	unsigned long	pi_minflt;
	unsigned long	pi_cminflt;
	unsigned long	pi_majflt;
	unsigned long	pi_cmajflt;
	unsigned long	pi_utime;
	unsigned long	pi_stime;
	unsigned long	pi_cutime;
	unsigned long	pi_cstime;
	unsigned long	pi_priority;
	unsigned long	pi_nice;
	unsigned long	pi_dummy1;
	unsigned long	pi_itrealvalue;
	unsigned long	pi_starttime;
	unsigned long	pi_vsize;
	unsigned long	pi_rss;
};


typedef struct _RuleStat	RuleStat;
struct _RuleStat {
	int		eventrc;
	unsigned long	proccount;
	SList		*msglist;
	unsigned long	res1;
	unsigned long	res2;
};

	
typedef struct _ProcessRule	ProcessRule;
struct _ProcessRule {
	RuleStat	rulestat;
	char		*objectname;
	SList		*matchlist;
	unsigned long flags;
	unsigned long	warnmemuse;
	unsigned long	critmemuse;
	unsigned long	cpuinterval;
	unsigned long	cpuwarnuse;
	unsigned long	cpucrituse;
	unsigned long	minprocs;
	unsigned long	maxprocs;
	unsigned long	res1;
	unsigned long	res2;
};


typedef struct _Trigger {
	void		*handle;
	int		pipe[2];
} Trigger;


typedef struct _ModuleData {
	ModuleInfo      *moduleinfo;
	u_int32_t	procarraysize;
	ProcessInfo	*procarray;
	u_int32_t	proccount;
	ProcessInfo	*procarrayold;
	u_int32_t	procoldcount;
	u_int32_t       pollintervalsecs;
	u_int32_t       pollinterval;
	u_int32_t       curinterval;
	u_int32_t       tickssec;
	pthread_t       polltid;
	SList           *rulelist;
/*
	int		syncpipe[2];
*/
	Trigger		*trigger;
} ModuleData;



/* Global Variables */
ModuleData      module_data;



/* Function Prototypes */
void *poll_thread (void *arg);
int parse_configfile (char *filepath);
int enumerate_processes (void **enumstat, ProcessInfo *proclist, int *listsize);
int processrule_add_message (ProcessRule *pr, int returncode, char *message, int messagelen);
int processrule_free_messages (ProcessRule *pr);
ProcessRule *rulelist_find_process_match (SList *rulelist, char *processname);
ProcessInfo *procinfolist_find_pid (ProcessInfo *proclist, int listsize, int pid);
int trigger_new (Trigger **tp);
int trigger_sleep (Trigger *t, int sleepseconds);
int trigger_activate (Trigger *t);
int trigger_free(Trigger **tp);
int terminate_process (int pid);


int netinfo_module_init (ModuleInfo *mi)
{
	int		rc;
	SList		*ritem = NULL;
	ProcessRule	*pr = NULL;
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
	memset (&module_data, 0, sizeof (ModuleData));
	module_data.moduleinfo = mi;
	module_data.moduleinfo->name = MODULE_NAME;
	module_data.pollintervalsecs = POLL_INTERVAL_DEFAULT_SECS;
	module_data.curinterval = 0;
	module_data.procarraysize = PROCARRAY_SIZE;
#ifndef WIN32
	module_data.tickssec = sysconf (_SC_CLK_TCK);
#else
	module_data.tickssec = 1;
#endif
	rc = parse_configfile (SYSCONFDIR DEFAULT_CONFIGFILE);
	if (rc != NETINFO_RC_SUCCESS) {
		vector_log_entry (mi,
			"Error parsing configuration data. rc=%d(%s)\n",
			rc,
			strerror (rc));
		return NETINFO_RC_MODULEERROR;
	}
	module_data.procarray = calloc (module_data.procarraysize, sizeof (ProcessInfo));
	if (module_data.procarray == NULL) {
		vector_log_entry (mi,
			"Failed to allocate process array memory.\n");
		return NETINFO_RC_MODULEERROR;
	}
	module_data.procarrayold = calloc (module_data.procarraysize, sizeof (ProcessInfo));
	if (module_data.procarray == NULL) {
		vector_log_entry (mi,
			"Failed to allocate process array memory.\n");
		return NETINFO_RC_MODULEERROR;
	}
/*
	rc = pipe (module_data.syncpipe);
*/
	rc = trigger_new (&module_data.trigger);
	if (rc != NETINFO_RC_SUCCESS) {
		vector_log_entry (mi,
			"Failed to allocate trigger.\n");
		return NETINFO_RC_MODULEERROR;
	}
	/* Set the initial eventrc to cause the first check to send events */
	ritem = module_data.rulelist;
	while (ritem != NULL) {
		pr = (ProcessRule *)slist_get_data (ritem);
		pr->rulestat.eventrc = -1;
		ritem = slist_get_next (ritem);
	}
	rc = pthread_create (&module_data.polltid,
			NULL,
			poll_thread,
			NULL);
	if (rc == 0) {
		vector_log_entry (mi,
			"Started poll thread (tid=%d) interval=%d\n",
			module_data.polltid,
			module_data.pollintervalsecs);
	} else {
		vector_log_entry (mi,
			"Failed to start poll thread.\n");
		return NETINFO_RC_MODULEERROR;
	}
	return NETINFO_RC_SUCCESS;
}


int netinfo_module_close (ModuleInfo *mi)
{
	int		rc;
	char		*str = NULL;
	SList		*ritem = NULL;
	SList		*mitem = NULL;
	ProcessRule	*pr = NULL;
	
	if (mi == NULL)
		return NETINFO_RC_INVALIDARG;
/*
	rc = write (module_data.syncpipe[SYNCPIPE_WRITE], "C", 1);
*/
	rc = trigger_activate (module_data.trigger);
	ritem = module_data.rulelist;
	while (ritem != NULL) {
		pr = (ProcessRule *)slist_get_data (ritem);
		mitem = pr->matchlist;
		while (mitem != NULL) {
			str = (char *)slist_get_data (mitem);
			if (str != NULL)
				free (str);
			mitem = slist_get_next (mitem);
		}
		slist_free (&pr->matchlist);
		mitem = pr->rulestat.msglist;
		while (mitem != NULL) {
			str = (char *)slist_get_data (mitem);
			if (str != NULL)
				free (str);
			mitem = slist_get_next (mitem);
		}
		slist_free (&pr->rulestat.msglist);
		free (pr);
		ritem = slist_get_next (ritem);
	}
	slist_free (&module_data.rulelist);
	if (module_data.procarray != NULL)
		free (module_data.procarray);
	if (module_data.procarrayold != NULL)
		free (module_data.procarrayold);
/*
	rc = close (module_data.syncpipe[SYNCPIPE_WRITE]);
	rc = close (module_data.syncpipe[SYNCPIPE_READ]);
*/
	rc = trigger_free (&module_data.trigger);
	return NETINFO_RC_SUCCESS;
}


int netinfo_module_execute_method (char *method,
			NetinfoHeader *nh,
			char **responsedata,
			unsigned long *responsedatalen)
{
	int rc;
	int len;
	int i;
	int argdatalen;
	int proccount;
	char argdata[NETINFO_HEADER_ARGDATA_SIZE];
	char buffer[BUFFER_SIZE];
	char *p = NULL;
	char *rd = NULL;
	ProcessInfo *pi = NULL;
	if (method == NULL ||
	    nh == NULL ||
	    responsedata == NULL ||
	    responsedatalen == NULL)
		return NETINFO_RC_INVALIDARG;
	*responsedata = NULL;
	*responsedatalen = 0;
	argdatalen = NETINFO_HEADER_ARGDATA_SIZE;
	rc = netinfo_header_get_argdata (nh,
				argdata,
				&argdatalen);
	if (!strcmp (method, "curinterval")) {
		len = sprintf (buffer,
				"%lu",
				(long)module_data.curinterval);
		rd = calloc (1, len);
		memcpy (rd, buffer, len);
		*responsedata = rd;
		*responsedatalen = len;
		rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
		rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
		rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
	} else if (!strcmp (method, "processlist")) {
		proccount = module_data.procarraysize;
		pi = (ProcessInfo *)calloc (proccount, sizeof (ProcessInfo));
		if (pi == NULL)
			return NETINFO_RC_MEMORYERROR;
		rc = enumerate_processes (NULL, pi, &proccount);
		if (rc != 0) {
			vector_log_entry (module_data.moduleinfo,
				"enumerate_processes(): rc=%d(%s)\n",
				rc,
				strerror (rc));
			free (pi);
			return rc;
		}
		rd = calloc (1, BUFFER_SIZE * proccount);
		if (rd == NULL) {
			free (pi);
			return NETINFO_RC_MEMORYERROR;
		}
		p = rd;
		for (i=0 ; i<proccount ; i++) {
			rc = snprintf (buffer,
					BUFFER_SIZE,
					"%5d %-32s %c %9lu %9lu %9lu %9lu\n",
					pi[i].pi_pid,
					pi[i].pi_procname,
					pi[i].pi_state,
					pi[i].pi_utime,
					pi[i].pi_stime,
					pi[i].pi_vsize,
					pi[i].pi_rss);
			if (rc >= BUFFER_SIZE) {
				buffer[BUFFER_SIZE-2] = '\n';
				rc = BUFFER_SIZE;
			}
			strncpy (p, buffer, rc);
			p += rc;
		}
		*responsedata = rd;
		*responsedatalen = strlen (rd);
		rc = netinfo_header_set_status (nh,
					NETINFO_STATUS_RESPONSE);
		rc = netinfo_header_set_flags (nh,
					NETINFO_FLAG_DATAPRESENT);
		rc = netinfo_header_set_rc (nh,
					NETINFO_RC_SUCCESS);
		free (pi);
	} else {
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	return NETINFO_RC_SUCCESS;
}


void *poll_thread (void *arg)
{
	int             rc;
	int             oldeventrc;
	int             len;
	int		i;
	int		l;
	int		procrc;
	unsigned long	mem;
	unsigned long	tctime;
	unsigned long	octime;
	unsigned long	cputime;
	unsigned long	cpuseconds;
	char            buffer[BUFFER_SIZE];
/*
	fd_set		spfds;
	struct timeval	tv;
*/
	char		*mstr = NULL;
	SList           *ritem = NULL;
	SList           *mitem = NULL;
	EventData       *event = NULL;
	EventChain      *eventchain = NULL;
	ProcessInfo	*opi = NULL;
	ProcessInfo	*patemp = NULL;
	ProcessRule	*pr = NULL;
	ModuleData      *md = &module_data;

	sleep (2 * SINGLE_INTERVAL); /*FIXME:  module is loaded before netinfod can initialize SSL context */

	rc = eventdata_new (&event);
	eventdata_clear (event);
	md->curinterval = 0;
	md->procoldcount = PROCARRAY_SIZE;
	rc = enumerate_processes (NULL, md->procarrayold, &md->procoldcount);
	if (rc != 0) {
		vector_log_entry (md->moduleinfo,
			"enumerate_processes(): rc=%d(%s)\n",
			rc,
			strerror (rc));
	}

	while (1) {
/*
		FD_ZERO (&spfds);
		FD_SET (md->syncpipe[SYNCPIPE_READ], &spfds);
		tv.tv_sec = md->pollintervalsecs;
		tv.tv_usec = 0;
		rc = select (md->syncpipe[SYNCPIPE_READ] + 1,
				&spfds,
				NULL,
				NULL,
				&tv);
*/
		rc = trigger_sleep (md->trigger, md->pollintervalsecs);
		if (rc > 0) {
			/* got the signal, cleanup and terminate the thread */
			vector_log_entry (md->moduleinfo,
				"Exiting poll thread\n");
/*
			rc = read (md->syncpipe[SYNCPIPE_READ], buffer, 1);
*/
			eventdata_free (&event);
			pthread_exit (NULL);
			return NULL;
		}
		rc = vector_eventchain_open (md->moduleinfo, &eventchain);
		md->proccount = PROCARRAY_SIZE;
		rc = enumerate_processes (NULL, md->procarray, &md->proccount);
		if (rc != 0) {
			vector_log_entry (md->moduleinfo,
				"enumerate_processes(): rc=%d(%s)\n",
				rc,
				strerror (rc));
		}
		ritem = md->rulelist;
		while (ritem != NULL) {
			pr = (ProcessRule *)slist_get_data (ritem);
			pr->rulestat.proccount = 0;
			ritem = slist_get_next (ritem);
		}
		for (i=0 ; i<md->proccount ; i++) {
			pr = rulelist_find_process_match (md->rulelist, md->procarray[i].pi_procname);
			if (pr == NULL)
				continue;
			procrc = 0;
			pr->rulestat.proccount++;

			opi = procinfolist_find_pid (md->procarrayold,
							md->procoldcount,
							md->procarray[i].pi_pid);
			if (opi != NULL) {
				memcpy (&md->procarray[i].pi_ps, &opi->pi_ps, sizeof (ProcessStat));
				mem = md->procarray[i].pi_vsize;
				tctime = md->procarray[i].pi_utime + md->procarray[i].pi_stime;
				octime = opi->pi_utime + opi->pi_stime;
				cputime = tctime - octime;
				cpuseconds = cputime / module_data.tickssec;
				if (pr->flags & RULE_FLAG_DEBUGINFO) {
					vector_log_entry (md->moduleinfo,
						"DEBUG: pid=%d name=%s  cputime=%lu cpuseconds=%lu memuse=%lu\n",
						md->procarray[i].pi_pid,
						md->procarray[i].pi_procname,
						cputime,
						cpuseconds,
						mem);
				}
				len = 0;
				oldeventrc = pr->rulestat.eventrc;
				pr->rulestat.eventrc = RC_OKAY;
				if (pr->critmemuse != 0 && mem >= pr->critmemuse) {
					len = snprintf (buffer,
						BUFFER_SIZE,
						"%s: pid=%d  owner=%s  Memory use is %lu bytes.\n",
						md->procarray[i].pi_procname,
						md->procarray[i].pi_pid,
						md->procarray[i].pi_ownername,
						mem);
/* vector_log_entry (md->moduleinfo, buffer); */
					processrule_add_message (pr,
							RC_CRITICAL,
							buffer,
							len);
					procrc = RC_CRITICAL;
				} else if (pr->warnmemuse != 0 && mem >= pr->warnmemuse) {
					len = snprintf (buffer,
						BUFFER_SIZE,
						"%s: pid=%d  owner=%s  Memory use is %lu bytes.\n",
						md->procarray[i].pi_procname,
						md->procarray[i].pi_pid,
						md->procarray[i].pi_ownername,
						mem);
/* vector_log_entry (md->moduleinfo, buffer); */
					processrule_add_message (pr,
							RC_WARNING,
							buffer,
							len);
					procrc = RC_WARNING;
				}
				if (pr->cpucrituse != 0 && cpuseconds >= pr->cpucrituse) {
					md->procarray[i].pi_ps.cpualertintervals++;
					if (md->procarray[i].pi_ps.cpualertintervals == pr->cpuinterval) {
						len = snprintf (buffer,
							BUFFER_SIZE,
							"%s: pid=%d  owner=%s  CPU use is %lu seconds.\n",
							md->procarray[i].pi_procname,
							md->procarray[i].pi_pid,
							md->procarray[i].pi_ownername,
							cpuseconds);
/* vector_log_entry (md->moduleinfo, buffer); */
						processrule_add_message (pr,
								RC_CRITICAL,
								buffer,
								len);
						procrc = RC_CRITICAL;
					}
				} else if (pr->cpuwarnuse != 0 && cpuseconds >= pr->cpuwarnuse) {
					md->procarray[i].pi_ps.cpualertintervals++;
					if (md->procarray[i].pi_ps.cpualertintervals == pr->cpuinterval) {
						len = snprintf (buffer,
							BUFFER_SIZE,
							"%s: pid=%d  owner=%s  CPU use is %lu seconds.\n",
							md->procarray[i].pi_procname,
							md->procarray[i].pi_pid,
							md->procarray[i].pi_ownername,
							cpuseconds);
/* vector_log_entry (md->moduleinfo, buffer); */
						processrule_add_message (pr,
								RC_WARNING,
								buffer,
								len);
						procrc = RC_WARNING;
					}
				} else {
					md->procarray[i].pi_ps.cpualertintervals = 0;
				}
				if (procrc == RC_CRITICAL && (pr->flags & RULE_FLAG_TERMINATE)) {
					rc = terminate_process (md->procarray[i].pi_pid);
					len = snprintf (buffer,
						BUFFER_SIZE,
						"%s: pid=%d  owner=%s  Has been terminated.\n",
						md->procarray[i].pi_procname,
						md->procarray[i].pi_pid,
						md->procarray[i].pi_ownername);
					processrule_add_message (pr,
							RC_CRITICAL,
							buffer,
							len);
				}
			}
		}

		ritem = md->rulelist;
		while (ritem != NULL) {
			pr = (ProcessRule *)slist_get_data (ritem);
			len = 0;
			if (pr->maxprocs != 0 && pr->rulestat.proccount > pr->maxprocs) {
				len = snprintf (buffer,
					BUFFER_SIZE,
					"%lu processes running. Maximum allowed is %lu.\n",
					pr->rulestat.proccount,
					pr->maxprocs);
				processrule_add_message (pr,
						RC_CRITICAL,
						buffer,
						len);
			} else if (pr->minprocs != 0 && pr->rulestat.proccount < pr->minprocs) {
				len = snprintf (buffer,
					BUFFER_SIZE,
					"%lu processes running. Minimum allowed is %lu.\n",
					pr->rulestat.proccount,
					pr->minprocs);
				processrule_add_message (pr,
						RC_CRITICAL,
						buffer,
						len);
			}
			if (pr->rulestat.eventrc != RC_OKAY) {
				buffer[0] = 0;
				len = 0;
				mitem = pr->rulestat.msglist;
				while (mitem != NULL && len < BUFFER_SIZE) {
					mstr = slist_get_data (mitem);
					l = strlen (mstr);
					if ( (len + l + 1) >= BUFFER_SIZE)
						break;
					strcat (buffer, mstr);
					len += l;
					mitem = slist_get_next (mitem);
				}
			} else {
				len = snprintf (buffer,
						BUFFER_SIZE,
						"Okay");
			}
			if (pr->rulestat.eventrc != oldeventrc) {
				eventdata_clear (event);
				rc = eventdata_set (event,
						EVENT_FLAG_ONESHOT,
						pr->rulestat.eventrc,
						pr->objectname,
						buffer,
						len);
				rc = vector_eventchain_emit_event (eventchain,
						event);
			}
			rc = processrule_free_messages (pr);
			ritem = slist_get_next (ritem);
		}

		patemp = md->procarrayold;
		md->procarrayold = md->procarray;
		md->procarray = patemp;
		md->procoldcount = md->proccount;


		rc = vector_eventchain_close (&eventchain);
		md->curinterval++;

	}
	eventdata_free (&event);
	pthread_exit (NULL);
	return NULL;
}


int parse_configfile (char *filepath)
{
	int		rc;
	char		buffer[BUFFER_SIZE];
	char		*directive = NULL;
	char		*attribs = NULL;
	char		*attrib = NULL;
	char		*matchstr = NULL;
	char		*matchlist = NULL;
	ConfigFile	*cf = NULL;
	ProcessRule	*pr = NULL;
	if (filepath == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = configfile_open (&cf, filepath);
	if (rc != NETINFO_RC_SUCCESS)
		return rc;
	rc = configfile_search_begin (cf, "WatchProcess");
	while (1) {
		rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		directive = strtok (buffer, "'\" \t");
		if (directive[0] == '#') {
			continue;
		}
		pr = calloc (1, sizeof (ProcessRule));
		
		matchlist = strdup (strtok (NULL, "'\"\t"));
		if (matchlist[0] == 0) {
			vector_log_entry (module_data.moduleinfo,
					"WARNING: invalid 'WatchProcess' line "
					"in configuration file. Line will be "
					"skipped.");
			continue;
		}
		attribs = strtok (NULL, " \t\r\n");
		attrib = strtok (attribs, "= \t\r\n");
		while (attrib != NULL) {
			if (!strcmp (attrib, "objectname")) {
				pr->objectname = strdup (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "warnmemuse")) {
				pr->warnmemuse = atol (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "critmemuse")) {
				pr->critmemuse = atol (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "cpuuseinterval")) {
				pr->cpuinterval = atol (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "warnintervalcpuuse")) {
				pr->cpuwarnuse = atol (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "critintervalcpuuse")) {
				pr->cpucrituse = atol (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "minprocs")) {
				pr->minprocs = atol (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "maxprocs")) {
				pr->maxprocs = atol (strtok (NULL,
							", \t\r\n"));
			} else if (!strcmp (attrib, "ruledebug")) {
				if (atoi (strtok (NULL, ", \t\r\n")) == 1)
					pr->flags |= RULE_FLAG_DEBUGINFO;
			} else if (!strcmp (attrib, "terminate")) {
				if (atoi (strtok (NULL, ", \t\r\n")) == 1)
					pr->flags |= RULE_FLAG_TERMINATE;
			}
			attrib = strtok (NULL, "= \t\r\n");
		}
		if (pr->cpuinterval == 0)
			pr->cpuinterval=1;
		attrib = strtok (matchlist, ",");
		while (attrib != NULL) {
			matchstr = strdup (attrib);
			rc = slist_item_add (&pr->matchlist, matchstr);
			attrib = strtok (NULL, ",");
		}
		free (matchlist);
		vector_log_entry (module_data.moduleinfo,
				"Process watch: object='%s' warnmem=%lu critmem=%lu cpualertinterval=%lu warncpu=%lu critcpu=%lu terminate=%d\n",
				pr->objectname,
				pr->warnmemuse,
				pr->critmemuse,
				pr->cpuinterval,
				pr->cpuwarnuse,
				pr->cpucrituse,
				pr->flags & RULE_FLAG_TERMINATE);
		rc = slist_item_add (&module_data.rulelist, pr);
	}
	rc = configfile_search_end (cf);
	rc = configfile_close (&cf);
	return NETINFO_RC_SUCCESS;
}

#ifndef WIN32
int enumerate_processes (void **enumstat, ProcessInfo *proclist, int *listsize)
{
	int		rc;
	int		iproc;
	char		filepath[FILEPATH_SIZE];
	char		statdata[STATDATA_SIZE];
	struct stat	st;
	DIR		*dir = NULL;
	FILE		*fd = NULL;
	struct dirent	*de = NULL;
	ProcessInfo	*pi = NULL;
	struct passwd	*pwe = NULL;

	if (proclist == NULL || *listsize < 1)
		return EINVAL;
	if (enumstat == NULL || (enumstat != NULL && *enumstat == NULL)) {
		dir = opendir ("/proc");
		if (dir == NULL) {
			*listsize = 0;
			return errno;
		}
	} else {
		dir = (DIR *)*enumstat;
	}
	iproc = 0;
	while ( (de = readdir (dir)) != NULL && iproc < *listsize) {
		rc = snprintf (filepath,
				FILEPATH_SIZE,
				"/proc/%s",
				de->d_name);
		rc = stat (filepath, &st);
		if (!S_ISDIR (st.st_mode))
			continue;
		if (atoi (de->d_name) == 0)
			continue;
		rc = snprintf (filepath,
				FILEPATH_SIZE,
				"/proc/%s/stat",
				de->d_name);
		fd = fopen (filepath, "r");
		if (fd == NULL)
			continue;
		if (fgets (statdata, STATDATA_SIZE, fd) == NULL) {
			fclose (fd);
			continue;
		}
		fclose (fd);
		pi = &proclist[iproc];
		pi->pi_uid = st.st_uid;
		pwe = getpwuid (st.st_uid);
		if (pwe != NULL) {
			strncpy (pi->pi_ownername, pwe->pw_name, OWNERNAME_SIZE);
			pi->pi_ownername[OWNERNAME_SIZE - 1] = 0;
		/*	free (pwe); */
		} else {
			pi->pi_ownername[0] = 0;
		}
		rc = sscanf (statdata,
				"%d %32s %c %d %d %d %d %d %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu",
				&pi->pi_pid,
				pi->pi_procname,
				&pi->pi_state,
				&pi->pi_ppid,
				&pi->pi_pgrp,
				&pi->pi_session,
				&pi->pi_tty_nr,
				&pi->pi_tpgid,
				&pi->pi_flags,
				&pi->pi_minflt,
				&pi->pi_cminflt,
				&pi->pi_majflt,
				&pi->pi_cmajflt,
				&pi->pi_utime,
				&pi->pi_stime,
				&pi->pi_cutime,
				&pi->pi_cstime,
				&pi->pi_priority,
				&pi->pi_nice,
				&pi->pi_dummy1,
				&pi->pi_itrealvalue,
				&pi->pi_starttime,
				&pi->pi_vsize,
				&pi->pi_rss);
		iproc++;
	}
	*listsize = iproc;
	if (de == NULL) {
		rc = closedir (dir);
		if (enumstat != NULL)
			*enumstat = NULL;
	} else {
		if (enumstat != NULL) {
			*enumstat = dir;
			return EAGAIN;
		} else {
			rc = closedir (dir);
		}
	}
	return 0;
}
#else
int enumerate_processes (void **enumstat, ProcessInfo *proclist, int *listsize)
{
	int			rc;
	int			iproc;
	DWORD			i;
	HANDLE			hp;
	HANDLE			ht;
	PROCESS_MEMORY_COUNTERS	pmc;
	FILETIME		ftCreation;
	FILETIME		ftExit;
	FILETIME		ftKernel;
	FILETIME		ftUser;
	FILETIME		ftCpu;
	SYSTEMTIME		stCpu;
	ULARGE_INTEGER		ulKernel;
	ULARGE_INTEGER		ulUser;
	ULARGE_INTEGER		ulCpu;
	TOKEN_PRIVILEGES	tp;
	DWORD			process_count = 0;
	WTS_PROCESS_INFO	*process_info = NULL;
	char			*servername = NULL;
	ProcessInfo	*pi = NULL;

	if (proclist == NULL || *listsize < 1)
		return EINVAL;
	OpenProcessToken (GetCurrentProcess(),
			TOKEN_ADJUST_PRIVILEGES,
			&ht);
	tp.PrivilegeCount = 1;
	LookupPrivilegeValue(NULL, SE_DEBUG_NAME, &tp.Privileges[0].Luid);
	tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
	AdjustTokenPrivileges(ht, FALSE, &tp, sizeof(tp), NULL, NULL);
	servername = WTS_CURRENT_SERVER_HANDLE;
	
/* TODO
	if (enumstat == NULL || (enumstat != NULL && *enumstat == NULL)) {
		dir = opendir ("/proc");
		if (dir == NULL) {
			*listsize = 0;
			return errno;
		}
	} else {
		dir = (DIR *)*enumstat;
	}
*/
	if (!WTSEnumerateProcesses(servername,  
			0, 1, &process_info, &process_count)) {
		vector_log_entry (module_data.moduleinfo,
				"Error: WTSEnumerateProcesses failed with error %d.\n",
				GetLastError());
		CloseHandle (ht);
		return 1;
	}
	for (iproc = 0; iproc < process_count; iproc++) {
		if (iproc >= *listsize)
			break;
		pi = &proclist[iproc];
/* vector_log_entry (module_data.moduleinfo,
			"DEBUG: pid=%d\n", process_info[iproc].ProcessId); */
		if (process_info[iproc].ProcessId == 0) {
			strcpy (pi->pi_procname, "idleprocess");
			strcpy (pi->pi_ownername, "SYSTEM");
		} else {
			char user_name[UNLEN];
			char domain_name[DNLEN];
			DWORD user_length = sizeof(user_name);
			DWORD domain_length = sizeof(domain_name);
			SID_NAME_USE name_use;

			hp = OpenProcess (PROCESS_QUERY_INFORMATION |
					PROCESS_VM_READ,
					FALSE,
					process_info[iproc].ProcessId);
			if (hp != NULL) {
				GetProcessMemoryInfo (hp,
						&pmc,
						sizeof (pmc));
				GetProcessTimes (hp,
						&ftCreation,
						&ftExit,
						&ftKernel,
						&ftUser);
				memcpy (&ulKernel, &ftKernel, sizeof (ULARGE_INTEGER));
				memcpy (&ulUser, &ftUser, sizeof (ULARGE_INTEGER));
				ulCpu.QuadPart = ulKernel.QuadPart + ulUser.QuadPart;
				memcpy (&ftCpu, &ulCpu, sizeof (ULARGE_INTEGER));
				FileTimeToSystemTime (&ftCpu, &stCpu);
				CloseHandle (hp);
			} else {
				vector_log_entry (module_data.moduleinfo,
					"Error: OpenProcess() failed\n");
				memset (&pmc, 0, sizeof (pmc));
				memset (&ftCreation, 0, sizeof (FILETIME));
				memset (&ftExit, 0, sizeof (FILETIME));
				memset (&ftKernel, 0, sizeof (FILETIME));
				memset (&ftUser, 0, sizeof (FILETIME));
				memset (&ulKernel, 0, sizeof (ULARGE_INTEGER));
				memset (&ulUser, 0, sizeof (ULARGE_INTEGER));
				memset (&ulCpu, 0, sizeof (ULARGE_INTEGER));
			}
			if (!LookupAccountSid(NULL, process_info[iproc].pUserSid,
						user_name,
						&user_length,
						domain_name,
						&domain_length,
						&name_use)) {
				vector_log_entry (module_data.moduleinfo,
					"Error: LookupAccountSid failed on PID %d with error %d.\n",
					process_info[iproc].ProcessId,
					GetLastError());
				continue;
			}
			pi->pi_pid = process_info[iproc].ProcessId;
			pi->pi_session = process_info[iproc].SessionId;
			pi->pi_utime = ulUser.QuadPart / 10000000;
			pi->pi_stime = ulKernel.QuadPart / 10000000;
			pi->pi_vsize = pmc.WorkingSetSize;
			strcpy (pi->pi_procname, process_info[iproc].pProcessName);
			strcpy (pi->pi_ownername, user_name);
		}
	}
	WTSFreeMemory(process_info);
	CloseHandle (ht);
	*listsize = iproc;

	/* TODO
	if (de == NULL) {
		rc = closedir (dir);
		if (enumstat != NULL)
			*enumstat = NULL;
	} else {
		if (enumstat != NULL) {
			*enumstat = dir;
			return EAGAIN;
		} else {
			rc = closedir (dir);
		}
	}
*/
	return 0;
}
#endif

int processrule_add_message (ProcessRule *pr, int returncode, char *message, int messagelen)
{
	char		*str = NULL;
	if (pr == NULL || message == NULL)
		return NETINFO_RC_INVALIDARG;
	str = strdup (message);
	slist_item_add (&pr->rulestat.msglist, str);
	if ( returncode >  pr->rulestat.eventrc){
		pr->rulestat.eventrc = returncode;
	}
	return 0;
}


int processrule_free_messages (ProcessRule *pr)
{
	int		rc;
	char		*str = NULL;
	SList		*litem = NULL;
	if (pr == NULL)
		return NETINFO_RC_INVALIDARG;
	litem = pr->rulestat.msglist;
	while (litem != NULL) {
		str = (char *)slist_get_data (litem);
		if (str != NULL)
			free (str);
		litem = slist_get_next (litem);
	}
	rc = slist_free (&pr->rulestat.msglist);
	return 0;
}


ProcessRule *rulelist_find_process_match (SList *rulelist, char *processname)
{
	int		r;
	int		matched;
	char		*p = NULL;
	char		*mstr = NULL;
	SList		*ritem = NULL;
	SList		*mitem = NULL;
	ProcessRule	*pr = NULL;
	if (rulelist == NULL || processname == NULL)
		return NULL;
	ritem = rulelist;
	while (ritem != NULL) {
		matched = 1;
		pr = (ProcessRule *)slist_get_data (ritem);
		mitem = pr->matchlist;
		if (mitem == NULL) /* must have at least one match string */
			matched = 0;
		while (mitem != NULL) {
			mstr = (char *)slist_get_data (mitem);
			if (!strcmp (mstr, "ANY")) {
				break;
			} else {
				if (mstr[0] == '!') {
					r = 1;
					mstr++;
				} else {
					r = 0;
				}
				p = strstr (processname, mstr);
				if ( (r == 0 && p == NULL) ||
				     (r == 1 && p != NULL) ) {
					matched = 0;
					break;
				}
			}
			mitem = slist_get_next (mitem);
		}
		if (matched == 1)
			return pr;
		ritem = slist_get_next (ritem);
	}
	return NULL;
}



ProcessInfo *procinfolist_find_pid (ProcessInfo *proclist, int listsize, int pid)
{
	int		i;
	if (proclist == NULL || listsize == 0)
		return NULL;
	for (i=0 ; i<listsize ; i++) {
		if (proclist[i].pi_pid == pid)
			return &proclist[i];
	}
	return NULL;
}


int trigger_new (Trigger **tp)
{
	int		rc;
	Trigger		*t = NULL;
	if (tp == NULL)
		return EINVAL;
	*tp = NULL;
	t = calloc (1, sizeof (Trigger));
	if (t == NULL)
		return ENOMEM;
	rc = 0;
#ifndef WIN32
	rc = pipe (t->pipe);
	if (rc != NETINFO_RC_SUCCESS) {
		free (t);
		rc = errno;
	}
#else
	t->handle = CreateEvent (NULL,
				TRUE,
				FALSE,
				"NetinfoPollThread");
	if (t->handle == NULL) {
		free (t);
		rc = GetLastError();
	}
#endif
	*tp = t;
	return ESUCCESS;
}


int trigger_sleep (Trigger *t, int sleepseconds)
{
	int		rc;
#ifndef WIN32
	char		buffer[2];
	fd_set		spfds;
	struct timeval	tv;
#endif
	if (t == NULL)
		return EINVAL;
#ifndef WIN32
	FD_ZERO (&spfds);
	FD_SET (t->pipe[SYNCPIPE_READ], &spfds);
	tv.tv_sec = sleepseconds;
	tv.tv_usec = 0;
	rc = select (t->pipe[SYNCPIPE_READ] + 1,
			&spfds,
			NULL,
			NULL,
			&tv);
	if (rc != 0) {
		/* trigger occurred */
		read (t->pipe[SYNCPIPE_READ], buffer, 1);
		rc = 1;
	} 
#else
	rc = WaitForSingleObject (t->handle, sleepseconds * 1000);
	if (rc == WAIT_OBJECT_0)
		rc = 1;
	else if (rc == WAIT_TIMEOUT)
		rc = 0;
#endif
	return rc;
}


int trigger_activate (Trigger *t)
{
	int		rc;
	if (t == NULL)
		return EINVAL;
#ifndef WIN32
	rc = write (t->pipe[SYNCPIPE_WRITE], "C", 1);
	if (rc == 1)
		rc = 0;
#else
	rc = SetEvent (t->handle);
	if (rc != 0)
		rc = 0;
#endif
	return rc;
}


int trigger_free(Trigger **tp)
{
	if (tp == NULL || *tp == NULL)
		return EINVAL;
	free (*tp);
	*tp = NULL;
	return ESUCCESS;
}


int terminate_process (int pid)
{
	int			rc;
#ifdef WIN32
	HANDLE			hp;
	HANDLE			ht;
	TOKEN_PRIVILEGES	tp;
#endif

	if (pid == 0 || pid == 1) {
		vector_log_entry (module_data.moduleinfo,
		"Error: Requested to terminate pid# %d. Not allowing this action.\n",
		pid);
		return EINVAL;
	}
#ifndef WIN32
	rc = kill (pid, SIGKILL);
#else
        OpenProcessToken (GetCurrentProcess(),
			TOKEN_ADJUST_PRIVILEGES,
			&ht);
	tp.PrivilegeCount = 1;
	LookupPrivilegeValue(NULL, SE_DEBUG_NAME, &tp.Privileges[0].Luid);
	tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
	AdjustTokenPrivileges(ht, FALSE, &tp, sizeof(tp), NULL, NULL);
	hp = OpenProcess (PROCESS_QUERY_INFORMATION |
			PROCESS_VM_READ | PROCESS_TERMINATE,
			FALSE,
			pid);
	if (hp != NULL) {
		rc = TerminateProcess (hp, 9);
		if (rc != 0)
			rc = 0;
		else
			rc = GetLastError();
		CloseHandle (hp);
	} else {
		rc = GetLastError();
	}
#endif
	return rc;
}


