/************************************************************************/
/* nieh_perf2rrd.c - Netinfoevtd performance data to RRD plugin		*/
/************************************************************************/
/*
 *  (c)2007,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/



#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include <rrd.h>
#include <cutil/cutil.h>

#include <netinfo/netinfo.h>
#include <netinfo/event.h>
#include <netinfo/confxml.h>


#define MODULE_NAME			"nieh_perf2rrd"
#define MODULE_VERSION			"0.1.0"

#define NAME_SIZE			64
#define BUFFER_SIZE			512

#define	CONFIG_DELIMITERS		",\r\n"
#define LOGFILE_DEFAULT			"netinfo-perf2rrd.log"
#define CONFFILE_DEFAULT		"perf2rrd.conf"
#define	UNKNOWN_DOMAIN_DEFAULT		"unknown"

#define	NETINFO_EVENT_FLAG_ONESHOT	0x00000001

#define MODULE_FLAG_DEBUG		0x00000001
#define MODULE_FLAG_USE_EVENT_TIME	0x00000002

#define	DIRECTORY_CREATION_MODE		0755
#define	FILE_CREATION_MODE		0644

#ifndef WIN32
#define	DIR_DELIMITER			"/"
#else
#define	DIR_DELIMITER			"\\"
#endif


typedef struct _RRDInfo		RRDInfo;
struct	_RRDInfo {
	char		*typename;
	SList		*ds_list;
	SList		*rra_list;
};


typedef struct	_PerfModInfo		PerfModInfo;
struct	_PerfModInfo {
	unsigned int	flags;
	char		*config_file;
	char		*log_file;
	FILE		*log_fd;
	char		*rrd_directory;

	AVLTree		*rrdtype_tree;
	SList		*rrdtype_list;
};



/* Global Variables */
PerfModInfo	mod_info;



void debug_log_arg_array (char **arg, int count)
{
	int	i;
	for (i = 0 ; i<count ; i++)
		fprintf (mod_info.log_fd, "DEBUG: arg[%d] = '%s'\n", i, arg[i]);
	return;
}


int rrdutil_clear_config (void)
{
	char		*p;
	SList		*litem;
	SList		*li;
	RRDInfo		*ri;
	
	litem = mod_info.rrdtype_list;
	while (litem != NULL) {
		ri = slist_get_data (litem);
		// free name
		free (ri->typename);
		// free DS list
		li = ri->ds_list;
		while (li != NULL) {
			p = slist_get_data (li);
			free (p);
			li = slist_get_next (li);
		}
		slist_free (&ri->ds_list);
		// free RRA list
		li = ri->rra_list;
		while (li != NULL) {
			p = slist_get_data (li);
			free (p);
			li = slist_get_next (li);
		}
		slist_free (&ri->rra_list);
		// free RRDInfo
		free (ri);
		litem = slist_get_next (litem);
	}
	// free rrdtype_list
	slist_free (&mod_info.rrdtype_list);
	// free rrdtype_tree
	avltree_free (&mod_info.rrdtype_tree);
	return 0;
}


int cleanup (void)
{
	rrdutil_clear_config ();
	if (mod_info.rrd_directory != NULL)
		free (mod_info.rrd_directory);
	if (mod_info.log_fd != NULL)
		fclose (mod_info.log_fd);
	if (mod_info.log_file != NULL)
		free (mod_info.log_file);
	return NETINFO_RC_SUCCESS;
}


int rrdinfo_compare (RRDInfo *ri1, RRDInfo *ri2)
{
	return strcmp (ri1->typename, ri2->typename);
}


int rrdinfo_compare_typename (char *typename, RRDInfo *ri)
{
	return strcmp (typename, ri->typename);
}


int rrdutil_parse_config (char *configfile)
{
	int		rc;
	char		*type;
	char		*value;
	Configuration	*c;
	ConfigSearch	*cs;
	ConfigItem	*ci;
	RRDInfo		*ri;

	rc = avltree_new (&mod_info.rrdtype_tree, (AVLNodeCompare)rrdinfo_compare, NULL);
	if (rc != 0)
		return rc;
	
	rc = configuration_init (NULL);

	rc = configuration_new (&c);
	if (rc != 0)
		return rc;

	rc = configuration_load_file (c, configfile);
	if (rc != 0)
		return rc;

	rc = configuration_search_new (c,
					&cs,
					NULL);
	rc = configuration_search_begin (cs,
					"/configuration/perf2rrd/rrdinfo");

	// loop over the rrdinfo elements
	while (1) {
		rc = configuration_search_next_item (cs, &ci);
		if (rc != 0)
			break;

		ri = calloc (1, sizeof (RRDInfo));
		if (ri == NULL) {
			configuration_search_end (cs);
			configuration_search_free (&cs);
			configuration_free (&c);
			return ENOMEM;
		}
			
		ri->typename = configuration_item_get_attribute (ci,
							"typename");
		if (mod_info.flags & MODULE_FLAG_DEBUG)
			fprintf (mod_info.log_fd, "DEBUG: Config creating '%s' rrdinfo.\n", ri->typename);

		// loop over the rrdinfo child elements (DSs and RRAs)
		ci = configuration_item_get_first_child (ci);
		while (ci != NULL) {
			type = configuration_item_get_attribute (ci, "type");
			value = configuration_item_get_content (ci);
			if (mod_info.flags & MODULE_FLAG_DEBUG) {
				fprintf (mod_info.log_fd,
					"DEBUG: Config rrdinfo type='%s' value='%s'\n",
					type,
					value);
			}
			if (!strcmp (type, "DS")) {
				rc = slist_item_add (&ri->ds_list, value);
				// the value will be free'd with the slist
			} else if (!strcmp (type, "RRA")) {
				rc = slist_item_add (&ri->rra_list, value);
				// the value will be free'd with the slist
			} else {
				// unknown type
				free (value);
			}
			free (type);
			ci = configuration_item_get_next_sibling (ci);
		}
		rc = avltree_insert (mod_info.rrdtype_tree, ri);
		rc = slist_item_add (&mod_info.rrdtype_list, ri);
		fprintf (mod_info.log_fd,
			"INFO: Added typename '%s' to configuration\n",
			ri->typename);
	}
	rc = configuration_search_end (cs);

	// cleanup
	rc = configuration_search_free (&cs);
	rc = configuration_free (&c);

	return 0;
}


int rrdutil_update_rrd (char *rrdpath, char *updatestr)
{
	int	rc;
	char	*args[2];

	args[0] = updatestr;
	args[1] = "\0";

	if (mod_info.flags & MODULE_FLAG_DEBUG)
		debug_log_arg_array (args, 1);

	rrd_clear_error ();
	rc = rrd_update_r (rrdpath, NULL, 1, (const char **)args);
	if (rc == -1) {
		fprintf (mod_info.log_fd, "ERROR: rrd_update_r(): '%s'\n", rrd_get_error ());
	}
	return rc;
}


int rrdutil_create_rrd (char *rrdpath, RRDInfo *ri)
{
	int	i;
	int	rc;
	time_t	ct;
	char	*args[BUFFER_SIZE];
	char	buffer[BUFFER_SIZE];
	SList	*litem;
	char	*str;

	if (ri == NULL)
		return EINVAL;

	ct = time (NULL);
	snprintf (buffer, BUFFER_SIZE, "%lu", ct - 1);

	// start adding arguments at offset #0
	i = 0;

	// DS arguments
	litem = ri->ds_list;
	while (litem != NULL) {
		str = slist_get_data (litem);
		args[i] = str;
		litem = slist_get_next (litem);
		i++;
	}

	// RRA arguments
	litem = ri->rra_list;
	while (litem != NULL) {
		str = slist_get_data (litem);
		args[i] = str;
		litem = slist_get_next (litem);
		i++;
	}
	
	// last argument is a zero-length string
	args[i] = "\0";

	if (mod_info.flags & MODULE_FLAG_DEBUG)
		debug_log_arg_array (args, i);

	// create the RRD and set it's permissions
	rrd_clear_error ();
	rc = rrd_create_r (rrdpath, 300, ct, i, (const char **)args);
	if (rc == -1) {
		fprintf (mod_info.log_fd, "ERROR: rrd_create_r(): '%s'\n", rrd_get_error ());
	}
	rc = chmod (rrdpath, FILE_CREATION_MODE);
	return rc;
}


int handle_event (EventData *event)
{
	int		rc;
	int		len;
	uint32_t	eventflags;
	/* uint64_t	eventtimestamp; */
	time_t		ct;
	struct stat	st;
	char		hostname[NAME_SIZE];
	char		objectname[NAME_SIZE];
	char		data[BUFFER_SIZE];
	char		rrdupdate[BUFFER_SIZE];
	char		rrdpath[BUFFER_SIZE];
	char		dir[BUFFER_SIZE];
	char		*p;
	char		*perftype;
	char		*perfdata;
	char		*domainname;
	RRDInfo		*ri;

	if (event == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = eventdata_get_flags (event, &eventflags);

	// We only want to process performance-data events
	if (eventflags != 0)
		return NETINFO_RC_SUCCESS;

	// break up hostname and domainname
	len = NAME_SIZE;
	rc = eventdata_get_hostname (event, hostname, &len);
	hostname[len] = 0;
	strtok_r (hostname, ".", &p);
	domainname = strtok_r (NULL, "\r\n", &p);
	if (domainname == NULL) {
		fprintf (mod_info.log_fd,
			"INFO: domain is NULL for '%s', using '%s'\n",
			hostname,
			UNKNOWN_DOMAIN_DEFAULT);
		domainname = UNKNOWN_DOMAIN_DEFAULT;
	}

	// break up data element
	len = BUFFER_SIZE;
	rc = eventdata_get_data (event, data, &len);
	data[len] = 0;
	// discard service info line
	strtok_r (data, "|", &p);
	// get perftype
	perftype = strtok_r (NULL, "=", &p);
	if (perftype == NULL) {
		if (mod_info.flags & MODULE_FLAG_DEBUG)
			fprintf (mod_info.log_fd, "DEBUG: perftype is NULL\n");
		return NETINFO_RC_SUCCESS;
	}

	// find the RRDInfo structure for the given perftype
	ri = avltree_find_by_compare (mod_info.rrdtype_tree,
					(AVLNodeCompare)rrdinfo_compare_typename,
					perftype);
	if (ri == NULL) {
		fprintf (mod_info.log_fd,
			"WARNING: Invalid typename '%s', nothing to do for this event.\n",
			perftype);
		return NETINFO_RC_SUCCESS;
	}

	// get perfdata
	perfdata = strtok_r (NULL, "\r\n", &p);
	if (perfdata == NULL) {
		if (mod_info.flags & MODULE_FLAG_DEBUG)
			fprintf (mod_info.log_fd, "DEBUG: perfdata is NULL\n");
		return NETINFO_RC_SUCCESS;
	}

	// Convert the semicolons to colons (for RRD update)
	// This will obviously only work if the performance data is
	// semicolon delimited, and the data is in the same order as
	// the DS elements in the RRD file.
	for (p=perfdata ; *p != '\0' ; p++) {
		if (*p == ';')
			*p = ':';
	}

	// Get the objectname and replace any forbidden(filesystem) characters
	len = NAME_SIZE;
	rc = eventdata_get_object (event, objectname, &len);
	objectname[len] = 0;
	for (p=objectname ; *p != '\0' ; p++) {
		if (*p == '/'
		    || *p == '\\') {
			*p = '~';
		}
	}

	// debug information
	if (mod_info.flags & MODULE_FLAG_DEBUG) {
		fprintf (mod_info.log_fd, "DEBUG: hostname=%s\n", hostname);
		fprintf (mod_info.log_fd, "DEBUG: domainname=%s\n", domainname);
		fprintf (mod_info.log_fd, "DEBUG: objectname=%s\n", objectname);
		fprintf (mod_info.log_fd, "DEBUG: perftype=%s\n", perftype);
		fprintf (mod_info.log_fd, "DEBUG: perfdata=%s\n", perfdata);
	}

	// determine which timestamp to use
	//if (mod_info.flags & MODULE_FLAG_USE_EVENT_TIME)
	//	rc = eventdata_get_timestamp (event, &ct);
	//else
		ct = time (NULL);

	// build RRD update string
	rc = snprintf (rrdupdate,
			BUFFER_SIZE,
			"%lu:%s",
			time (NULL),
			perfdata);

	// build RRD path
	rc = snprintf (rrdpath,
			BUFFER_SIZE,
			"%s/%s/%s/%s__%s.rrd",
			mod_info.rrd_directory,
			domainname,
			hostname,
			perftype,
			objectname);

	// make sure the RRD file exists
	rc = stat (rrdpath, &st);
	if (mod_info.flags & MODULE_FLAG_DEBUG)
		fprintf (mod_info.log_fd,
			"DEBUG: stat(%s): rc=%d errno=%d\n",
			rrdpath,
			rc,
			errno);
	if (rc != 0) {
		// Make sure the domain directory exists, create it if needed
		rc = snprintf (dir,
				BUFFER_SIZE,
				"%s/%s",
				mod_info.rrd_directory,
				domainname);
		rc = stat (dir, &st);
		if (mod_info.flags & MODULE_FLAG_DEBUG)
			fprintf (mod_info.log_fd,
				"DEBUG: stat(%s): rc=%d errno=%d\n",
				dir,
				rc,
				errno);
		if (rc != 0) {
			rc = mkdir (dir, DIRECTORY_CREATION_MODE);
			if (rc != 0) {
				rc = errno;
				fprintf (mod_info.log_fd,
					"ERROR: mkdir('%s') failed: rc=%d\n",
					dir, errno);
				return rc;
			}
			fprintf (mod_info.log_fd,
				"INFO: Created '%s' directory (new host).\n", dir);
		}

		// Make sure the host directory exists, create it if needed
		rc = snprintf (dir,
				BUFFER_SIZE,
				"%s/%s/%s",
				mod_info.rrd_directory,
				domainname,
				hostname);
		rc = stat (dir, &st);
		if (mod_info.flags & MODULE_FLAG_DEBUG)
			fprintf (mod_info.log_fd,
				"DEBUG: stat(%s): rc=%d errno=%d\n",
				dir,
				rc,
				errno);
		if (rc != 0) {
			rc = mkdir (dir, DIRECTORY_CREATION_MODE);
			if (rc != 0) {
				rc = errno;
				fprintf (mod_info.log_fd,
					"WARNING: mkdir('%s') failed: rc=%d\n",
					dir, errno);
				return rc;
			}
			fprintf (mod_info.log_fd,
				"INFO: Created '%s' directory (new host).\n", dir);
		}

		// create the RRD file
		rc = rrdutil_create_rrd (rrdpath, ri);
		if (rc != 0) {
			fprintf (mod_info.log_fd,
				"WARNING: rrdutil_create_rrd(%s, %s): rc=%d\n",
				rrdpath,
				ri->typename,
				rc);
			return rc;
		}
	}

	// update the RRD file
	rc = rrdutil_update_rrd (rrdpath, rrdupdate);
	if (rc != 0) {
		fprintf (mod_info.log_fd,
			"WARNING: rrdutil_update_rrd(%s, '%s'): rc=%d\n",
			rrdpath,
			rrdupdate,
			rc);
	}

	return rc;
}


int module_init (char *args)
{
	int		rc;
	struct stat	st;
	char		*str = NULL;
	char		*argtokp = NULL;
	char		*valtokp = NULL;

	// initialize variables
	mod_info.flags = 0;
	mod_info.log_file = NULL;
	mod_info.log_fd = NULL;
	mod_info.rrd_directory = NULL;

	// parse module arguments
	if (args == NULL)
		return NETINFO_RC_INVALIDARG;
	str = strtok_r (args, CONFIG_DELIMITERS, &argtokp);
	while (1) {
		if (str == NULL)
			break;
		if (strstr (str, "rrddir=")) {
			str = strtok_r (str, "=", &valtokp);
			mod_info.rrd_directory = strdup (strtok_r (NULL,
						CONFIG_DELIMITERS, &valtokp));
		} else if (strstr (str, "debug=")) {
			str = strtok_r (str, "=", &valtokp);
			if (atoi((strtok_r (NULL,
					CONFIG_DELIMITERS, &valtokp))) == 1)
				mod_info.flags |= MODULE_FLAG_DEBUG;
		} else if (strstr (str, "useeventtime=")) {
			str = strtok_r (str, "=", &valtokp);
			if (atoi((strtok_r (NULL,
					CONFIG_DELIMITERS, &valtokp))) == 1)
				mod_info.flags |= MODULE_FLAG_USE_EVENT_TIME;
		} else if (strstr (str, "log=")) {
			str = strtok_r (str, "=", &valtokp);
			mod_info.log_file = strdup (strtok_r (NULL,
						CONFIG_DELIMITERS, &valtokp));
		} else if (strstr (str, "config=")) {
			str = strtok_r (str, "=", &valtokp);
			mod_info.config_file = strdup (strtok_r (NULL,
						CONFIG_DELIMITERS, &valtokp));
		} else { 
			mod_info.log_file = strdup (str);
		}
		str = strtok_r (NULL, CONFIG_DELIMITERS, &argtokp);
	}

	// config file
	if (mod_info.config_file == NULL) {
		mod_info.config_file = strdup (SYSCONFDIR DIR_DELIMITER CONFFILE_DEFAULT);
	}

	// open log file
	if (mod_info.log_file == NULL) {
		mod_info.log_file = strdup (LOGDIR DIR_DELIMITER LOGFILE_DEFAULT);
	}
	mod_info.log_fd = fopen (mod_info.log_file, "a");
	if (mod_info.log_fd == NULL) {
		cleanup ();
		return NETINFO_RC_INVALIDARG;
	}

	// log module information
	fprintf (mod_info.log_fd, "INFO: %s version %s\n", MODULE_NAME, MODULE_VERSION);
	fprintf (mod_info.log_fd, "INFO: Parameter rrd_directory=%s\n",
		mod_info.rrd_directory);
	fprintf (mod_info.log_fd, "INFO: Config flags=%d\n", mod_info.flags);

	// Validate module parameters
	if (mod_info.rrd_directory == NULL) {
		fprintf (mod_info.log_fd,
			"ERROR: rrd_directory was not specified.\n");
		cleanup ();
		return NETINFO_RC_INVALIDARG;
	}
	rc = stat (mod_info.rrd_directory, &st);
	if (rc == -1) {
		rc = errno;
		fprintf (mod_info.log_fd,
			"ERROR: rrd_directory stat(): rc=%d(%s)\n",
			rc,
			strerror (rc));
		cleanup ();
		return rc;
	}
	if (S_ISDIR(st.st_mode) == 0) {
		fprintf (mod_info.log_fd,
			"ERROR: rrd_directory does not point to a directory.\n");
		cleanup ();
		return rc;
	}

	// Load configuration
	rc = rrdutil_parse_config (mod_info.config_file);
	if (rc != 0) {
		fprintf (mod_info.log_fd,
			"ERROR: Failure while parsing configuration file.\n");
		cleanup ();
		return rc;
	}

	fprintf (mod_info.log_fd, "INFO: Module initialized successfully.\n");
	return NETINFO_RC_SUCCESS;
}


int module_cleanup (void)
{
	int	rc;
	fprintf (mod_info.log_fd, "INFO: Unloading module.\n");
	rc = cleanup ();
	return NETINFO_RC_SUCCESS;
}


