/************************************************************************/
/* pcrelog.c - PCRE Log parsing library source                          */
/************************************************************************/
/*
 *  (c)2008 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/


#define	_FILE_OFFSET_BITS	64
#define	_LARGEFILE64_SOURCE


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <cutil/cutil.h>

#include <pcre.h>

#include <pcrelog.h>

#define	BUFFER_SIZE		4096


#define	RULE_FLAG_USE_INTERVAL	0x00000001


/* win32 doesn't have ftello, but ftell on that platform returns a 'long' */
/* TODO: Move to netinfo_portable.h */
#ifndef ftello
#define ftello(x)	ftell(x)
#endif


/* structure used to pass pcre data to callbacks */
/*
	.subject	  The string which matched the PCRE rule
	.ivectab	  The vector table data from pcre_exec
	.substrings	  The count of substrings from prce_exec
*/
struct	_MatchData {
	char		*subject;
	int		*ivectab;
	int		substrings;
};


/* one structure per input (file) */
/*
	.name		   A name for this input source
	.path		   The path to the source file
	.fd		   The file descriptor for the input source
	.flags		   Flags for this source (FLAG_NOREOPEN, etc)
	.state		   State of this source (STATE_OPEN, etc)
	.offset		   Current read offset within the source
	.last_read	   Timestamp of the last read from the source
	.last_match	   Timestamp of the last match from the source
	.matchgroup	   The matchgroup associated with this source
	.buffer		   Storage for the last line(s) read from the
			   input source
	.bufalloclen	   The allocated length of .buffer
	.bufdatalen	   The length of the data currently in .buffer
*/
struct	_InputSource {
	char		*name;
	char		*path;
	FILE		*fd;
	int		flags;
	int		state;
	ino_t		inode;
	off_t		offset;
	time_t		last_read;
	time_t		last_match;
	MatchGroup	*matchgroup;
	char		*buffer;
	int		bufalloclen;
	int		bufdatalen;
};


/* one structure per match template part */
/*
	.type		   Type of this part (MT_TYPE_STRING|MT_TYPE_BACKREF)
	.number		   For TYPE_BACKREF, the substring #
	.str		   For TYPE_STRING, the text string to use
*/
struct	_MatchTemplate {
	int	type;
	union {
		unsigned int	number;
		char		*string;
	} value;
};


/* structure used to hold MatchRule attributes (for user application) */
/*
	.name		  The attribute name (string)
	.value		  The attribute value (string)
*/
struct	_MatchRuleAttribute {
	char		*name;
	char		*value;
};


/* one structure per pcre rule */
/*
	.refcount	   Reference counter for this rule
	.id		   An ID for this rule (application provided)
	.desc		   A description for this rule (app provided)
	.pcre		   The compiled pcre
	.ruleflags	   Flags for this rule
	.matches_total	   Running total of the times this rule 'matched'
	.last_match_time   Timestamp from the last time this rule 'matched'
	.intv_matches	   # of matches in the configured match interval
	.intv_first_match  Timestamp of the first match in this interval
	.intv_seconds	   The # of seconds in a match interval
	.intv_req_matches  The # of seconds in a match interval
	.action_function   Function to call in a 'match'
	.action_arg	   Argument to send to the action function
	.template_str	   Output template string (raw string)
	.template_list	   List of template string parts 
	.attribute_list	   List of attributes for application use 
*/
struct	_MatchRule {
	int			refcount;
	unsigned int		id;
	char			*desc;
	pcre			*pcre_match;
	unsigned int		ruleflags;
	unsigned int		matches_total;
	time_t			last_match_time;
	unsigned int		intv_matches;
	time_t			intv_first_match;
	unsigned int		intv_seconds;
	unsigned int		intv_req_matches;
	MatchActionCallback	action_function;
	void			*action_arg;
	char			*template_str;
	SList			*template_list;
	SList			*attribute_list;
};


/* one structure per 'match-group' */
/*
	.name		   A name for this match group
	.pcre_prematch	   compiled pcre which must 'match' in order to
			   continue processing rules from the group
	.matchrule_list	   linked list of match rules in this group
	.match_action      The 'action' function to be attached to each
			   rule added by the 'load_rule_file' function.
	.action_arg	   The (optional) argument which will be set in
			   the MatchRule when a match occurs.
	.erroffset	   If the PCRE fails to compile, this holds the
			   offset into the string where the error occured
	.errstr		   If the PCRE fails to compile, this holds the
			   error string
*/
struct	_MatchGroup {
	char			*name;
	pcre			*pcre_prematch;
	SList			*matchrule_list;
	MatchActionCallback	match_action;
	void			*action_arg;
	int			erroffset;
	char			*errstr;
};


/* Function Prototypes */
int inputsource_line_match (InputSource *is);


int matchrule_new (MatchRule **mrp, pcre *p, MatchActionCallback action,
		   void *actionarg)
{
	MatchRule	*mr = NULL;
	if (mrp == NULL || p == NULL || action == NULL)
		return EINVAL;
	*mrp = NULL;
	mr = calloc (1, sizeof (MatchRule));
	if (mr == NULL)
		return errno;
	mr->action_function = action;
	mr->pcre_match = p;
	mr->action_arg = actionarg;
	*mrp = mr;
	mr->refcount++;
	return 0;
}


int matchrule_free (MatchRule **mrp)
{
	MatchRule	*mr = NULL;
	if (mrp == NULL || *mrp == NULL)
		return EINVAL;
	mr = *mrp;
	mr->refcount--;
	if (mr->refcount > 0)
		return 0;
	pcre_free (mr->pcre_match);
	if (mr->desc != NULL)
		free (mr->desc);
	if (mr->template_str != NULL)
		free (mr->template_str);
	if (mr->template_list != NULL)
		slist_free (&mr->template_list);
	free (mr);
	*mrp = NULL;
	return 0;
}


int matchrule_set_id (MatchRule *mr, unsigned int id)
{
	if (mr == NULL)
		return EINVAL;
	mr->id = id;
	return 0;
}


int matchrule_get_id (MatchRule *mr, unsigned int *idp)
{
	if (mr == NULL || idp == NULL)
		return EINVAL;
	*idp = mr->id;
	return 0;
}


int matchrule_set_desc (MatchRule *mr, char *desc)
{
	if (mr == NULL || desc == NULL)
		return EINVAL;
	if (mr->desc != NULL)
		free (mr->desc);
	mr->desc = strdup (desc);
	if (mr->desc == NULL)
		return ENOMEM;
	return 0;
}


int matchrule_get_desc (MatchRule *mr, char *desc, int desclen)
{
	int		len;
	if (mr == NULL || desc == NULL)
		return EINVAL;
	len = strlen (mr->desc);
	if (len > desclen)
		len = desclen-1;
	memcpy (desc, mr->desc, len);
	desc[len] = 0;
	return 0;
}

int matchrule_enable_match_interval (MatchRule *mr,
				unsigned int intvseconds,
				unsigned int intvmatches)
{
	if (mr == NULL)
		return EINVAL;
	mr->intv_seconds = intvseconds;
	mr->intv_req_matches = intvmatches;
	mr->ruleflags |= RULE_FLAG_USE_INTERVAL;
	return 0;
}


int matchrule_disable_match_interval (MatchRule *mr)
{
	if (mr == NULL)
		return EINVAL;
	mr->ruleflags -= (mr->ruleflags & RULE_FLAG_USE_INTERVAL);
	return 0;
}


int matchrule_get_match_interval (MatchRule *mr,
				unsigned int *intvsecondsp,
				unsigned int *intvmatchesp)
{
	if (mr == NULL || intvsecondsp == NULL || intvmatchesp == NULL)
		return EINVAL;
	*intvsecondsp = mr->intv_seconds;
	*intvmatchesp = mr->intv_req_matches;
	return 0;
}


int matchrule_set_message (MatchRule *mr, char *message)
{
	int		rc;
	int		len;
	int		val;
	int		inbr;
	char		c;
	char		*p = NULL;
	char		*s = NULL;
	char		*e = NULL;
	char		*m = NULL;
	MatchTemplate	*mt = NULL;
	if (mr == NULL || message == NULL)
		return EINVAL;
	if (mr->template_str != NULL)
		free (mr->template_str);
	if (mr->template_list != NULL)
		slist_free (&mr->template_list);
	mr->template_str = strdup (message);
	if (mr->template_str == NULL)
		return ENOMEM;
	/* Break the message up into the Template parts */
	m = strdup (message);
	len = strlen (m);
	p = m;
	s = m;
	inbr = 0;
	while (p < m+len) {
		if (*p == '$') {
			inbr = 1;
			*p = 0;
			if (p > m) {
				e = strdup (s);
				mt = calloc (1, sizeof (MatchTemplate));
				mt->type = MT_TYPE_STRING;
				mt->value.string = e;
				rc = slist_item_add (&mr->template_list,
						mt);
			}
			p++;
			s = p;
			continue;
		}
		if (inbr == 1) {
			if (*p < '0' || *p > '9') {
				inbr = 0;
				c = *p;
				*p = 0;
				val = atoi (s);
				*p = c;
				mt = calloc (1, sizeof (MatchTemplate));
				mt->type = MT_TYPE_BACKREF;
				mt->value.number = val;
				rc = slist_item_add (&mr->template_list,
						mt);
				s = p;
			} else {
				p++;
			}
		} else {
			p++;
		}
	}
	if (inbr == 1) {
		val = atoi (s);
		mt = calloc (1, sizeof (MatchTemplate));
		mt->type = MT_TYPE_BACKREF;
		mt->value.number = val;
		rc = slist_item_add (&mr->template_list,
				mt);
	} else {
		e = strdup (s);
		mt = calloc (1, sizeof (MatchTemplate));
		mt->type = MT_TYPE_STRING;
		mt->value.string = e;
		rc = slist_item_add (&mr->template_list,
				mt);
	}
	free (m);
	return 0;
}


int matchrule_set_argument (MatchRule *mr, void *arg)
{
	if (mr == NULL)
		return EINVAL;
	mr->action_arg = arg;
	return 0;
}


int matchrule_get_argument (MatchRule *mr, void **argp)
{
	if (mr == NULL || argp == NULL)
		return EINVAL;
	*argp = mr->action_arg;
	return 0;
}


int matchrule_set_attribute (MatchRule *mr, char *name, char *value)
{
	int			rc;
	MatchRuleAttribute	*mra = NULL;
	if (mr == NULL || name == NULL || value == NULL)
		return EINVAL;
	if (matchrule_get_attribute (mr, name) != NULL)
		return EEXIST;
	mra = calloc (1, sizeof (MatchRuleAttribute));
	if (mra == NULL)
		return ENOMEM;
	mra->name = strdup (name);
	if (mra == NULL) {
		free (mra);
		return ENOMEM;
	}
	mra->value = strdup (value);
	if (mra == NULL) {
		free (mra->name);
		free (mra);
		return ENOMEM;
	}
	rc = slist_item_add (&mr->attribute_list, mra);
	if (rc != 0) {
		free (mra->name);
		free (mra->value);
		free (mra);
		return ENOMEM;
	}
	return 0;
}


MatchRuleAttribute *matchrule_get_attribute (MatchRule *mr, char *name)
{
	SList			*litem = NULL;
	MatchRuleAttribute	*mra = NULL;
	if (mr == NULL || name == NULL)
		return NULL;
	litem = mr->attribute_list;
	while (litem != NULL) {
		mra = slist_get_data (litem);
		if (mra != NULL) {
			if (!strcmp (mra->name, name))
				return mra;
		}
		litem = slist_get_next (litem);
	}
	return NULL;
}


char *matchrule_get_attribute_value (MatchRule *mr, char *name)
{
	SList			*litem = NULL;
	MatchRuleAttribute	*mra = NULL;
	if (mr == NULL || name == NULL)
		return NULL; 
	litem = mr->attribute_list;
	while (litem != NULL) {
		mra = slist_get_data (litem);
		if (mra != NULL) {
			if (!strcmp (mra->name, name))
				return mra->value;
		}
		litem = slist_get_next (litem);
	}
	return NULL;
}


int matchrule_delete_attribute (MatchRule *mr, char *name)
{
	MatchRuleAttribute	*mra = NULL;
	if (mr == NULL || name == NULL)
		return EINVAL;
	mra = matchrule_get_attribute (mr, name);
	if (mra == NULL)
		return ENOENT;
	slist_item_remove (&mr->attribute_list, mra);
	free (mra->name);
	free (mra->value);
	free (mra);
	return 0;
}


int matchrule_construct_message (MatchRule *mr, MatchData *md, char *buffer,
				int *buflenp)
{
	int		rc;
	int		len;
	char		str[BUFFER_SIZE];
	SList		*litem = NULL;
	MatchTemplate	*mt = NULL;
	if (mr == NULL || md == NULL || buffer == NULL)
		return EINVAL;
        buffer[0] = 0;
        len = *buflenp - 1;
        litem = mr->template_list;
        while (litem != NULL) {
                mt = slist_get_data (litem);
                if (mt->type == MT_TYPE_STRING) {
                        strncat (buffer, mt->value.string, len);
                        len -= strlen (mt->value.string);
                } else if (mt->type == MT_TYPE_BACKREF) {
                        if (mt->value.number <= md->substrings) {
                                rc = pcre_copy_substring (md->subject,
                                                (int *)md->ivectab,
                                                md->substrings,
                                                mt->value.number,
                                                str,
                                                BUFFER_SIZE);
                                strncat (buffer, str, len);
                                len -= strlen (str);
                        }
                } else {
                }
                litem = slist_get_next (litem);
        }
	*buflenp = *buflenp - len;
	return 0;
}


int matchgroup_new (MatchGroup **mgp, char *name, char *prematch)
{
	MatchGroup	*mg = NULL;
	if (mgp == NULL || prematch == NULL)
		return EINVAL;
	*mgp = NULL;
	mg = calloc (1, sizeof (MatchGroup));
	if (mg == NULL)
		return errno;
	if (name != NULL)
		mg->name = strdup (name);
	mg->pcre_prematch = pcre_compile (prematch,
			0,
			(const char **)&mg->errstr,
			&mg->erroffset,
			NULL);
	if (mg->pcre_prematch == NULL) {
		if (mg->name != NULL)
			free (mg->name);
		free (mg);
		return 1;
	}
	*mgp = mg;
	return 0;
}


int matchgroup_free (MatchGroup **mgp)
{
	int		rc;
	MatchGroup	*mg = NULL;
	MatchRule	*mr = NULL;
	SList		*litem = NULL;
	if (mgp == NULL || *mgp == NULL)
		return EINVAL;
	mg = *mgp;
	if (mg->name != NULL)
		free (mg->name);
	if (mg->pcre_prematch != NULL)
		pcre_free (mg->pcre_prematch);
	litem = mg->matchrule_list;
	while (litem != NULL) {
		mr = slist_get_data (litem);
		rc = matchrule_free (&mr);
		litem = slist_get_next (litem);
	}
	free (mg);
	*mgp = NULL;
	return 0;
}


int matchgroup_set_action_callback (MatchGroup *mg, MatchActionCallback maf,
					void *callbackarg)
{
	if (mg == NULL)
		return EINVAL;
	mg->match_action = maf;
	mg->action_arg = callbackarg;
	return 0;
}


int matchgroup_add_rule (MatchGroup *mg, MatchRule *mr)
{
	int	rc;
	if (mg == NULL || mr == NULL)
		return EINVAL;
	rc = slist_item_add (&mg->matchrule_list, mr);
	mr->refcount++;
	return rc;
}


int matchgroup_load_rule_file (MatchGroup *mg, char *rulefile, int filetype)
{
	int		rc;
	int		fileline;
	int		erroffset;
	unsigned int	id;
	unsigned int	intvmatches;
	unsigned int	intvseconds;
	char		line[BUFFER_SIZE];
	char		*errstr = NULL;
	char		*str = NULL;
	char		*value = NULL;
	MatchRule	*mr = NULL;
	pcre		*p = NULL;
	FILE		*rfd = NULL;
	intvmatches = 0;
	intvseconds = 0;
	if (mg == NULL || rulefile == NULL)
		return EINVAL;
	rfd = fopen (rulefile, "r");
	if (rfd == NULL)
		return errno;
	fileline = 0;
	while (!feof (rfd)) {
		str = fgets (line, BUFFER_SIZE, rfd);
		if (str == NULL)
			break;
		fileline++;
		if (str[0] == '#')
			continue;
		rc = strlen (line);
		if (line[rc-1] == '\n')
			line[rc-1] = 0;
		/* rule.id */
		str = strstr (line, "rule.id=");
		if (str != NULL) {
			str += 8;
			id = atol (str);
			rc = matchrule_set_id (mr, id);
		}
		/* rule.interval_seconds */
		str = strstr (line, "rule.interval_seconds=");
		if (str != NULL) {
			str += 22;
			intvseconds = atol (str);
		}
		/* rule.interval_matches */
		str = strstr (line, "rule.interval_matches=");
		if (str != NULL) {
			str += 22;
			intvmatches = atol (str);
		}
		/* rule.interval_enable */
		str = strstr (line, "rule.interval_enable=yes");
		if (str != NULL) {
			rc = matchrule_enable_match_interval (mr,
							intvseconds,
							intvmatches);
		}
		/* rule.desc */
		str = strstr (line, "rule.desc=");
		if (str != NULL) {
			str += 10;
			rc = matchrule_set_desc (mr, str);
		}
		/* rule.msg */
		str = strstr (line, "rule.msg=");
		if (str != NULL) {
			str += 9;
			rc = matchrule_set_message (mr, str);
		}
		/* rule.attribute */
		str = strstr (line, "rule.attribute=");
		if (str != NULL) {
			str += 15;
			value = strstr (str, "=");
			if (value != NULL) {
				*value = 0;
				value++;
				rc = matchrule_set_attribute (mr, str, value);
			}
		}
		/* rule.regex */
		str = strstr (line, "rule.regex=");
		if (str != NULL) {
			str += 11;
			p = pcre_compile (str,
					0,
					(const char **)&errstr,
					&erroffset,
					NULL);
			if (p == NULL) {
				continue;
			}
			rc = matchrule_new (&mr, p, mg->match_action, NULL);
			if (rc != 0) {
				pcre_free (p);
				fclose (rfd);
				return rc;
			}
			rc = matchgroup_add_rule (mg, mr);
			if (rc != 0) {
				fclose (rfd);
				return rc;
			}
		}
	}
	fclose (rfd);
	return 0;
}


int inputsource_new (InputSource **isp, char *name, char *path, int buflen)
{
	InputSource	*is = NULL;
	if (isp == NULL || path == NULL)
		return EINVAL;
	*isp = NULL;
	is = calloc (1, sizeof (InputSource));
	if (is == NULL)
		return errno;
	if (path != NULL)
		is->path = strdup (path);
	if (is->path == NULL) {
		free (is);
		return ENOMEM;
	}
	if (buflen < BUFFER_SIZE)
		buflen = BUFFER_SIZE;
	is->buffer = calloc(1, buflen);	
	if (is->buffer == NULL) {
		free (is->path);
		free (is);
		return ENOMEM;
	}
	is->bufalloclen = buflen;

	if (name != NULL)
		is->name = strdup (name);
	is->fd = NULL;
	*isp = is;
	return 0;
}


int inputsource_free (InputSource **isp)
{
	InputSource	*is = NULL;
	if (isp == NULL || *isp == NULL)
		return EINVAL;
	is = *isp;
	if (is->name != NULL)
		free (is->name);
	if (is->path != NULL)
		free (is->path);
	if (is->buffer != NULL)
		free (is->buffer);
	free (is);
	*isp = NULL;
	return 0;
}


int inputsource_set_matchgroup (InputSource *is, MatchGroup *mg)
{
	if (is == NULL || mg == NULL)
		return EINVAL;
	is->matchgroup = mg;
	return 0;
}


int inputsource_open (InputSource *is)
{
	int		rc;
	struct stat	st;
	if (is == NULL)
		return EINVAL;
	if (is->fd != NULL)
		return EBUSY;
	is->fd = fopen (is->path, "r");
	if (is->fd == NULL)
		return errno;
	rc = stat (is->path, &st);
	is->inode = st.st_ino;	
	is->offset = fseek (is->fd, 0, SEEK_END);
	if (is->offset == -1) {
		rc = fclose (is->fd);
		return errno;
	}
	is->state |= IS_STATE_OPEN;
	return 0;
}


int inputsource_close (InputSource *is)
{
	int		rc;
	if (is == NULL)
		return EINVAL;
	if (is->fd != NULL) {
		rc = fclose (is->fd);
		is->fd = NULL;
		is->state =- (is->state & IS_STATE_OPEN);
	}
	return 0;
}


int inputsource_line_match (InputSource *is)
{
	int		rc;
	int		matchrc;
	int		td;
	int		ivectab[32];
	MatchData	md;
	SList		*litem = NULL;
	MatchRule	*mr = NULL;
	if (is == NULL || is->fd == NULL)
		return EINVAL;
	fgets (is->buffer, is->bufalloclen, is->fd);
	is->buffer[is->bufalloclen-1] = 0;
	is->last_read = time (NULL);
	/* Does it match the group's prematch */
	matchrc = pcre_exec (is->matchgroup->pcre_prematch,
			NULL,
			is->buffer,
			strlen (is->buffer),
			0,
			0,
			(int *)&ivectab,
			32);
	if (matchrc < 0)
		return matchrc;
	/* Find a rule match in the group, if one exists */
	litem = is->matchgroup->matchrule_list;
	while (litem != NULL) {
		mr = slist_get_data (litem);
		matchrc = pcre_exec (mr->pcre_match,
				NULL,
				is->buffer,
				strlen (is->buffer),
				0,
				0,
				(int *)&ivectab,
				32);
		if (matchrc <= 0) {
			/* didn't match... next... */
			litem = slist_get_next (litem);
			continue;
		}
		/* it matched process it */
		mr->last_match_time = time (NULL);
		is->last_match = mr->last_match_time;
		mr->matches_total++;
		if (mr->ruleflags & RULE_FLAG_USE_INTERVAL) {
			td = mr->last_match_time - mr->intv_first_match;
			if (td > mr->intv_seconds) {
				/* match was outside interval, reset */
				mr->intv_first_match = mr->last_match_time;
				mr->intv_matches = 0;
				return 1; /* it matched, but no action */
			} else {
				mr->intv_matches++;
				if (mr->intv_matches < mr->intv_req_matches)
					return 1; /* it matched, no action */

				/* We met the interval match requirements, */
				/* reset before we call the action function */
				mr->intv_first_match = mr->last_match_time;
				mr->intv_matches = 0;
			}
		}
		md.subject = is->buffer;
		md.ivectab = ivectab;
		md.substrings = matchrc;
		mr->action_arg = is->matchgroup->action_arg;
		rc = (mr->action_function) (mr,	&md);
		return 1; /* a match was found and we took action */
	}
	return 0;
}


int inputsource_process (InputSource *is)
{
	int		rc;
	struct stat	st;
	if (is == NULL || is->fd == NULL)
		return EINVAL;
	while (!feof (is->fd)) {
		rc = inputsource_line_match (is);
		is->offset = ftello (is->fd);
	}
	rc = stat (is->path, &st);
	if (st.st_ino != is->inode) {
		/* log was rotated */
		rc = inputsource_close (is);
		rc = inputsource_open (is);
		return EROTATED;
	}
	if (st.st_size > is->offset)
		clearerr (is->fd);
	return EAGAIN;
}


