/****************************************************************/
/* netinfod - provide host information to monitor hosts		*/
/****************************************************************/
/* History:							*/
/* (02-10-2005) v0.8.3 Added code for v2 protocol stuff. Todd	*/
/* (02-27-2002) v0.8.2 Added code to allow IP address or	*/
/*		hostname in 'HOST' entry in config file. Todd	*/
/* (12-16-2001) v0.8.1 Added protocol negotiation code. Todd	*/
/* (11-27-2001) v0.8 Add OpenSSL support for encryption. Todd	*/
/* (03-06-2001) v0.7 Compile as a service under W32. Todd	*/
/* (09-07-2000) v0.5 Port to win32 complete. Todd Shadburn	*/
/* (08-18-2000) v0.4.1 Begin port to win32 platform. Added the	*/
/*		clientparms stuff to easy porting for different	*/
/*		threading structures. Todd			*/
/* (09-13-1999) v0.4 Added the commandline stuff.Todd		*/
/* (09-02-1999) v0.3 Fix the bug where the cli_socket was left	*/
/*		in a CLOSE_WAIT state after the conversation	*/
/*		was over. Added 'delete_lists()'. Todd Shadburn	*/
/* (08-31-1999) v0.2 Changes to config file to specify both	*/
/*		allowed hosts and allowed commands.Todd		*/
/* (08-24-1999) Orginal creation date. Todd Shadburn		*/
/****************************************************************/


#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
/*
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <pthread.h>
*/
#else
#include <wtypes.h>
#include <process.h>
#include <winsock2.h>
#include <winsvc.h>
#include <io.h>
#endif

#include <signal.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>

#include <cutil/cutil.h>
#include <comm/comm.h>

#include <netinfo/netinfo_portable.h>
#include <netinfo/netinfo.h>
#include <netinfo/nmodule_p.h>
#include <netinfo/conffile.h>
#include <netinfo/event.h>
#include <netinfo/event_p.h>
#include <netinfo/netbuffer.h>
#include <netinfo/rpc.h>


/****************************************************************/
/* Preprocessor definitions					*/
/****************************************************************/

#define TEMPFILE_NAME_SIZE	64
#define ARGUMENT_IDENTIFIER	"$ARGS$"

#define NETINFOD_LOG_ENTRY_SIZE	128

#define NETINFOD_FLAG_NODAEMON	0x00000001

#define NETINFOD_HEARTBEAT_INTERVAL	60

#ifndef WIN32
#define NETINFOD_CONFIG_FILE	"/netinfod.conf"
#define NETINFOD_LOG_FILE	"/netinfod.log"
#else
#define NETINFOD_CONFIG_FILE	"\\netinfod.conf"
#define NETINFOD_LOG_FILE	"\\netinfod.log"
#define	MSG_STARTED		"Netinfod has started."
#define	MSG_ENDED		"Netinfod has ended."
#define W32_SERVICE		1
#endif

#define IO_BUFFER_SIZE		4096

#define ADDED_EVENT_CONCAT_20050827
#ifdef ADDED_EVENT_CONCAT_20050827
#define NETINFO_EVENTCHAIN_TIMEOUT_SECONDS	120
#endif

#ifdef W32_SERVICE
#ifdef __GNUC__
#define RTMAIN			int
#define RETURNMAIN		return 0
#else
#define RTMAIN			VOID
#define RETURNMAIN		return
#endif
#endif


#ifndef WIN32
#define socket_t int
#define close_socket(s)	close(s)
#define open_file(f, a, m) open(f, a, m)
#define close_file(f) close(f)
#define INVALID_SOCKET -1
#define THREAD_EXIT(x)	exit(x)
#else
#define socket_t SOCKET
#define close_socket(s)	closesocket(s)
#define open_file(f, a, m) _open(f, a)
#define close_file(f) _close(f)
#define unlink(f) DeleteFile(f)
#define read(f, b, l) _read(f, b, l)
#define THREAD_EXIT(x)	ExitThread(x)
#define stat(x, y)	_stat(x, y)
#endif

/****************************************************************/
/* Structure Prototypes						*/
/****************************************************************/
typedef struct
	{
	socket_t		clientsocket;
	struct sockaddr_in	*clientsockaddr;
	int			sockaddrsize;
	/* added for protocol v2 */
	CommChannel		*cc;
	NetBuffer		*nb;
	int			rpcfd;
	u_int64_t		rpcfdoffset;
	} clientparms_t;


typedef struct
	{
	u_int32_t	serverflags;
	SList		*HostList;
	SList		*CommandList;
	SList		*cfgmodulelist;
	SList		*ModuleList;
	unsigned short	serverport;
	char		config_file[120];
	char		log_file[120];
	FILE		*lfd;
	pthread_mutex_t	log_mutex;
	char		*pubkeyfile;
	char		*privkeyfile;
	char		*pemphrase;
	ModuleVector	modvectors;
	char		*eventhost;
	unsigned short	eventport;
	CommModule	*cm;
	CommChannel	*ccs;
	Attribute	*commattribs;
#ifdef ADDED_EVENT_CONCAT_20050827
	CommChannel	*eventcc;
	u_int32_t	lasteventtime;
	u_int32_t	eventtimeout;
#endif
	u_int32_t	serverpid;
	CommGroup	*cg;
	u_int32_t	conncount;
	EventClient	*ec;
	} serverparms_t;
	

/****************************************************************/
/* Global variables						*/
/****************************************************************/
/*
char unknown_message[] = "I don't talk to strangers.\n";
char unknown_request[] = "I don't know that command.\n";
*/

serverparms_t		Server;

#ifdef W32_SERVICE
SERVICE_STATUS		ServiceStatus;
SERVICE_STATUS_HANDLE	ServiceStatusHandle;
#endif

#define BUFFER_SIZE		256

/****************************************************************/
/* Function Prototypes						*/
/****************************************************************/
void signal_handler(int sig);
void handle_client(clientparms_t *clientparms);
int access_file_load(char *filename);
void delete_lists(void);
int host_has_access(long hostname);
/*char * get_command_from_request(char * cli_request); */
int pem_passwd_cb(char *buf, int size, int rwflag, void *password);
int send_response_v2 (CommChannel *cc,
			u_int32_t status,
			u_int32_t flags,
			u_int32_t responserc,
			char *command,
			char *data,
			u_int32_t datalen);
int netinfod_log_entry (char *format, ...);
/* int netinfod_method_marshall (clientparms_t * cp, char *cmd, NetinfoHeader *nh); */
int netinfod_vector_log_entry (ModuleInfo *mi, char *format, va_list args);
int netinfod_vector_emit_event (ModuleInfo *mi, EventData *ed);
/* int netinfod_emit_event (EventData *ed); */
int netinfod_module_load (clientparms_t *cp, NetinfoHeader *nh);
int netinfod_module_unload (clientparms_t *cp, NetinfoHeader *nh);
int netinfod_vector_eventchain_open (ModuleInfo *mi, EventChain **ec);
int netinfod_vector_eventchain_close (EventChain **ec);
int netinfod_vector_eventchain_emit_event (EventChain *ec, EventData *ed);
#ifdef ADDED_EVENT_CONCAT_20050827
void *netinfod_housekeeper_thread (void * arg);
#endif
/* int netinfod_restart (clientparms_t *cp, NetinfoHeader *nh); */
int netinfod_restart (void);
int handle_request_rpcnb (clientparms_t *cp);
int handle_connection_io (CommChannel *cc, int iotype, void *ccarg);


#ifndef W32_SERVICE
void usage(void);
#endif

#ifdef W32_SERVICE
VOID NetinfodServiceStart(DWORD argc, LPTSTR *argv);
VOID NetinfodServiceCtrlHandler(DWORD opcode);

RTMAIN main(void)
	{
	char str[256];
	SERVICE_TABLE_ENTRY DispatchTable[] =
		{
			{"Netinfod",
			  (LPSERVICE_MAIN_FUNCTION)NetinfodServiceStart},
			{NULL, NULL}
		};
	if(!StartServiceCtrlDispatcher(DispatchTable))
		{
		sprintf(str,
			"Netinfod: StartServiceCtrlDispatcher error = %d\n",
			GetLastError());
		OutputDebugStringA(str);
		}
	RETURNMAIN;
	}
#endif


/****************************************************************/
/* main - Start of execution					*/
/****************************************************************/
#ifdef WIN32
#ifdef W32_SERVICE
void NetinfodServiceStart(DWORD argc, LPTSTR *argv)
#else
int main(DWORD argc, char * argv[])
#endif
#else
int main(int argc, char * argv[])
#endif
{
	int		rc;
#ifndef WIN32
	int		option;
	pid_t		pid;
#else
	DWORD		option;
	WSADATA		wsaData;
#endif

#ifdef W32_SERVICE
	char		str[256];
#endif
	char		*modpath = NULL;
	SList		*litem = NULL;
	NetinfoModule	*nm = NULL;
	EventData	*event = NULL;

	memset (&Server, 0, sizeof (Server));
	Server.serverflags = 0;
	Server.serverport = NETINFO_DEFAULT_PORT;
	Server.pubkeyfile = NULL;
	Server.privkeyfile = NULL;
	Server.modvectors.log_entry = netinfod_vector_log_entry;
	Server.modvectors.emit_event = netinfod_vector_emit_event;
	Server.modvectors.eventchain_open = netinfod_vector_eventchain_open;
	Server.modvectors.eventchain_close = netinfod_vector_eventchain_close;
	Server.modvectors.eventchain_emit_event =
				netinfod_vector_eventchain_emit_event;
	Server.eventport = NIEVENT_DEFAULT_PORT;
#ifdef ADDED_EVENT_CONCAT_20050827
	Server.eventtimeout = NETINFO_EVENTCHAIN_TIMEOUT_SECONDS;
#endif

#ifdef W32_SERVICE
	ServiceStatus.dwServiceType		= SERVICE_WIN32;
	ServiceStatus.dwCurrentState		= SERVICE_START_PENDING;
	ServiceStatus.dwControlsAccepted	= SERVICE_ACCEPT_STOP;
	ServiceStatus.dwWin32ExitCode		= 0;
	ServiceStatus.dwServiceSpecificExitCode = 0;
	ServiceStatus.dwCheckPoint		= 0;
	ServiceStatus.dwWaitHint		= 0;
	ServiceStatusHandle = RegisterServiceCtrlHandler("Netinfod",
			(LPHANDLER_FUNCTION)NetinfodServiceCtrlHandler);
	if(ServiceStatusHandle == 0) {
		sprintf(str,
			"Netinfod: RegisterServiceCtrlHandler error = %d\n",
			GetLastError());
		OutputDebugStringA(str);
	}
#endif

	strcpy(Server.config_file, SYSCONFDIR NETINFOD_CONFIG_FILE);
	strcpy(Server.log_file, LOGDIR NETINFOD_LOG_FILE);

#ifndef WIN32
	while (( option = getopt(argc, argv, "Df:hl:V")) != EOF) {
		switch(option) {
			case 'D':
				Server.serverflags |= NETINFOD_FLAG_NODAEMON;
				fprintf (stderr, "Not daemonizing.\n");
				break;
			case 'f':
				strcpy(Server.config_file, optarg);
				break;
			case 'l':
				strcpy(Server.log_file, optarg);
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'V':
				printf("\nnetinfod - version %s\n\n",
					VERSION);
				exit(0);
				break;
			default:
				usage();
				exit(-1);
				break;
		}
	}
#else
	option = 1;
	while (option < argc) {
		if(argv[option][0] == '-') {
			switch(argv[option][1]) {
				case 'f':
					if((option+1) < argc) {
						option++;
						strcpy(Server.config_file, argv[option]);
					}
					break;
				case 'l':
					if((option+1) < argc) {
						option++;
						strcpy(Server.log_file, argv[option]);
					}
					break;
#ifndef W32_SERVICE
				case 'h':
					usage();
					exit(0);
					break;
				case 'V':
					printf("\nnetinfod - version %s\n\n",
						VERSION);
					exit(0);
					break;
				default:
					usage();
					exit(-1);
					break;
#else
				default:
					exit(-1);
					break;
#endif
			}
		} else {
#ifndef W32_SERVICE
			usage();
#endif
			exit(-1);
		}
		option++;
	}
#endif

#ifndef WIN32
	if (!(Server.serverflags & NETINFOD_FLAG_NODAEMON)) {
		if ( (pid=fork()) < 0)
			exit(-1);
		else if(pid != 0)
			exit(0);
		chdir ("/");
		setsid ();
		umask (0);
		close (2);
		close (1);
		close (0);
		Server.serverpid = getpid();
	}
#endif

	rc = pthread_mutex_init (&Server.log_mutex, NULL);
	Server.lfd=fopen(Server.log_file, "a");
	if(!Server.lfd) {
		exit(-4);
	}
	netinfod_log_entry ("netinfod starting...\n");

#ifdef WIN32
	WSAStartup(MAKEWORD(2,0), &wsaData);
#endif
	rc = access_file_load(Server.config_file);
	if(rc) {
		netinfod_log_entry ("Error loading config file. rc=%d\n", rc);
		exit(-1);
	}

	rc = comm_plugin_load (&Server.cm,
			"comm_openssl",
			Server.commattribs);
	if(rc != 0) {
		netinfod_log_entry ("comm_plugin_load failed. rc=%d\n", rc);
		exit(-1);
	}

	litem = Server.cfgmodulelist;
	while (litem != NULL) {
		modpath = (char *)slist_get_data (litem);
		netinfod_log_entry ("Loading module '%s'\n", modpath);
		rc = netinfo_module_load (&nm,
					modpath,
					&Server.modvectors);
		if (rc != NETINFO_RC_SUCCESS) {
			netinfod_log_entry (
					"Failed to load module. rc=%d\n", rc);
		} else {
			slist_item_add(&Server.ModuleList, (void *)nm);
		}
		litem = slist_get_next (litem);
	}

	rc = comm_group_new(&Server.cg, handle_connection_io);
	if(rc != 0) {
		netinfod_log_entry ("comm_group_new failed. rc=%d\n", rc);
		rc = comm_plugin_unload (&Server.cm);
		exit(-1);
	}

	rc = comm_bind (Server.cm, &Server.ccs, 1, Server.serverport);
	if(rc != 0) {
		netinfod_log_entry ("comm_bind failed. rc=%d\n", rc);
		rc = comm_group_free (&Server.cg);
		rc = comm_plugin_unload (&Server.cm);
		exit(-1);
	}
	
	rc = comm_group_add (Server.cg, Server.ccs, NULL, COMM_IO_READ);
	if(rc != 0) {
		netinfod_log_entry ("comm_group_add failed. rc=%d\n", rc);
		rc = comm_close (&Server.ccs, 0);
		rc = comm_group_free (&Server.cg);
		rc = comm_plugin_unload (&Server.cm);
		exit(-1);
	}

#ifndef WIN32
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGHUP, signal_handler);
	signal(SIGPIPE, signal_handler);
	signal(SIGCHLD, SIG_IGN);
#endif
#ifdef W32_SERVICE
	ServiceStatus.dwCurrentState	= SERVICE_RUNNING;
	ServiceStatus.dwCheckPoint	= 0;
	ServiceStatus.dwWaitHint	= 0;
	if(!SetServiceStatus(ServiceStatusHandle, &ServiceStatus)) {
		sprintf(str, "Netinfod: SetServiceStatus error = %d\n",
				GetLastError());
		OutputDebugStringA(str);
	}
#endif


	rc = eventclient_init (&Server.ec, 0);
	if(rc != 0) {
		netinfod_log_entry ("eventclient_init failed. rc=%d\n", rc);
		rc = comm_close (&Server.ccs, 0);
		rc = comm_group_free (&Server.cg);
		rc = comm_plugin_unload (&Server.cm);
		exit(-1);
	}
	rc = eventclient_configure_heartbeat (Server.ec,
					"NETINFOD",
					NETINFOD_HEARTBEAT_INTERVAL);
	rc = eventclient_set_option (Server.ec,
					EC_OPTION_LOGGING_FUNCTION,
					netinfod_log_entry);
	rc = eventdata_new (&event);
	rc = eventdata_set (event,
			0,
			0,
			"SYSEVENT",
			"Netinfod agent started",
			strlen ("Netinfod agent started"));
/*	rc = netinfod_emit_event (event); */
	rc = eventclient_emit_event (Server.ec, event);
	rc = eventdata_free (&event);


	rc = comm_group_loop (Server.cg);

	rc = eventclient_uninit (&Server.ec);
	rc = comm_close (&Server.ccs, 0);
	rc = comm_plugin_unload (&Server.cm);
	/*rc = openssl_comm_cleanup (&Server.cmd);*/
	attribute_freeall (&Server.commattribs);
#ifdef WIN32
	WSACleanup();
#endif
	exit(0);
#ifdef W32_SERVICE
	return;
#else
	return(0);
#endif
}


#ifndef WIN32
void signal_handler(int sig)
{
	int rc;
	EventData *event = NULL;
	switch(sig) {
		case SIGPIPE:
			/*netinfod_log_entry ("SIGPIPE: Attempted write to an unconnected socket.\n"); */
			break;
		case SIGCHLD:
			wait(&rc);
			break;
		case SIGINT:
		case SIGQUIT:
		case SIGTERM:

			/*
			 * This is too much in the signal handler and is
			 * probably causing the hangs (__lll_lock_wait_private).
			 * It needs to be moved out! (unsafe, see CWE-364)...
			*/
			/* ... so cheat, for now. */
			exit (0);

			rc = comm_close (&Server.ccs, 0);
			rc = eventdata_new (&event);
			rc = eventdata_set (event,
					0,
					2,
					"SYSEVENT",
					"Netinfod agent ended",
					strlen ("Netinfod agent ended"));
			rc = eventclient_emit_event (Server.ec, event);
			/* rc = netinfod_emit_event (event); */
			rc = eventdata_free (&event);

			rc = comm_plugin_unload (&Server.cm);
			/*openssl_comm_cleanup (&Server.cmd);*/
			delete_lists();
			/*
			free(Server.pubkeyfile);
			free(Server.privkeyfile);
			*/
			netinfod_log_entry ("Server shutdown.\n");
			fflush(Server.lfd);
			fclose(Server.lfd);
			exit(0);
			break;
		case SIGHUP:
			/*
			 * Not complete and unsafe (CWE-364)
			netinfod_log_entry ("SIGHUP: Reloading Server.config file.\n");
			delete_lists();
			access_file_load(Server.config_file);
			*/
			break;
	}
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGHUP, signal_handler);
	signal(SIGPIPE, signal_handler);
	return;
}
#endif


void handle_client(clientparms_t *clientparms)
{
	int rc;
	int cli_size;
	long cli_ipaddress;
	char cli_addr[20];
	socket_t cli_socket;
	struct sockaddr_in * cli_sockaddr = NULL;
	if (clientparms == NULL || clientparms->cc == NULL) {
		free (clientparms->cc->cmodule);
		free (clientparms->cc);
		free (clientparms->clientsockaddr);
		free (clientparms);
		THREAD_EXIT(0);
		return ;
	}

	netinfod_log_entry ("Entered handle_client\n");
	cli_socket = clientparms->clientsocket;
	cli_size = clientparms->sockaddrsize;
	cli_sockaddr = clientparms->clientsockaddr;
	cli_ipaddress = htonl(cli_sockaddr->sin_addr.s_addr);
	sprintf(cli_addr, "%d.%d.%d.%d",
		(unsigned int)cli_ipaddress >> 24,
		(unsigned int)(cli_ipaddress << 8) >> 24,
		(unsigned int)(cli_ipaddress << 16) >> 24,
		(unsigned int)(cli_ipaddress << 24) >> 24);
	if(!host_has_access(cli_ipaddress)) {
		netinfod_log_entry ("Attempted Connection from: %s\n", cli_addr);
		rc = comm_close (&clientparms->cc, 0);
		free (clientparms->clientsockaddr);
		free (clientparms);
		THREAD_EXIT(0);
		return;
	}
	netinfod_log_entry ("Connection accepted from: %s\n", cli_addr);
	netinfod_log_entry ("Calling: handle_request_rpcnb\n");
	rc = handle_request_rpcnb (clientparms);
	if (rc != NETINFO_RC_SUCCESS)
		netinfod_log_entry ("handle_request_rpcnb(): rc=%d\n", rc);

	netinfod_log_entry ("Disconnecting.\n");
	rc = comm_close (&clientparms->cc, 0);
	free (clientparms->clientsockaddr);
	free (clientparms);
	THREAD_EXIT(0);
	return;
}


int access_file_load(char *filename)
{
	int rc;
	int count;
	int len;
	char line[BUFFER_SIZE];
	long host;
	ConfigFile *cf = NULL;
	struct hostent *hentry = NULL;
	char *command = NULL;
	char *directive = NULL;
	char *target = NULL;
	len=120;
	
	rc = configfile_open (&cf, filename);
	if(rc != NETINFO_RC_SUCCESS)
		return(NETINFO_RC_INVALIDARG);
	while(1) {
		rc = configfile_read_line (cf, line, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		if(line[0] == '#' || strlen (line) < 2)
			continue;
		directive = strtok (line, " \t\r\n");
		if (directive == NULL)
			continue;
		if(!strcmp(directive, "HOST")) {
			target = strtok (NULL, " \t\r\n");
			if (target == NULL)
				continue;
			hentry=gethostbyname(target);
			if(!hentry) {
				netinfod_log_entry ("HOST: %s: DNS lookup error.\n",
					target);
				continue;
			}
			count=0;
			while(hentry->h_addr_list[count] != NULL) {
				netinfod_log_entry ("HOST: %s -> %d.%d.%d.%d added.\n",
					hentry->h_name,
					(unsigned char)hentry->h_addr_list[count][0],
					(unsigned char)hentry->h_addr_list[count][1],
					(unsigned char)hentry->h_addr_list[count][2],
					(unsigned char)hentry->h_addr_list[count][3]);
				host=*((long *)hentry->h_addr_list[count]);
				slist_item_add (&Server.HostList, (void *)host);
				count++;
			}
		} else if(!strcmp(directive, "COMMAND")) {
			target = strtok (NULL, "\r\n");
			if (target == NULL)
				continue;
			command = strdup (target);
			slist_item_add (&Server.CommandList, (void *)command);
		} else if(!strcmp(directive, "PUBLICKEY")) {
			target = strtok (NULL, "\r\n");
			if (target == NULL)
				continue;
			command = strdup (target);
			Server.pubkeyfile = command;
			Server.commattribs = attribute_add (Server.commattribs,
							"SSLCertificateFile",
							command);
		} else if(!strcmp(directive, "PRIVATEKEY")) {
			target = strtok (NULL, "\r\n");
			if (target == NULL)
				continue;
			command = strdup (target);
			Server.privkeyfile = command;
			Server.commattribs = attribute_add (Server.commattribs,
							"SSLPrivateKeyFile",
							command);
		} else if(!strcmp(directive, "PEMPHRASE")) {
			target = strtok (NULL, "\r\n");
			if (target == NULL)
				continue;
			command = strdup (target);
			Server.pemphrase = command;
		} else if(!strcmp(directive, "PORT")) {
			target = strtok (NULL, " \t\r\n");
			if (target == NULL)
				continue;
			Server.serverport=atoi(target);
		} else if(!strcmp(directive, "MODULE")) {
			target = strdup (strtok (NULL, " \t\r\n"));
			if (target == NULL)
				continue;
			slist_item_add(&Server.cfgmodulelist, (void *)target);
/*
			netinfod_log_entry ("Loading module '%s'\n",
						target);
			rc = netinfo_module_load (&nm,
						target,
						&Server.modvectors);
			if (rc != NETINFO_RC_SUCCESS) {
				netinfod_log_entry (
					"Failed to load module. rc=%d\n",
					rc);
			} else {
				slist_item_add(&Server.ModuleList, (void *)nm);
			}
*/
		} else if(!strcmp(directive, "EVENTHOST")) {
			target = strdup (strtok (NULL, " \t\r\n"));
			if (target == NULL)
				continue;
			Server.eventhost = target;
		} else if(!strcmp(directive, "EVENTPORT")) {
			target = strtok (NULL, " \t\r\n");
			if (target == NULL)
				continue;
			Server.eventport = atoi (target);
		} else {
			fprintf(stderr, "Invalid directive: '%s'\n", directive);
			configfile_close (&cf);
			return(-2);
		}
	}
	configfile_close (&cf);
	return(0);
}


int host_has_access(long hostname)
{
	long host;
	long nohost;
	SList *litem = NULL;
	litem = Server.HostList;
	while (litem != NULL) {
		host=(long)slist_get_data (litem);
		nohost=htonl(hostname);
		if(nohost == host)
			return(1);
		litem = slist_get_next (litem);
	}
	return(0);
}

/*
char * get_command_from_request(char * cli_request)
{
	char *command = NULL;
	char *item = NULL;
	SList *litem = NULL;
	if (cli_request == NULL)
		return (NULL);
	litem = Server.CommandList;
	while (litem != NULL) {
		item = strdup(slist_get_data (litem));
		command = strtok (item, " \t");
		if (!strcmp (command, cli_request)) {
			command = strdup (strtok (NULL, "\r\n"));
			free (item);
			return command;
		}
		free(item);
		litem = slist_get_next (litem);
	}
	return(NULL);
}
*/

void delete_lists(void)
{
	int rc;
	void *p = NULL;
	SList *litem = NULL;
	NetinfoModule *nm = NULL;
	netinfod_log_entry ("Freeing module, host & command lists.\n");
	/* HostList is now a list of longs not char*... */
	/* dont try to free the items, just the list */
	slist_free (&Server.HostList);
	litem = Server.CommandList;
	while (litem != NULL) {
		p = slist_get_data (litem);
		free (p);
		litem = slist_get_next (litem);
	}
	slist_free (&Server.CommandList);
	litem = Server.ModuleList;
	while (litem != NULL) {
		nm = (NetinfoModule *)slist_get_data (litem);
		rc = netinfo_module_unload (&nm);
		litem = slist_get_next (litem);
	}
	slist_free (&Server.ModuleList);
	return;
}


int pem_passwd_cb(char *buf, int size, int rwflag, void *password)
{
	int len;
	len = strlen(Server.pemphrase);
	if(len > size)
		len = size - 1;
	strncpy(buf, Server.pemphrase, len);
	buf[len] = 0x00;
	return(len);
}


#ifndef W32_SERVICE
void usage(void)
{
	printf("\nnetinfod - Information Server.n");
	printf(" Options:\n");
#ifndef WIN32
	printf("   -D                   Netinfod will not daemonize itself.\n");
#endif
	printf("   -f configfile	Specifies the path to the ");
	printf("configuration file to use.\n");
	printf("   -h			This help screen.\n");
	printf("   -l logfile		Specifies the path to the ");
	printf("log file to use.\n");
	printf("   -V			Display version information ");
	printf("and exit.\n\n");
	return;
}
#endif

#ifdef W32_SERVICE
void NetinfodServiceCtrlHandler(DWORD opcode)
{
	CHAR str[256];
	switch(opcode) {
		case SERVICE_CONTROL_STOP:
			ServiceStatus.dwWin32ExitCode	= 0;
			ServiceStatus.dwCurrentState	= SERVICE_STOPPED;
			ServiceStatus.dwCheckPoint	= 0;
			ServiceStatus.dwWaitHint	= 0;
			break;
	}
	if(!SetServiceStatus(ServiceStatusHandle, &ServiceStatus)) {
		sprintf(str, "Netinfod: SetServiceStatus error = %lu\n",
				GetLastError());
		OutputDebugStringA(str);
	}
	return;
}
#endif


int send_response_v2 (CommChannel *cc,
			u_int32_t status,
			u_int32_t flags,
			u_int32_t responserc,
			char *command,
			char *data,
			u_int32_t datalen)
{
	int rc;
	int bytes;
	int count;
	NetinfoHeader nh;
	char *pdata = NULL;
	if (cc == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = netinfo_header_clear (&nh);
	rc = netinfo_header_set_status (&nh, status);
	rc = netinfo_header_set_rc (&nh, responserc);
	if (command != NULL)
		rc = netinfo_header_set_command (&nh, command);
	if (data != NULL && datalen > 0)
		rc = netinfo_header_set_datalen (&nh, datalen);
	rc = netinfo_header_to_networkorder (&nh);
	bytes = sizeof (NetinfoHeader);
	rc = comm_write(cc, (char*)&nh, &bytes);
	if (data != NULL && datalen > 0) {
		bytes = datalen;
		pdata = data;
		while (bytes > 0) {
			count = (bytes > 4096 ? 4096 : bytes);
			rc = comm_write(cc,
					pdata,
					&count);
			bytes -= count;
			pdata += count;
		}
	}
	return NETINFO_RC_SUCCESS;
}


int netinfod_log_entry (char *format, ...)
{
	va_list args;
	int rc;
	time_t t;
	struct tm ltr;
	struct tm *lt;
	char *str = NULL;
	if (format == NULL || Server.lfd == NULL)
		return NETINFO_RC_INVALIDARG;
	str = calloc (1, NETINFOD_LOG_ENTRY_SIZE);
	if (str == NULL)
		return NETINFO_RC_MEMORYERROR;
	va_start (args, format);
	t = time (NULL);
	lt = localtime_r (&t, &ltr);
	fprintf (Server.lfd,
		"%04d-%02d-%02d %02d:%02d:%02d ",
		ltr.tm_year+1900,
		ltr.tm_mon+1,
		ltr.tm_mday,
		ltr.tm_hour,
		ltr.tm_min,
		ltr.tm_sec);
	rc = vsnprintf (str,
			NETINFOD_LOG_ENTRY_SIZE,
			format,
			args);
	pthread_mutex_lock (&Server.log_mutex);
	fprintf (Server.lfd, str);
	pthread_mutex_unlock (&Server.log_mutex);
	fflush (Server.lfd);
	free (str);
	va_end (args);
	return NETINFO_RC_SUCCESS;
}

/*
int netinfod_method_marshall (clientparms_t * cp, char *cmd, NetinfoHeader *nh)
{
	int		rc;
	int		count;
	u_int32_t	bytes;
	u_int32_t	responsedatalen;
	char		*responsedata = NULL;
	char		*p = NULL;
	char		*modname = NULL;
	char		*modmethod = NULL;
	SList		*litem = NULL;
	NetinfoModule	*nm = NULL;
	if (cp == NULL || cmd == NULL || nh == NULL)
		return NETINFO_RC_INVALIDARG;
	if (strstr (cmd, ".") == NULL)
		return NETINFO_RC_NOTIMPLEMENTED;
	p = strdup (cmd);
	modname = strtok (p, ".");
	modmethod = strtok (NULL, " ");
	if (modname == NULL || modmethod == NULL) {
		free (p);
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	litem = Server.ModuleList;
	while (litem != NULL) {
		nm = (NetinfoModule *)slist_get_data (litem);
		if (!strcmp (netinfo_module_get_name (nm), modname))
			break;
		litem = slist_get_next (litem);
	}
	if (nm == NULL) {
		netinfod_log_entry ("Module '%s' not loaded.\n", modname);
		free (p);
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	netinfod_log_entry ("Calling method '%s' in module '%s'\n",
				modmethod,
				modname);
	rc = netinfo_module_call_method (nm,
					modmethod,
					nh,
					&responsedata,
					&responsedatalen);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfod_log_entry ("netinfo_module_call_method(): rc=%d\n",
				rc);
		free (p);
		return rc;
	}
	free (p);
	if (responsedata == NULL)
		responsedatalen = 0;
	rc = netinfo_header_set_datalen (nh, responsedatalen);
	rc = netinfo_header_set_command (nh, cmd);
	rc = netinfo_header_to_networkorder (nh);
	count = sizeof (NetinfoHeader);
	rc = comm_write(cp->cc, (char*)nh, &count);
	if (responsedata != NULL) {
		bytes = responsedatalen;
		p = responsedata;
		while (bytes > 0) {
			count = (bytes > 4096 ? 4096 : bytes);
			rc = comm_write(cp->cc, p, &count);
			if (rc != 0) {
				netinfod_log_entry ("ERROR: comm_write() failed."
						    " %d bytes remaining. rc=%d\n",
						    bytes,
						    rc);
				free (responsedata);
				return NETINFO_RC_SSLERROR;
				break;
			}
			bytes -= count;
			p += count;
		}
		free (responsedata);
	}
	return NETINFO_RC_SUCCESS;
}
*/

int netinfod_vector_log_entry (ModuleInfo *mi, char *format, va_list args)
{
	int rc;
	time_t t;
	struct tm ltr;
	struct tm *lt;
	char *str = NULL;
	if (mi == NULL || format == NULL || Server.lfd == NULL)
		return NETINFO_RC_INVALIDARG;
	str = calloc (1, NETINFOD_LOG_ENTRY_SIZE);
	if (str == NULL)
		return NETINFO_RC_MEMORYERROR;
	t = time (NULL);
	lt = localtime_r (&t, &ltr);
	fprintf (Server.lfd,
		"%04d-%02d-%02d %02d:%02d:%02d (%s) ",
		ltr.tm_year+1900,
		ltr.tm_mon+1,
		ltr.tm_mday,
		ltr.tm_hour,
		ltr.tm_min,
		ltr.tm_sec,
		mi->name);
	rc = vsnprintf (str,
			NETINFOD_LOG_ENTRY_SIZE,
			format,
			args);
	pthread_mutex_lock (&Server.log_mutex);
	fprintf (Server.lfd, str);
	pthread_mutex_unlock (&Server.log_mutex);
	fflush (Server.lfd);
	free (str);
	return NETINFO_RC_SUCCESS;
}

/* replaced by the same function below (eventclient based) */
/*
int netinfod_vector_emit_event (ModuleInfo *mi, EventData *ed)
{
	int rc;
	int bytes;
	CommChannel *cc = NULL;
	if (mi == NULL || ed == NULL)
		return NETINFO_RC_INVALIDARG;
	if (Server.eventhost == NULL) {
		rc = netinfod_log_entry (
			"emit_event: No event trap host configured.\n");
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	rc = comm_connect (Server.cm,
			&cc,
			Server.eventhost,
			Server.eventport);
	if (rc != 0) {
		rc = netinfod_log_entry (
			"emit_event: Failed to connect to trap host. rc=%d\n", rc);
		return rc;
	}
	rc = eventdata_to_networkorder (ed);
	rc = eventdata_sizeof (ed, &bytes);
	rc = comm_write (cc, (char *)ed, &bytes);
	rc = comm_close (&cc, 0);
	return NETINFO_RC_SUCCESS;
}
*/

/*
int netinfod_emit_event (EventData *ed)
{
	int rc;
	int bytes;
	CommChannel *cc = NULL;
	if (ed == NULL)
		return NETINFO_RC_INVALIDARG;
	if (Server.eventhost == NULL) {
		rc = netinfod_log_entry (
			"emit_event: No event trap host configured.\n");
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	rc = comm_connect (Server.cm,
			&cc,
			Server.eventhost,
			Server.eventport);
	if (rc != 0) {
		rc = netinfod_log_entry (
			"emit_event: Failed to connect to trap host. rc=%d\n", rc);
		return rc;
	}
	rc = eventdata_to_networkorder (ed);
	rc = eventdata_sizeof (ed, &bytes);
	rc = comm_write (cc, (char *)ed, &bytes);
	rc = comm_close (&cc, 0);
	return NETINFO_RC_SUCCESS;
}
*/

int netinfod_module_load (clientparms_t *cp, NetinfoHeader *nh)

{
	int		rc;
	int		lrc;
	int		bytes;
	char		filepath[NETINFO_HEADER_ARGDATA_SIZE];
	NetinfoModule	*nm = NULL;
	if (cp == NULL || nh == NULL)
		return NETINFO_RC_INVALIDARG;
	memset (filepath, 0, NETINFO_HEADER_ARGDATA_SIZE);
	bytes = NETINFO_HEADER_ARGDATA_SIZE;
	rc = netinfo_header_get_argdata (nh, filepath, &bytes);
	netinfod_log_entry ("module.load: path=%s\n", filepath);
	
	lrc = netinfo_module_load (&nm,
				filepath,
				&Server.modvectors);
	if (lrc != NETINFO_RC_SUCCESS) {
		netinfod_log_entry (
			"Failed to load module. rc=%d\n",
			lrc);
	} else {
		slist_item_add(&Server.ModuleList, (void *)nm);
	}
	rc = send_response_v2 (cp->cc, 
			NETINFO_STATUS_RESPONSE,
			0,
			lrc,
			"module.load",
			NULL,
			0);
	return NETINFO_RC_SUCCESS;
}


int netinfod_module_unload (clientparms_t *cp, NetinfoHeader *nh)
{
	int		rc;
	int		lrc;
	int		bytes;
	char		name[NETINFO_HEADER_ARGDATA_SIZE];
	NetinfoModule	*nm = NULL;
	SList		*litem = NULL;
	if (cp == NULL || nh == NULL)
		return NETINFO_RC_INVALIDARG;
	memset (name, 0, NETINFO_HEADER_ARGDATA_SIZE);
	bytes = NETINFO_HEADER_ARGDATA_SIZE;
	rc = netinfo_header_get_argdata (nh, name, &bytes);
	netinfod_log_entry ("module.unload: name=%s\n", name);
	
	litem = Server.ModuleList;
	while (litem != NULL) {
		nm = slist_get_data (litem);
		if (nm != NULL && nm->moduleinfo.name != NULL &&
			!strcmp (nm->moduleinfo.name, name)) {
			break;
		}
		litem = slist_get_next (litem);
	}
	if (litem == NULL) {
		rc = send_response_v2 (cp->cc, 
				NETINFO_STATUS_RESPONSE |
					NETINFO_STATUS_ARGINVALID,
				0,
				NETINFO_RC_INVALIDARG,
				"module.unload",
				NULL,
				0);
		return NETINFO_RC_INVALIDARG;
	}
	lrc = netinfo_module_unload (&nm);
	if (lrc != NETINFO_RC_SUCCESS) {
		netinfod_log_entry (
			"Failed to unload module. rc=%d\n",
			lrc);
	} else {
		slist_item_remove(&Server.ModuleList, (void *)nm);
	}
	rc = send_response_v2 (cp->cc, 
			NETINFO_STATUS_RESPONSE,
			0,
			lrc,
			"module.unload",
			NULL,
			0);
	return NETINFO_RC_SUCCESS;
}


/* New for eventclient */

int netinfod_vector_eventchain_open (ModuleInfo *mi, EventChain **ec)
{
	EventChain	*c = NULL;
	if (mi == NULL || ec == NULL)
		return NETINFO_RC_INVALIDARG;
	/* START just fake it for now, need to alter lib/netinfo_p.c */
	c = calloc (1, sizeof (EventChain));
	if (c == NULL)
		return NETINFO_RC_MEMORYERROR;
	c->moduleinfo = mi;
	*ec = c;
	/* END just fake it for now, need to alter lib/netinfo_p.c */
	return NETINFO_RC_SUCCESS;
}


int netinfod_vector_eventchain_close (EventChain **ec)
{
	if (ec == NULL || *ec == NULL)
		return NETINFO_RC_INVALIDARG;
	free (*ec);
	*ec = NULL;
	return NETINFO_RC_SUCCESS;
}


int netinfod_vector_eventchain_emit_event (EventChain *ec, EventData *ed)
{
	int		rc;
	if (ec == NULL || ed == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = eventclient_emit_event (Server.ec, ed);
	return rc;
}


int netinfod_vector_emit_event (ModuleInfo *mi, EventData *ed)
{
	int		rc;
	if (mi == NULL || ed == NULL)
		return NETINFO_RC_INVALIDARG;
	rc = eventclient_emit_event (Server.ec, ed);
	return rc;
}


/* Old for eventchain */
/*
int netinfod_vector_eventchain_open (ModuleInfo *mi, EventChain **ec)
{
	int		rc;
	CommChannel	*cc = NULL;
	EventChain	*c = NULL;
	if (mi == NULL || ec == NULL)
		return NETINFO_RC_INVALIDARG;
	*ec = NULL;
	if (Server.eventhost == NULL) {
		rc = netinfod_log_entry (
			"eventchain_open: No event trap host configured.\n");
		return NETINFO_RC_NOTIMPLEMENTED;
	}
	c = calloc (1, sizeof (EventChain));
	if (c == NULL)
		return NETINFO_RC_MEMORYERROR;
#ifdef ADDED_EVENT_CONCAT_20050827
	if (Server.eventcc != NULL) {
		c->commchannel = Server.eventcc;
		c->moduleinfo = mi;
		*ec = c;
		return NETINFO_RC_SUCCESS;
	}
#endif
	rc = comm_connect (Server.cm,
			&cc,
			Server.eventhost,
			Server.eventport);
	if (rc != 0) {
		free (c);
		return rc;
	}
	c->commchannel = cc;
	c->moduleinfo = mi;
#ifdef ADDED_EVENT_CONCAT_20050827
	Server.eventcc = cc;
#endif
	*ec = c;
	return NETINFO_RC_SUCCESS;
}


int netinfod_vector_eventchain_close (EventChain **ec)
{
	int		rc;
	EventChain	*c = NULL;
	CommChannel	*cc = NULL;
	if (ec == NULL || *ec == NULL)
		return NETINFO_RC_INVALIDARG;
	c = *ec;
	if (c->commchannel == NULL || c->moduleinfo == NULL)
		return NETINFO_RC_INVALIDARG;
#ifdef ADDED_EVENT_CONCAT_20050827
	if ( (time (NULL) - Server.lasteventtime) < Server.eventtimeout) {
		free (c);
		*ec = NULL;
		return NETINFO_RC_SUCCESS;
	}
#endif
	cc = c->commchannel;
#ifdef ADDED_EVENT_CONCAT_20050827
	Server.eventcc = NULL;
#endif
	rc = comm_close (&cc, 0);
	free (c);
	*ec = NULL;	
	return NETINFO_RC_SUCCESS;
}


int netinfod_vector_eventchain_emit_event (EventChain *ec, EventData *ed)
{
	int		rc;
	int		bytes;
	CommChannel	*cc = NULL;
	if (ec == NULL || ed == NULL)
		return NETINFO_RC_INVALIDARG;
	if (ec->commchannel == NULL || ec->moduleinfo == NULL)
		return NETINFO_RC_INVALIDARG;
	cc = ec->commchannel;
	rc = eventdata_to_networkorder (ed);
	rc = eventdata_sizeof (ed, &bytes);
	rc = comm_write (cc, (char *)ed, &bytes);
	if (rc != NETINFO_RC_SUCCESS)
		netinfod_log_entry (
			"eventchain_emit_event: Error sending event. rc=%d\n", rc);
#ifdef ADDED_EVENT_CONCAT_20050827
	Server.lasteventtime = time(NULL);
#endif
	return rc;
}
*/

#ifdef ADDED_EVENT_CONCAT_20050827
void *netinfod_housekeeper_thread (void * arg)
{
	int		seconds;
	while (1) {
		sleep (Server.eventtimeout);
		seconds = time (NULL) - Server.lasteventtime;
		if (seconds > Server.eventtimeout) {
			/* Close the event CommChannel */
		}
	}
#ifndef WIN32
	pthread_exit (NULL);
#else
	THREAD_EXIT(0);
#endif
	return (NULL);
}
#endif


#ifndef WIN32
int netinfod_restart (void)
{
	int	rc;
	int	trc;
	pid_t	pid;
	netinfod_log_entry (
		"Restarting netinfod...\n");
	if ( (pid=fork()) < 0)
		return errno;
	else if (pid != 0) {
		pid = getpid();
		rc = kill (Server.serverpid, SIGTERM);
		return NETINFO_RC_SUCCESS;
	}
	netinfod_log_entry (
		"(restart pid=%d) waiting for netinfod(pid=%d) to exit...\n",
		getpid(),
		Server.serverpid);
	rc = waitpid (Server.serverpid, &trc, 0);
	chdir ("/");
	setsid ();
	close (2);
	close (1);
	close (0);
	delete_lists();
	netinfod_log_entry (
		"(restart) restarting server...\n", getpid());
	fflush(Server.lfd);
	fclose(Server.lfd);
	sleep (2);
	rc = system (SBINDIR "/netinfod");
	exit (0);
	return rc;
}
#else
int netinfod_restart (void)
{
	/* Not supported on win32 platform */
	return 0;
}
#endif


int handle_request_rpcnb (clientparms_t *cp)
{
	int			rc;
	int			nfd;
	int			access;
	int			bytes;
	u_int32_t		iobytes;
	unsigned char		buffer[IO_BUFFER_SIZE];
	NetinfoRpcHeader	nrh;
	NetinfoRpcFileOpen	nrfo;
	NetinfoRpcFileClose	nrfc;
	NetinfoRpcFileData	nrfd;
	if (cp == NULL || cp->cc == NULL || cp->nb == NULL)
		return NETINFO_RC_INVALIDARG;

	access = 0;

/*
	while (1) {
*/
		netbuffer_clear (cp->nb);
		bytes = NET_BUFFER_SIZE;
		rc = comm_read (cp->cc, cp->nb->buffer, &bytes);
		if (rc != 0)
			goto handle_request_rpcnb_error;
		cp->nb->woffset = bytes;
		rc = netbuffer_read (cp->nb, (char*)&nrh, sizeof (NetinfoRpcHeader));
		if (rc != 0)
			goto handle_request_rpcnb_error;
		netinfo_rpcheader_ntoh (&nrh);
		switch (nrh.callid) {
			case NETINFO_RPC_CALLID_FILEOPEN:
				rc = netbuffer_read (cp->nb,
						(char*)&nrfo,
						sizeof (NetinfoRpcFileOpen));
				if (rc != 0 || cp->rpcfd != -1) {
					nrh.returncode = NETINFO_RC_PROTOCOLERROR;
					netinfo_rpcheader_hton (&nrh);
					bytes = sizeof (NetinfoRpcHeader);
					rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
					break;
				}
				netinfo_rpcfileopen_ntoh (&nrfo);
				access = 0;
				if (nrfo.access & NETINFO_RPC_READ)
					access += O_RDONLY;
				if (nrfo.access & NETINFO_RPC_WRITE)
					access += O_WRONLY | O_CREAT;
				if ((access & NETINFO_RPC_READWRITE) == NETINFO_RPC_READWRITE)
					access = O_RDWR | O_CREAT;
				if (access & O_CREAT)
					nfd = open (nrfo.filepath, access, 0640);
				else
					nfd = open (nrfo.filepath, access);
				if (nfd == -1) {
					nrh.returncode = errno;
					netinfo_rpcheader_hton (&nrh);
					bytes = sizeof (NetinfoRpcHeader);
					rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
					break;
				}
				cp->rpcfd = nfd;
				cp->rpcfdoffset = 0;
				netbuffer_clear (cp->nb);
				nrh.returncode = 0;
				netinfo_rpcheader_hton (&nrh);
				rc = netbuffer_write (cp->nb,
						(char*)&nrh,
						sizeof (NetinfoRpcHeader));
				nrfo.fileid = nfd;
				netinfo_rpcfileopen_hton (&nrfo);
				rc = netbuffer_write (cp->nb,
						(char*)&nrfo,
						sizeof (NetinfoRpcFileOpen));
				bytes = cp->nb->woffset;
				rc = comm_write (cp->cc,
						(char*)cp->nb->buffer,
						&bytes);
				break;
			case NETINFO_RPC_CALLID_FILECLOSE:
				rc = netbuffer_read (cp->nb,
						(char*)&nrfc,
						sizeof (NetinfoRpcFileClose));
				if (rc != 0 || cp->rpcfd == -1) {
					nrh.returncode = NETINFO_RC_PROTOCOLERROR;
					netinfo_rpcheader_hton (&nrh);
					bytes = sizeof (NetinfoRpcHeader);
					rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
					break;
				}
				netinfo_rpcfileclose_ntoh (&nrfc);
				if (nrfc.fileid != cp->rpcfd) {
					nrh.returncode = EBADF;
					netinfo_rpcheader_hton (&nrh);
					bytes = sizeof (NetinfoRpcHeader);
					rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
					break;
				}
				rc = close (cp->rpcfd);
				cp->rpcfd = -1;
				cp->rpcfdoffset = 0;
				nrh.returncode = rc;
				netinfo_rpcheader_hton (&nrh);
				bytes = sizeof (NetinfoRpcHeader);
				rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
				break;
			case NETINFO_RPC_CALLID_FILEDATA:
				rc = netbuffer_read (cp->nb,
						(char*)&nrfd,
						sizeof (NetinfoRpcFileData));
				if (rc != 0 || cp->rpcfd == -1) {
					nrh.returncode = NETINFO_RC_PROTOCOLERROR;
					netinfo_rpcheader_hton (&nrh);
					bytes = sizeof (NetinfoRpcHeader);
					rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
					break;
				}
				netinfo_rpcfiledata_ntoh (&nrfd);
				if (nrfd.fileid != cp->rpcfd) {
					nrh.returncode = EBADF;
					netinfo_rpcheader_hton (&nrh);
					bytes = sizeof (NetinfoRpcHeader);
					rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
					break;
				}
				if (nrfd.offset != cp->rpcfdoffset) {
					rc = lseek (cp->rpcfd, nrfd.offset, SEEK_SET);
					if (rc == -1) {
						nrh.returncode = errno;
						netinfo_rpcheader_hton (&nrh);
						bytes = sizeof (NetinfoRpcHeader);
						rc = comm_write (cp->cc,
								(char*)&nrh,
								&bytes);
						break;
					}
					cp->rpcfdoffset = nrfd.offset;
				}
				if (nrfd.direction == NETINFO_RPC_READ) {
					if (access == O_WRONLY) {
						nrh.returncode = EBADF;
						netinfo_rpcheader_hton (&nrh);
						bytes = sizeof (NetinfoRpcHeader);
						rc = comm_write (cp->cc,
								(char*)&nrh,
								&bytes);
						break;
					}
					iobytes = read (cp->rpcfd, buffer, IO_BUFFER_SIZE);
					if (iobytes < 1) {
						if (iobytes == 0)
							nrh.returncode = ENODATA;
						else
							nrh.returncode = errno;
						netinfo_rpcheader_hton (&nrh);
						bytes = sizeof (NetinfoRpcHeader);
						rc = comm_write (cp->cc,
								(char*)&nrh,
								&bytes);
						break;
					}
					cp->rpcfdoffset += iobytes;
					netbuffer_clear (cp->nb);
					nrh.returncode = 0;
					netinfo_rpcheader_hton (&nrh);
					rc = netbuffer_write (cp->nb,
							(char*)&nrh,
							sizeof (NetinfoRpcHeader));
					nrfd.length = iobytes;
					netinfo_rpcfiledata_hton (&nrfd);
					rc = netbuffer_write (cp->nb,
							(char*)&nrfd,
							sizeof (NetinfoRpcFileData));
					rc = netbuffer_write (cp->nb,
							buffer,
							iobytes);
					bytes = cp->nb->woffset;
					rc = comm_write (cp->cc,
							(char*)cp->nb->buffer,
							&bytes);
				} else if (nrfd.direction == NETINFO_RPC_WRITE) {
					if (access == O_RDONLY) {
						nrh.returncode = EBADF;
						netinfo_rpcheader_hton (&nrh);
						bytes = sizeof (NetinfoRpcHeader);
						rc = comm_write (cp->cc,
								(char*)&nrh,
								&bytes);
						break;
					}
					iobytes = nrfd.length;
					rc = netbuffer_read (cp->nb, buffer, iobytes);
					if (rc != 0) {
						nrh.returncode = NETINFO_RC_PROTOCOLERROR;
						netinfo_rpcheader_hton (&nrh);
						bytes = sizeof (NetinfoRpcHeader);
						rc = comm_write (cp->cc,
								(char*)&nrh,
								&bytes);
						break;
					}
					iobytes = write (cp->rpcfd, buffer, iobytes);
					if (iobytes == -1) {
						nrh.returncode = errno;
						netinfo_rpcheader_hton (&nrh);
						bytes = sizeof (NetinfoRpcHeader);
						rc = comm_write (cp->cc,
								(char*)&nrh,
								&bytes);
						break;
					}
					netbuffer_clear (cp->nb);
					nrh.returncode = 0;
					netinfo_rpcheader_hton (&nrh);
					rc = netbuffer_write (cp->nb,
							(char*)&nrh,
							sizeof (NetinfoRpcHeader));
					nrfd.length = iobytes;
					netinfo_rpcfiledata_hton (&nrfd);
					rc = netbuffer_write (cp->nb,
							(char*)&nrfd,
							sizeof (NetinfoRpcFileData));
					bytes = cp->nb->woffset;
					rc = comm_write (cp->cc,
							(char*)cp->nb->buffer,
							&bytes);
				}
				break;
			case NETINFO_RPC_CALLID_RESTART:
				nrh.returncode = NETINFO_RC_SUCCESS;
				netinfo_rpcheader_hton (&nrh);
				bytes = sizeof (NetinfoRpcHeader);
				rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
				rc = netinfod_restart ();
				break;
			default:
netinfod_log_entry ("NETINFO_RPC_PROTOCOLERROR: nrh.callid = %lu\n", nrh.callid);
				nrh.returncode = NETINFO_RC_PROTOCOLERROR;
				netinfo_rpcheader_hton (&nrh);
				bytes = sizeof (NetinfoRpcHeader);
				rc = comm_write (cp->cc,
							(char*)&nrh,
							&bytes);
				break;
		}
/*
	}
*/
	return NETINFO_RC_SUCCESS;

handle_request_rpcnb_error:
	return rc;
}



int handle_connection_io (CommChannel *cc, int iotype, void *ccarg)
{
        int                     rc;
        CommChannel             *ccc = NULL;
	clientparms_t		*clientparms = NULL;
        if (cc == NULL)
		return EINVAL;
        // If it's the server socket, accept the new connection
        if (cc == Server.ccs) {
                rc = comm_accept (cc, &ccc, 0);
                if (rc != 0) {
                        netinfod_log_entry (
                                "comm_accept(): failed rc=%d\n", rc);
                        return rc;
                }
		// allocate client structures
		clientparms = calloc (1, sizeof (clientparms_t));
		clientparms->clientsockaddr = calloc (1,
				sizeof(struct sockaddr_in));
		rc = comm_get_fd (ccc, &clientparms->clientsocket);
		clientparms->sockaddrsize = sizeof (struct sockaddr_in);
		rc = getpeername (clientparms->clientsocket,
				(struct sockaddr *)clientparms->clientsockaddr,
				&clientparms->sockaddrsize);
		clientparms->cc = ccc;
		clientparms->rpcfd = -1;
		rc = netbuffer_new (&clientparms->nb);
		// Validate the client
		// Hook the client connection
                rc = comm_group_add (Server.cg, ccc, clientparms, COMM_IO_READ);
                if (rc != 0) {
                        comm_close (&ccc, 0);
                        netinfod_log_entry (
                                "comm_group_add(): failed rc=%d\n", rc);
                        return rc;
                }
                Server.conncount++;
                if ( (Server.conncount % 100) == 0) {
                        netinfod_log_entry (
                                "Handled %lu connections so far.\n",
                                Server.conncount);
                }
                return rc;
        } else {
	        // It's a client socket, handle the request
		clientparms = (clientparms_t *)ccarg;
		if (clientparms == NULL) {
			netinfod_log_entry ("%s: ccarg is NULL!\n", __PRETTY_FUNCTION__);
			return EINVAL;
		}
		rc = handle_request_rpcnb (clientparms);
		if (rc != NETINFO_RC_SUCCESS) {
			if (rc != ENODATA && rc != EAGAIN) {
				netinfod_log_entry (
					"handle_request_rpcnb(): rc=%d\n",
					rc);
			}
			if (rc == EAGAIN || rc == 2)
				return 0;
			if (clientparms->rpcfd != -1) {
				netinfod_log_entry (
					"Client left file open...closing it.\n");
				close (clientparms->rpcfd);
			}
			clientparms->rpcfd = -1;
			comm_group_remove (Server.cg, clientparms->cc);
			comm_close (&clientparms->cc, 0);
/*
			if (clientparms->clientsockaddr != NULL)
				free (clientparms->clientsockaddr);
			rc = netbuffer_free (&clientparms->nb);
			free (clientparms);
*/
			return rc;
		}
		return rc;
	}
        return 0;
}



