/****************************************************************/
/* netinfoevtd 							*/
/****************************************************************/
/* History:                                                     */
/* (03-08-2005) Orginal creation date. Todd Shadburn            */
/****************************************************************/


#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <pthread.h>
#else
#include <windows.h>
#include <winsock2.h>
#include <winsvc.h>
#include <io.h>
#endif

#include <signal.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <cutil/cutil.h>
#include <comm/comm.h>

#include <netinfo/netinfo.h>
#include <netinfo/event.h>
#include <netinfo/event_p.h>
#include <netinfo/conffile.h>


#ifndef WIN32
#include <dlfcn.h>
#define MODULEHANDLE            void*
#define	SOCKET_CLEANUP()
#else
#define MODULEHANDLE            HMODULE
#define RTLD_NOW                1
#define dlopen(x,y)             LoadLibrary (x)
#define dlsym(x,y)              GetProcAddress (x, y)
#define dlclose(x)              FreeLibrary (x)
#define	SOCKET_CLEANUP()	WSACleanup()
#endif


/* Enable Debugging */
/*
#define DEBUG	1
*/

#ifndef WIN32
#define NETINFOEVTD_CONFIG_FILE    "/netinfoevtd.conf"
#define NETINFOEVTD_LOG_FILE       "/netinfoevtd.log"
#else
#define NETINFOEVTD_CONFIG_FILE    "\\netinfoevtd.conf"
#define NETINFOEVTD_LOG_FILE       "\\netinfoevtd.log"
#endif

#define	PATH_SIZE			256
#define BUFFER_SIZE			256
#define	STRING_SIZE			64
#define CONFIG_DELIMITERS		" \t\r\n"
#define COMM_LOOP_TIMEOUT		60000000

#define	FLAG_FLUSH_LOG_ENTRIES		1
#define	NETINFOEVTD_FLAG_NODAEMON	2


#ifdef	CONFIG_EVENTBUCKETS
#define	EVENTBUCKET_SIZE_DEFAULT	10
#define	EVENTBUCKET_COUNT_DEFAULT	100
#define	EVENTBUCKET_APPEND_COUNT	10
#define	EVENTBUCKET_FLAG_INUSE		1
#define	EVENTBUCKET_FLAG_READONLY	2
#else
#define	QUEUE_SIZE_DEFAULT		10
#define	QUEUE_COUNT_DEFAULT		10
#define	CONN_BUFFER_SIZE		4096
#endif


/* Prototypes for functions which must be exported by NIEH plugins */ 
typedef int (*EventHandlerInit)(char *args);
typedef int (*EventHandlerCleanup)(void);
typedef int (*EventHandler)(EventData *event);


typedef struct _NIEHModule	NIEHModule;
struct _NIEHModule {
	char			*modulepath;
	char			*moduleargs;
	MODULEHANDLE		module;
	EventHandlerInit	module_init;
	EventHandler		handle_event;
	EventHandlerCleanup	module_cleanup;
	pthread_mutex_t		mutex;
	unsigned long		usecondssum;
	unsigned long		eventsprocessed;
};


typedef struct	_ConnBuffer	ConnBuffer;
struct	_ConnBuffer {
	int		len;
	char		buffer[CONN_BUFFER_SIZE];
};


#ifdef	CONFIG_EVENTBUCKETS
typedef struct	_EventBucket	EventBucket;
struct	_EventBucket {
	unsigned int	flags;
	unsigned int	bucketsize;
	unsigned int	events;
	EventData	**eventarray;
};

typedef	struct	_EventBucketStats	EventBucketStats;
struct	_EventBucketStats {
	int		totalbuckets;
	int		inusebuckets;
	int		totalevents;
	int		inuseevents;
	double		fillaverage;
};
#endif


/* Global Variable */
struct	{
	CommModule	*cm;
	Attribute	*modattrib;
	CommChannel	*ccs;
	CommGroup	*cg;
	pthread_mutex_t	modmutex;
	SList		*modlist;
	int		serverport;
	unsigned long	flags;
	unsigned long	conncount;
	char		config_file[PATH_SIZE];
	char		log_file[PATH_SIZE];

	pthread_attr_t	threadattr;
	pthread_mutex_t	procmutex;
	pthread_cond_t	proccond;
	pthread_t	evtproc_threadid;
	int		eventsize;
	pthread_mutex_t	logmutex;
	char		logbuffer[BUFFER_SIZE];
	FILE		*logfile;
#ifdef	CONFIG_EVENTBUCKETS
	int		bucketsize;
	int		bucketcount;
	int		bucketappend;
	pthread_mutex_t	bucketmutex;
	SList		*bucketlist;
#else
	int		queuesize;
	int		queuecount;
	pthread_mutex_t	inqmutex;
	SList		*inqlist;
	SQueue		*inqcurrent;
	pthread_mutex_t	outqmutex;
	SList		*outqlist;
#endif
} EvtServer;


/* Function Prototypes */
void signal_handler(int sig);
int netinfoevtd_log_entry (char *format, ...);
int handle_connection_io (CommChannel *cc, int iotype, void *ccarg);
int netinfoevtd_process_events (void);
void *netinfoevtd_event_processing_thread (void *arg);
#ifdef	CONFIG_EVENTBUCKETS
int netinfoevtd_eventbuckets_alloc (int count, int size);
EventBucket *eventbucket_new (int size);
EventBucket *netinfoevtd_eventbucket_grab (void);
int netinfoevtd_eventbucket_release(EventBucket *eb);
int netinfoevtd_eventbucket_stats (EventBucketStats *ebs);
#else
int netinfoevtd_queues_allocate (int qsize, int qcount);
int netinfoevtd_queues_deallocate (void);
int netinfoevtd_queues_statistics (void);
#endif
void usage(void);


int main (int argc, char **argv)
{
	int			rc;
/*	pthread_t		threadid;	*/

#ifndef WIN32
	pid_t			pid;
	int			option;
#else
	WSADATA			wsaData;
	DWORD			option;
#endif
	char			buffer[BUFFER_SIZE];
	char			*directive = NULL;
	EventData		*event = NULL;
	ConfigFile		*cf = NULL;
	NIEHModule		*nm = NULL;
	SList			*listitem = NULL;

	memset (&EvtServer, 0, sizeof (EvtServer));
	strncpy (EvtServer.config_file, SYSCONFDIR NETINFOEVTD_CONFIG_FILE, PATH_SIZE);
	EvtServer.config_file[PATH_SIZE-1] = 0;
	strncpy (EvtServer.log_file, LOGDIR NETINFOEVTD_LOG_FILE, PATH_SIZE);
	EvtServer.log_file[PATH_SIZE-1] = 0;
	EvtServer.serverport = NIEVENT_DEFAULT_PORT;
#ifdef	CONFIG_EVENTBUCKETS
	EvtServer.bucketsize = EVENTBUCKET_SIZE_DEFAULT;
	EvtServer.bucketcount = EVENTBUCKET_COUNT_DEFAULT;
	EvtServer.bucketappend = EVENTBUCKET_APPEND_COUNT;
#else
	EvtServer.queuesize = QUEUE_SIZE_DEFAULT;
	EvtServer.queuecount = QUEUE_COUNT_DEFAULT;
	EvtServer.eventsize = sizeof (EventData);
#endif

/* Parse commandline arguments */
#ifndef WIN32
	while (( option = getopt(argc, argv, "Df:hl:V")) != EOF) {
		switch(option) {
			case 'D':
				EvtServer.flags |= NETINFOEVTD_FLAG_NODAEMON;
				fprintf (stderr, "Not daemonizing.\n");
				break;
			case 'f':
				strncpy(EvtServer.config_file, optarg, PATH_SIZE);
				EvtServer.config_file[PATH_SIZE-1] = 0;
				break;
			case 'l':
				strncpy(EvtServer.log_file, optarg, PATH_SIZE);
				EvtServer.log_file[PATH_SIZE-1] = 0;
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'V':
				printf("\nnetinfoevtd - version %s\n\n",
					VERSION);
				exit(0);
				break;
			default:
				usage();
				exit(-1);
				break;
		}
	}
#else
	option = 1;
	while (option < argc) {
		if(argv[option][0] == '-') {
			switch(argv[option][1]) {
				case 'f':
					if((option+1) < argc) {
						option++;
						strncpy(EvtServer.config_file, argv[option], PATH_SIZE);
						EvtServer.config_file[PATH_SIZE] = 0;
					}
					break;
				case 'l':
					if((option+1) < argc) {
						option++;
						strncpy(EvtServer.log_file, argv[option], PATH_SIZE);
						EvtServer.log_file[PATH_SIZE] = 0;
					}
					break;
#ifndef W32_SERVICE
				case 'h':
					usage();
					exit(0);
					break;
				case 'V':
					printf("\nnetinfoevtd - version %s\n\n",
						VERSION);
					exit(0);
					break;
				default:
					usage();
					exit(-1);
					break;
#else
				default:
					exit(-1);
					break;
#endif
			}
		} else {
#ifndef W32_SERVICE
			usage();
#endif
			exit(-1);
		}
		option++;
	}
#endif


	EvtServer.logfile = fopen (EvtServer.log_file, "a");
	if (EvtServer.logfile == NULL) {
		fprintf (stderr, "Failed to open logfile\n");
		return 1;
	}
	netinfoevtd_log_entry ("netinfoevtd - Version %s\n", VERSION);
	netinfoevtd_log_entry ("Starting system...\n");
	pthread_attr_init (&EvtServer.threadattr);
	pthread_attr_setdetachstate (&EvtServer.threadattr,
					PTHREAD_CREATE_DETACHED);
	pthread_cond_init (&EvtServer.proccond, NULL);
	pthread_mutex_init (&EvtServer.logmutex, NULL);
	pthread_mutex_init (&EvtServer.procmutex, NULL);
	pthread_mutex_init (&EvtServer.modmutex, NULL);
#ifdef	CONFIG_EVENTBUCKETS
	pthread_mutex_init (&EvtServer.bucketmutex, NULL);
#else
	pthread_mutex_init (&EvtServer.inqmutex, NULL);
	pthread_mutex_init (&EvtServer.outqmutex, NULL);
#endif

	rc = comm_group_new (&EvtServer.cg, handle_connection_io);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfoevtd_log_entry (
			"Failed to create comm group.\n");
		fclose (EvtServer.logfile);
		return 2;
	}
	
	rc = configfile_open (&cf, EvtServer.config_file);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfoevtd_log_entry (
			"Failed to open configuration file.\n");
		fclose (EvtServer.logfile);
		return 2;
	}

	rc = configfile_search_begin (cf, "PUBLICKEY");
	rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfoevtd_log_entry (
			"Failed to find PUBLICKEY directive.\n");
		rc = configfile_close (&cf);
		fclose (EvtServer.logfile);
		return 2;
	}
	directive = strtok (buffer, CONFIG_DELIMITERS);
	EvtServer.modattrib = attribute_add (EvtServer.modattrib,
				"SSLCertificateFile",
				strtok (NULL, CONFIG_DELIMITERS));
	rc = configfile_search_end (cf);

	rc = configfile_search_begin (cf, "PRIVATEKEY");
	rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfoevtd_log_entry (
			"Failed to find PRIVATEKEY directive.\n");
		rc = configfile_close (&cf);
		attribute_freeall (&EvtServer.modattrib);
		fclose (EvtServer.logfile);
		return 2;
	}
	directive = strtok (buffer, CONFIG_DELIMITERS);
	EvtServer.modattrib = attribute_add (EvtServer.modattrib,
				"SSLPrivateKeyFile",
				strtok (NULL, CONFIG_DELIMITERS));
	rc = configfile_search_end (cf);

	rc = configfile_search_begin (cf, "EVENTPORT");
	rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
	if (rc == NETINFO_RC_SUCCESS) {
		directive = strtok (buffer, CONFIG_DELIMITERS);
		directive = strtok (NULL, CONFIG_DELIMITERS);
		if (directive != NULL) {
			EvtServer.serverport = atoi (directive);
			netinfoevtd_log_entry (
				"Listener will be setup on port %d\n",
				EvtServer.serverport);
		}
	}
	rc = configfile_search_end (cf);

	rc = configfile_search_begin (cf, "QUEUESIZE");
	rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
	if (rc == NETINFO_RC_SUCCESS) {
		directive = strtok (buffer, CONFIG_DELIMITERS);
		directive = strtok (NULL, CONFIG_DELIMITERS);
		if (directive != NULL) {
			EvtServer.queuesize = atoi (directive);
			netinfoevtd_log_entry (
				"Queue size set to %d\n",
				EvtServer.serverport);
		}
	}
	rc = configfile_search_end (cf);

	rc = configfile_search_begin (cf, "FLUSHLOG");
	rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
	if (rc == NETINFO_RC_SUCCESS) {
		directive = strtok (buffer, CONFIG_DELIMITERS);
		directive = strtok (NULL, CONFIG_DELIMITERS);
		if (directive != NULL) {
			if (directive[0] == 'Y' || directive[0] == 'y') {
				EvtServer.flags |= FLAG_FLUSH_LOG_ENTRIES;
				netinfoevtd_log_entry (
					"Log entries will be flushed.\n");
			}
		}
	}
	rc = configfile_search_end (cf);

	rc = configfile_search_begin (cf, "MODULE");
	while (1) {
		rc = configfile_search_next (cf, buffer, BUFFER_SIZE);
		if (rc != NETINFO_RC_SUCCESS)
			break;
		if (buffer[0] == '#')
			continue;
		nm = calloc (1, sizeof (NIEHModule));
		directive = strtok (buffer, CONFIG_DELIMITERS);
		directive = strtok (NULL, CONFIG_DELIMITERS);
		if (directive == NULL) {
			free (nm);
			netinfoevtd_log_entry (
				"No module path specified.\n");
			continue;
		}
		nm->modulepath = strdup(directive);
		directive = strtok (NULL, CONFIG_DELIMITERS);
		if (directive != NULL)
			nm->moduleargs = strdup (directive);
		netinfoevtd_log_entry (
			"Loading module: %s\n",
			nm->modulepath);
		nm->module = dlopen (nm->modulepath, RTLD_NOW);
		if (nm->module == NULL) {
			if (nm->moduleargs != NULL)
				free (nm->moduleargs);
			free (nm->modulepath);
			free (nm);
			netinfoevtd_log_entry (
				"Failed to open module.\n");
			continue;
		}
		nm->module_init = dlsym (nm->module, "module_init");
		nm->module_cleanup = dlsym (nm->module, "module_cleanup");
		nm->handle_event = dlsym (nm->module, "handle_event");
		if (nm->module_init == NULL ||
		    nm->module_cleanup == NULL ||
		    nm->handle_event == NULL) {
			dlclose (nm->module);
			if (nm->moduleargs != NULL)
				free (nm->moduleargs);
			free (nm->modulepath);
			free (nm);
			netinfoevtd_log_entry (
				"Failed to locate module symbols.\n");
			continue;
		}
		rc = (*nm->module_init) (nm->moduleargs);
		if (rc != NETINFO_RC_SUCCESS) {
			dlclose (nm->module);
			if (nm->moduleargs != NULL)
				free (nm->moduleargs);
			free (nm->modulepath);
			free (nm);
			netinfoevtd_log_entry (
				"Calling module_init() failed.\n");
			continue;
		}
		pthread_mutex_init (&nm->mutex, NULL);
		pthread_mutex_lock (&EvtServer.modmutex);
		rc = slist_item_add (&EvtServer.modlist, nm);
		pthread_mutex_unlock (&EvtServer.modmutex);
		netinfoevtd_log_entry (
			"Module loaded successfully.\n");
	}
	rc = configfile_search_end (cf);

	rc = configfile_close (&cf);

#ifndef WIN32
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGHUP, signal_handler);
	signal(SIGPIPE, signal_handler);
	signal(SIGCHLD, SIG_IGN);

	if (!(EvtServer.flags & NETINFOEVTD_FLAG_NODAEMON)) {
		if ( (pid=fork()) < 0)
			exit(-1);
		else if(pid != 0)
			exit(0);
		chdir ("/");
		setsid ();
		umask (0);
		close (2);
		close (1);
		close (0);
	}

#endif



#ifdef WIN32
	WSAStartup(MAKEWORD(2,0), &wsaData);
#endif
	rc = eventdata_new (&event);
	rc = comm_plugin_load (&EvtServer.cm, "comm_openssl", EvtServer.modattrib);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfoevtd_log_entry (
			"Failed to initialize comm layer.\n");
		eventdata_free (&event);
		fclose (EvtServer.logfile);
		SOCKET_CLEANUP();
		return rc;
	}

#ifdef	CONFIG_EVENTBUCKETS
	rc = netinfoevtd_eventbuckets_alloc(EvtServer.bucketcount,
					EvtServer.bucketsize);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfoevtd_log_entry (
			"Failed to allocate event buckets.\n");
		rc = comm_plugin_unload (&EvtServer.cm);
		eventdata_free (&event);
		netinfoevtd_log_entry ("Exiting.\n");
		fclose (EvtServer.logfile);
		SOCKET_CLEANUP();
		return rc;
	}
	netinfoevtd_log_entry (
		"Successfully allocated %d event buckets, of size %d.\n",
		EvtServer.bucketcount,
		EvtServer.bucketsize);
#else
	rc = netinfoevtd_queues_allocate(EvtServer.queuesize,
					EvtServer.queuecount);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfoevtd_log_entry (
			"Failed to allocate event queues.\n");
		rc = comm_plugin_unload (&EvtServer.cm);
		eventdata_free (&event);
		netinfoevtd_log_entry ("Exiting.\n");
		fclose (EvtServer.logfile);
		SOCKET_CLEANUP();
		return rc;
	}
	netinfoevtd_log_entry (
		"Successfully allocated %d event queues, of size %d.\n",
		EvtServer.queuecount,
		EvtServer.queuesize);
#endif	

	rc = comm_bind (EvtServer.cm, &EvtServer.ccs, NULL, EvtServer.serverport);
	if (rc != NETINFO_RC_SUCCESS) {
		netinfoevtd_log_entry (
			"Failed to bind to port.\n");
		rc = comm_plugin_unload (&EvtServer.cm);
		eventdata_free (&event);
		netinfoevtd_log_entry ("Exiting.\n");
		fclose (EvtServer.logfile);
		SOCKET_CLEANUP();
		return rc;
	}

	rc = comm_group_add (EvtServer.cg, EvtServer.ccs, NULL, COMM_IO_READ);
	if (rc != 0) {
		netinfoevtd_log_entry (
				"Failed in comm_group_add(). rc=%d\n",
				rc);
		rc = comm_close (&EvtServer.ccs, 0);
		rc = comm_plugin_unload (&EvtServer.cm);
		eventdata_free (&event);
		netinfoevtd_log_entry ("Exiting.\n");
		fclose (EvtServer.logfile);
		SOCKET_CLEANUP();
		return rc;
	}

	rc = comm_group_set_option (EvtServer.cg, COMM_OPTION_TIMEOUT, COMM_LOOP_TIMEOUT);
	if (rc != 0) {
		netinfoevtd_log_entry (
				"Failed in comm_group_set_option(). rc=%d\n",
				rc);
		rc = comm_group_remove (EvtServer.cg, EvtServer.ccs);
		rc = comm_close (&EvtServer.ccs, 0);
		rc = comm_plugin_unload (&EvtServer.cm);
		eventdata_free (&event);
		netinfoevtd_log_entry ("Exiting.\n");
		fclose (EvtServer.logfile);
		SOCKET_CLEANUP();
		return rc;
	}

	rc = pthread_create (&EvtServer.evtproc_threadid,
				&EvtServer.threadattr,
				netinfoevtd_event_processing_thread,
				NULL);
	if (rc != 0) {
		netinfoevtd_log_entry (
			"Failed to spawn event processing thread. rc=%d\n",
			rc);
		rc = comm_group_remove (EvtServer.cg, EvtServer.ccs);
		rc = comm_close (&EvtServer.ccs, 0);
		rc = comm_plugin_unload (&EvtServer.cm);
		eventdata_free (&event);
		netinfoevtd_log_entry ("Exiting.\n");
		fclose (EvtServer.logfile);
		SOCKET_CLEANUP();
		return rc;
	}
	netinfoevtd_log_entry ("Successfully spawned event handler thread.\n");

	netinfoevtd_log_entry ("System ready.\n");

	/* Grab a queue off the inqlist, before entering the loop */
	pthread_mutex_lock (&EvtServer.inqmutex);
	EvtServer.inqcurrent = slist_get_data (EvtServer.inqlist);
	rc = slist_item_remove (&EvtServer.inqlist, EvtServer.inqcurrent);
	pthread_mutex_unlock (&EvtServer.inqmutex);

	rc = comm_group_loop (EvtServer.cg);

	rc = comm_group_remove (EvtServer.cg, EvtServer.ccs);
	rc = comm_close (&EvtServer.ccs, 0);
	rc = comm_plugin_unload (&EvtServer.cm);
	rc = eventdata_free (&event);
	pthread_mutex_lock (&EvtServer.modmutex);
	listitem = EvtServer.modlist;
	while (listitem != NULL) {
		nm = slist_get_data (listitem);
		if (nm != NULL) {
			rc = (*nm->module_cleanup) ();
			dlclose (nm->module);
			if (nm->moduleargs != NULL)
				free (nm->moduleargs);
			free (nm->modulepath);
			free (nm);
		}
		listitem = slist_get_next (listitem);
	}
	pthread_mutex_unlock (&EvtServer.modmutex);
	SOCKET_CLEANUP();
	return 0;
}


#ifndef WIN32
void signal_handler(int sig)
{
	int rc;
	netinfoevtd_log_entry ("Caught signal %d\n", sig);
	switch(sig) {
		case SIGCHLD:
			wait(&rc);
			break;
/*
		case SIGPIPE:
			fprintf (EvtServer.logfile,
				"Writing to a pipe without a reader!\n");
			break;
*/
		case SIGINT:
		case SIGQUIT:
		case SIGTERM:
			rc = comm_close (&EvtServer.ccs, 0);
			rc = comm_plugin_unload (&EvtServer.cm);
			netinfoevtd_log_entry ("Server shutdown.\n");
			fclose(EvtServer.logfile);
			exit(0);
			break;
	}
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGHUP, signal_handler);
	signal(SIGPIPE, signal_handler);
	return;
}
#endif


int netinfoevtd_log_entry (char *format, ...)
{
	va_list		args;
	int		rc;
	time_t		t;
	struct tm	*lt = NULL;
	if (format == NULL || EvtServer.logfile == NULL)
		return NETINFO_RC_INVALIDARG;
	/* Depricated by EvtServer.logbuffer
	str = calloc (1, BUFFER_SIZE);
	if (str == NULL)
		return NETINFO_RC_MEMORYERROR;
	*/
	va_start (args, format);
	t = time (NULL);
	lt = localtime (&t);
	pthread_mutex_lock (&EvtServer.logmutex);
        fprintf (EvtServer.logfile,
		"%04d-%02d-%02d %02d:%02d:%02d ",
		lt->tm_year+1900,
		lt->tm_mon+1,
		lt->tm_mday,
		lt->tm_hour,
		lt->tm_min,
		lt->tm_sec);
	rc = vsnprintf (EvtServer.logbuffer,
			BUFFER_SIZE,
			format,
			args);
	fprintf (EvtServer.logfile, EvtServer.logbuffer);
	pthread_mutex_unlock (&EvtServer.logmutex);
	if (EvtServer.flags & FLAG_FLUSH_LOG_ENTRIES)
		fflush (EvtServer.logfile);
	/* Depricated by EvtServer.logbuffer
	free (str);
	*/
	va_end (args);
	return NETINFO_RC_SUCCESS;
}


#ifdef	CONFIG_EVENTBUCKETS
int handle_connection_io (CommChannel *cc, int iotype, void *ccarg)
{
	int			rc;
	int			i;
	int			bytes;
	int			eventsize;
#ifdef DEBUG
	int			len;
	char			hostname[256];
	char			object[256];
#endif
	EventBucketStats	ebs;
	EventBucket		*eb = NULL;
	CommChannel		*ccc = NULL;
	if (cc == NULL)
		return EINVAL;
	/* If it's the server socket, accept the new connection */
	if (cc == EvtServer.ccs) {
		rc = comm_accept (cc, &ccc, 0);
		if (rc != 0) {
			netinfoevtd_log_entry (
				"comm_accept(): failed rc=%d\n", rc);
			return rc;
		}
		rc = comm_group_add (EvtServer.cg, ccc, NULL, COMM_IO_READ);
		if (rc != 0) {
			comm_close (&ccc, 0);
			netinfoevtd_log_entry (
				"comm_group_add(): failed rc=%d\n", rc);
			return rc;
		}
		EvtServer.conncount++;
		if ( (EvtServer.conncount % 100) == 0) {
			netinfoevtd_log_entry (
				"Handled %lu connections so far.\n",
				EvtServer.conncount);
			rc = netinfoevtd_eventbucket_stats (&ebs);
			netinfoevtd_log_entry (
				"Buckets: %d/%d in use, fillavg = %0.4f%%\n",
				ebs.inusebuckets,
				ebs.totalbuckets,
				ebs.fillaverage);
		}
		return rc;
	}

	eb = netinfoevtd_eventbucket_grab ();
	if (eb == NULL) {
		netinfoevtd_log_entry (
			"netinfoevtd_eventbucket_grab(): failed\n");
		rc = netinfoevtd_eventbuckets_alloc(EvtServer.bucketappend,
						EvtServer.bucketsize);
		if (rc != NETINFO_RC_SUCCESS) {
			netinfoevtd_log_entry (
				"Failed to allocate the requested event buckets.\n");
			return rc;
		}
		netinfoevtd_log_entry (
			"Successfully allocated %d event buckets, of size %d.\n",
			EvtServer.bucketappend,
			EvtServer.bucketsize);
		eb = netinfoevtd_eventbucket_grab ();
		if (eb == NULL) {
			netinfoevtd_log_entry (
				"netinfoevtd_eventbucket_grab(): failed (after bucket allocation)\n");
			return EINVAL;
		}
	}
	rc = eventdata_sizeof (eb->eventarray[0], &eventsize);
	for (i=0 ; i<eb->bucketsize ; i++) {
#ifdef DEBUG
		netinfoevtd_log_entry ("%s: Reading event...\n", __PRETTY_FUNCTION__);
#endif
		rc = eventdata_clear (eb->eventarray[i]);
		bytes = eventsize;
		rc = comm_read (cc,
				(char *)eb->eventarray[i],
				&bytes);
#ifdef DEBUG
		netinfoevtd_log_entry ("openssl_comm_read(): rc=%d bytes=%d\n",
			rc, bytes);
#endif
		if (rc != NETINFO_RC_SUCCESS) {
			if (rc != ENODATA && rc != EAGAIN) {
				netinfoevtd_log_entry (
					"openssl_comm_read(): rc=%d\n",
					rc);
			}
			if (eb->events > 0) {
				eb->flags |= EVENTBUCKET_FLAG_READONLY;
				pthread_cond_signal (&EvtServer.proccond);
			} else {
				netinfoevtd_eventbucket_release (eb);
			}
			if (rc == EAGAIN || rc == 2)
				return 0;
			comm_group_remove (EvtServer.cg, cc);
			comm_close (&cc, 0);
			return rc;
		}
		eventdata_to_hostorder (eb->eventarray[i]);
		eb->events++;
#ifdef DEBUG
		len = STRING_SIZE;
		rc = eventdata_get_hostname (eb->eventarray[i],
				hostname,
				&len);
		len = STRING_SIZE;
		rc = eventdata_get_object (eb->eventarray[i],
				object,
				&len);
		netinfoevtd_log_entry ("Event: hostname=%s  object=%s\n", hostname, object);
#endif
	}
	eb->flags |= EVENTBUCKET_FLAG_READONLY;
	pthread_cond_signal (&EvtServer.proccond);
	return 0;
}

#else

int handle_connection_io (CommChannel *cc, int iotype, void *ccarg)
{
	int			rc;
	int			bytes;
	int			ccflags;
#ifdef DEBUG
	int			len;
	char			hostname[256];
	char			object[256];
#endif
	CommChannel		*ccc = NULL;
	ConnBuffer		*connbuffer = NULL;
	EventData		*event = NULL;
	if (cc == NULL)
		return EINVAL;
	/* If it's the server socket, accept the new connection */
	if (cc == EvtServer.ccs) {
		rc = comm_accept (cc, &ccc, 0);
		if (rc != 0 && rc != COMM_RC_WANT_READ && rc != COMM_RC_WANT_WRITE) {
			netinfoevtd_log_entry (
				"comm_accept(): failed rc=%d\n", rc);
			return rc;
		}
		connbuffer = calloc (1, sizeof (ConnBuffer));
		if (connbuffer == NULL) {
			comm_close (&ccc, 0);
			rc = ENOMEM;
			netinfoevtd_log_entry (
				"calloc(): failed to allocate connbuffer rc=%d\n", rc);
			return rc;
		}
		if (rc == COMM_RC_WANT_WRITE)
			rc = comm_set_ioneeds (cc, COMM_IO_WRITE);
		rc = comm_group_add (EvtServer.cg, ccc, connbuffer, COMM_IO_READ);
		if (rc != 0) {
			free (connbuffer);
			comm_close (&ccc, 0);
			netinfoevtd_log_entry (
				"comm_group_add(): failed rc=%d\n", rc);
			return rc;
		}
		EvtServer.conncount++;
		if ( (EvtServer.conncount % 100) == 0) {
			netinfoevtd_log_entry (
				"Handled %lu connections so far.\n",
				EvtServer.conncount);
		}
		return rc;
	}

	if (iotype == COMM_IO_WRITE)
		rc = comm_set_ioneeds (cc, 0);
	
	rc = comm_get_flags (cc, &ccflags);
	if (ccflags & COMMFLAG_INACCEPT) {
		rc = comm_accept_continue (cc);
		return rc;
	}

#ifdef DEBUG
	netinfoevtd_log_entry ("%s: Reading event data...\n", __PRETTY_FUNCTION__);
#endif
	connbuffer = ccarg;
	bytes = EvtServer.eventsize - connbuffer->len;
	rc = comm_read (cc,
			(char *)connbuffer->buffer,
			&bytes);
#ifdef DEBUG
	netinfoevtd_log_entry ("openssl_comm_read(): rc=%d bytes=%d\n",
				rc, bytes);
#endif
	if (rc != NETINFO_RC_SUCCESS) {
		switch (rc) {
			case COMM_RC_WANT_READ:
			case EAGAIN:
				return 0;
				break;
			case COMM_RC_WANT_WRITE:
				rc = comm_set_ioneeds (cc, COMM_IO_WRITE);
				return 0;
				break;
			default:
				if (rc != ENODATA) {
					netinfoevtd_log_entry (
						"openssl_comm_read(): rc=%d\n",
						rc);
				}
				comm_group_remove (EvtServer.cg, cc);
				free (connbuffer);
				comm_close (&cc, 0);
				return rc;
		}
	}
	connbuffer->len += bytes;
	if (connbuffer->len == EvtServer.eventsize) {
		event = (EventData *)connbuffer->buffer;
		eventdata_to_hostorder (event);
		rc = squeue_item_enqueue (EvtServer.inqcurrent, event);
		if (rc != 0) {
#ifdef DEBUG
			netinfoevtd_log_entry (
				"DEBUG: inqueue full, moving to outqueue...\n");
#endif
			/* Add the full queue to the outqlist */
			pthread_mutex_lock (&EvtServer.outqmutex);
			rc = slist_item_add (&EvtServer.outqlist, EvtServer.inqcurrent);
			EvtServer.inqcurrent = NULL;
			pthread_mutex_unlock (&EvtServer.outqmutex);
			pthread_cond_signal (&EvtServer.proccond);
			/* Grab a queue off the inqlist */
			pthread_mutex_lock (&EvtServer.inqmutex);
			if (EvtServer.inqlist != NULL) {
				EvtServer.inqcurrent = slist_get_data (EvtServer.inqlist);
				rc = slist_item_remove (&EvtServer.inqlist, EvtServer.inqcurrent);
				pthread_mutex_unlock (&EvtServer.inqmutex);
			} else {
				pthread_mutex_unlock (&EvtServer.inqmutex);
				netinfoevtd_log_entry (
					"Allocating a new inqueue!\n");
				rc = squeue_new (&EvtServer.inqcurrent,
						EvtServer.eventsize,
						EvtServer.queuesize);
				if (rc != 0) {
					netinfoevtd_log_entry (
						"Failed to allocate a new inqueue! (rc=%d)\n", rc);
				}
			}
			/* Attempt the enqueue again */
			rc = squeue_item_enqueue (EvtServer.inqcurrent, event);
			if (rc != 0) {
				netinfoevtd_log_entry (
					"Enqueue failed after queue switch (rc=%d)!\n", rc);
			}
		}

		connbuffer->len = 0;
#ifdef DEBUG
		len = STRING_SIZE;
		rc = eventdata_get_hostname (event, hostname, &len);
		len = STRING_SIZE;
		rc = eventdata_get_object (event, object, &len);
		netinfoevtd_log_entry ("Event: hostname=%s  object=%s\n", hostname, object);
#endif
	}
	return 0;
}
#endif


#ifdef	CONFIG_EVENTBUCKETS
int netinfoevtd_eventbuckets_alloc (int count, int size)
{
	int		rc;
	int		i;
	EventBucket	*eb = NULL;
	if (count < EVENTBUCKET_COUNT_DEFAULT)
		count = EVENTBUCKET_COUNT_DEFAULT;
	if (size < EVENTBUCKET_SIZE_DEFAULT)
		size = EVENTBUCKET_SIZE_DEFAULT;
	for (i=0 ; i<count ; i++) {
		eb = eventbucket_new (size);
		if (eb != NULL) {
			pthread_mutex_lock (&EvtServer.bucketmutex);
			rc = slist_item_add (&EvtServer.bucketlist, eb);
			pthread_mutex_unlock (&EvtServer.bucketmutex);
		}
	}
	return 0;
}


EventBucket *eventbucket_new (int size)
{
	int		rc;
	int		x;
	EventBucket	*eb = NULL;
	eb = calloc (1, sizeof (EventBucket));
	if (eb == NULL)
		return NULL;
	eb->flags = 0;
	eb->bucketsize = size;
	eb->events = 0;
	eb->eventarray = calloc (size, sizeof (void *));
	if (eb->eventarray == NULL) {
		free (eb);
		return NULL;
	}
	for (x=0 ; x<size ; x++) {
		rc = eventdata_new (&eb->eventarray[x]);
	}
	return eb;
}


EventBucket *netinfoevtd_eventbucket_grab (void)
{
	SList		*litem = NULL;
	EventBucket	*eb = NULL;
	pthread_mutex_lock (&EvtServer.bucketmutex);
	litem = EvtServer.bucketlist;
	while (litem != NULL) {
		eb = slist_get_data (litem);
		if (!(eb->flags & EVENTBUCKET_FLAG_INUSE)) {
			eb->flags = EVENTBUCKET_FLAG_INUSE;
			eb->events = 0;
			pthread_mutex_unlock (&EvtServer.bucketmutex);
			return eb;
		}
		litem = slist_get_next (litem);
	}
	pthread_mutex_unlock (&EvtServer.bucketmutex);
	return NULL;
}


int netinfoevtd_eventbucket_release(EventBucket *eb)
{
	if (eb == NULL)
		return NETINFO_RC_INVALIDARG;
	pthread_mutex_lock (&EvtServer.bucketmutex);
	eb->flags = 0;
	pthread_mutex_unlock (&EvtServer.bucketmutex);
	return NETINFO_RC_SUCCESS;
}


int eventbucket_add_event(EventBucket *eb, EventData *event)
{
	int	rc;
	int	bytes;
	if (eb == NULL || event == NULL)
		return NETINFO_RC_INVALIDARG;
	if (eb->events == eb->bucketsize)
		return NETINFO_RC_ENDOFFILE;
	rc = eventdata_sizeof (event, &bytes);
	memcpy (&eb->eventarray[eb->events], event, bytes);
	eb->events++;
	return NETINFO_RC_SUCCESS;
}
#endif


#ifdef	CONFIG_EVENTBUCKETS
void *netinfoevtd_event_processing_thread (void *arg)
{
	int		rc;
	int		i;
	int		len;
	char		hostname[STRING_SIZE];
	char		object[STRING_SIZE];
	sigset_t	sigset;
	struct timeval	stv;
	struct timeval	etv;
	SList		*mitem = NULL;
	NIEHModule	*nm = NULL;
	EventBucket	*eb = NULL;
	SList		*ebitem = NULL;

	sigfillset (&sigset);
	pthread_sigmask (SIG_BLOCK, &sigset, NULL);

	while (1) {
		pthread_mutex_lock (&EvtServer.procmutex);
		pthread_cond_wait (&EvtServer.proccond, &EvtServer.procmutex);
		pthread_mutex_unlock (&EvtServer.procmutex);
#ifdef DEBUG
/* fprintf (stderr, "%s: woke up!\n", __PRETTY_FUNCTION__); */
#endif
		ebitem = EvtServer.bucketlist;
		while (ebitem != NULL) {
			eb = slist_get_data (ebitem);
			i = EVENTBUCKET_FLAG_INUSE | EVENTBUCKET_FLAG_READONLY;
			if ( (eb->flags & i) != i) {
				ebitem = slist_get_next (ebitem);
				continue;
			}
			for (i=0 ; i<eb->events ; i++) {
				pthread_mutex_lock (&EvtServer.modmutex);
				mitem = EvtServer.modlist;
				while (mitem != NULL) {
					nm = slist_get_data (mitem);
					if (nm == NULL) {
						mitem = slist_get_next (mitem);
						continue;
					}
					gettimeofday (&stv, NULL);
					pthread_mutex_lock (&nm->mutex);
					rc = (*nm->handle_event) (eb->eventarray[i]);
					pthread_mutex_unlock (&nm->mutex);
					if (rc != NETINFO_RC_SUCCESS) {
						len = STRING_SIZE;
						rc = eventdata_get_hostname (eb->eventarray[i],
									hostname,
									&len);
						len = STRING_SIZE;
						rc = eventdata_get_object (eb->eventarray[i],
									object,
									&len);
						netinfoevtd_log_entry (
							"(%s) Module method returned rc=%d (host='%s' object='%s')\n",
							nm->modulepath,
							rc,
							hostname,
							object);
					}
					gettimeofday (&etv, NULL);
					nm->usecondssum += ( (etv.tv_sec * 1000000 + etv.tv_usec) -
							     (stv.tv_sec * 1000000 + stv.tv_usec));
					nm->eventsprocessed++;
					if ( (nm->eventsprocessed % 10000) == 0) {
						netinfoevtd_log_entry (
							"(%s) Module processing avg = %lu usec/event\n",
							nm->modulepath,
							nm->usecondssum / 100);
						nm->usecondssum = 0;
					}
					mitem = slist_get_next (mitem);
				}
				pthread_mutex_unlock (&EvtServer.modmutex);
			}
			rc = netinfoevtd_eventbucket_release (eb);
			ebitem = slist_get_next (ebitem);
		}
	}
	pthread_exit (NULL);
	return NULL;
}
#else
void *netinfoevtd_event_processing_thread (void *arg)
{
	int		rc;
	int		len;
	char		hostname[STRING_SIZE];
	char		object[STRING_SIZE];
	sigset_t	sigset;
	struct timeval	stv;
	struct timeval	etv;
	SList		*mitem = NULL;
	NIEHModule	*nm = NULL;
	SQueue		*queue = NULL;
	EventData	*event = NULL;

	sigfillset (&sigset);
	pthread_sigmask (SIG_BLOCK, &sigset, NULL);
	rc = eventdata_new(&event);

	while (1) {
		/* Wait if needed, then grab a queue off the outqlist */
		pthread_mutex_lock (&EvtServer.procmutex);
		if (EvtServer.outqlist == NULL) {
			pthread_cond_wait (&EvtServer.proccond, &EvtServer.procmutex);
#ifdef DEBUG
/*			fprintf (stderr, "%s: woke up from condition wait!\n", __PRETTY_FUNCTION__); */
#endif
		}
		pthread_mutex_unlock (&EvtServer.procmutex);
		pthread_mutex_lock (&EvtServer.outqmutex);
		if (EvtServer.outqlist != NULL) {
			queue = slist_get_data (EvtServer.outqlist);
			rc = slist_item_remove (&EvtServer.outqlist, queue);
		}
		pthread_mutex_unlock (&EvtServer.outqmutex);
		/* Process each queue item with each plugin */
		while (1) {
			rc = squeue_item_dequeue (queue, event);
			if (rc != 0)
				break;
			pthread_mutex_lock (&EvtServer.modmutex);
			mitem = EvtServer.modlist;
			while (mitem != NULL) {
				nm = slist_get_data (mitem);
				if (nm == NULL) {
					mitem = slist_get_next (mitem);
					continue;
				}
				gettimeofday (&stv, NULL);
				pthread_mutex_lock (&nm->mutex);
				rc = (*nm->handle_event) (event);
				pthread_mutex_unlock (&nm->mutex);
				if (rc != NETINFO_RC_SUCCESS) {
					len = STRING_SIZE;
					rc = eventdata_get_hostname (event,
								hostname,
								&len);
					len = STRING_SIZE;
					rc = eventdata_get_object (event,
								object,
								&len);
					netinfoevtd_log_entry (
						"(%s) Module method returned rc=%d (host='%s' object='%s')\n",
						nm->modulepath,
						rc,
						hostname,
						object);
				}
				gettimeofday (&etv, NULL);
				nm->usecondssum += ( (etv.tv_sec * 1000000 + etv.tv_usec) -
						     (stv.tv_sec * 1000000 + stv.tv_usec));
				nm->eventsprocessed++;
				if ( (nm->eventsprocessed % 10000) == 0) {
					netinfoevtd_log_entry (
						"(%s) Module processing avg = %lu usec/event\n",
						nm->modulepath,
						nm->usecondssum / 100);
					nm->usecondssum = 0;
				}
				mitem = slist_get_next (mitem);
			}
			pthread_mutex_unlock (&EvtServer.modmutex);
		}
		/* Add the empty queue back to the inqlist */
#ifdef DEBUG
		netinfoevtd_log_entry (
			"DEBUG: Moving empty queue back onto the inqlist\n");
#endif
		pthread_mutex_lock (&EvtServer.inqmutex);
		rc = slist_item_add (&EvtServer.inqlist, queue);
		pthread_mutex_unlock (&EvtServer.inqmutex);
		queue = NULL;
	}
#ifdef DEBUG
	netinfoevtd_log_entry (
		"DEBUG: event_processing_thread exiting.\n");
#endif
	rc = eventdata_free(&event);
	pthread_exit (NULL);
	return NULL;
}
#endif


#ifdef	CONFIG_EVENTBUCKETS
int netinfoevtd_eventbucket_stats (EventBucketStats *ebs)
{
	SList		*litem = NULL;
	EventBucket	*eb = NULL;
	if (ebs == NULL)
		return EINVAL;
	ebs->totalbuckets = 0;
	ebs->inusebuckets = 0;
	ebs->totalevents = 0;
	ebs->inuseevents = 0;
	ebs->fillaverage = 0.0;
	pthread_mutex_lock (&EvtServer.bucketmutex);
	litem = EvtServer.bucketlist;
	while (litem != NULL) {
		eb = slist_get_data (litem);
		if (eb->flags & EVENTBUCKET_FLAG_INUSE) {
			ebs->inusebuckets++;
			ebs->totalevents += eb->bucketsize;
			ebs->inuseevents += eb->events;
		}
		ebs->totalbuckets++;
/*		ebs->fillaverage += (double)eb->events/(double)eb->bucketsize; */
		litem = slist_get_next (litem);
	}
	pthread_mutex_unlock (&EvtServer.bucketmutex);
/*	ebs->fillaverage = (ebs->fillaverage/(double)ebs->totalbuckets)*100.0; */
	if (ebs->inuseevents > 0 && ebs->totalevents > 0)
		ebs->fillaverage = ((double)ebs->inuseevents/(double)ebs->totalevents)*100.0;
	else
		ebs->fillaverage = 0.0;
	return 0;
}
#endif


#ifndef	CONFIG_EVENTBUCKETS
int netinfoevtd_queues_allocate (int qsize, int qcount)
{
	int	rc;
	int	i;
	SQueue	*queue = NULL;
	rc = 0;
	pthread_mutex_lock (&EvtServer.inqmutex);
	for (i=0 ; i<qcount ; i++) {
		rc = squeue_new (&queue, EvtServer.eventsize, qsize);
		if (rc != 0) {
			pthread_mutex_unlock (&EvtServer.inqmutex);
			return rc;
		}
		rc = slist_item_add (&EvtServer.inqlist, queue);
		if (rc != 0) {
			pthread_mutex_unlock (&EvtServer.inqmutex);
			return rc;
		}
	}
	pthread_mutex_unlock (&EvtServer.inqmutex);
	return rc;
}


int netinfoevtd_queues_deallocate (void)
{
	int	rc;
	SQueue	*queue = NULL;
	SList	*litem = NULL;
	rc = 0;
	pthread_mutex_lock (&EvtServer.outqmutex);
	litem = EvtServer.outqlist;
	while (litem != NULL) {
		queue = slist_get_data (litem);
		rc = squeue_free (&queue);
		litem = slist_get_next (litem);
	}
	rc = slist_free (&EvtServer.outqlist);
	pthread_mutex_unlock (&EvtServer.outqmutex);

	pthread_mutex_lock (&EvtServer.inqmutex);
	litem = EvtServer.inqlist;
	while (litem != NULL) {
		queue = slist_get_data (litem);
		rc = squeue_free (&queue);
		litem = slist_get_next (litem);
	}
	rc = slist_free (&EvtServer.inqlist);
	pthread_mutex_unlock (&EvtServer.inqmutex);
	return rc;
}


int netinfoevtd_queues_statistics (void)
{
	int	rc = 0;
	return rc;
}
#endif


#ifndef W32_SERVICE
void usage(void)
{
	printf("\nnetinfoevtd - Information Server\n");
	printf(" Options:\n");
#ifndef WIN32
	printf("   -D		   Netinfoevtd will not daemonize itself.\n");
#endif
	printf("   -f configfile   Specifies the path to the configuration file to use.\n");
	printf("   -h		   This help screen.\n");
	printf("   -l logfile	   Specifies the path to the log file to use.\n");
	printf("   -V		   Display version information and exit.\n\n");
	return;
}
#endif


