/************************************************************************/
/* nidaemon.c - Daemonize a process and generate events for it.		*/
/************************************************************************/
/*
 *  (c)2005 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <netinfo/netinfo_portable.h>
#include <netinfo/netinfo.h>
#include <netinfo/event.h>


#define	FLAG_APPEND		1
#define	STRING_SIZE		512
#define HEARTBEAT_INTERVAL	300
#define HEARTBEAT_NAME		"HEARTBEAT-"

#define	RC_OK			0
#define	RC_WARNING		1
#define	RC_CRITICAL		2
#define	RC_UNKNOWN		3


void usage (void);
void signal_handler (int signum);

int main (int argc, char **argv)
{
	int			rc;
	int			erc;
	int			option;
	int			fdin;
	int			fdout;
	int			fderr;
	int			flags;
	int			fdflags;
	int			waitseconds;
	int			procstatus;
	int			hbinterval;
	pid_t			pid;
	struct sigaction	saction;
	char			msgbuffer[STRING_SIZE];
	char			*login = NULL;
	char			*logout = NULL;
	char			*logerr = NULL;
	char			*cmd = NULL;
	char			**args = NULL;
	char			*objectname = NULL;
	char			*hbobject = NULL;
	EventClient		*ec = NULL;
	EventData		*ed = NULL;

	rc = 0;
	flags = 0;
	waitseconds = 0;
	hbinterval = HEARTBEAT_INTERVAL;
	while (( option = getopt(argc, argv, "0:1:2:aH:hO:o:Vw:")) != EOF) {
		switch (option) {
			case '0':
				login = optarg;
				break;
			case '1':
				logout = optarg;
				break;
			case '2':
				logerr = optarg;
				break;
			case 'a':
				flags |= FLAG_APPEND;
				break;
			case 'H':
				hbinterval = atoi (optarg);
				break;
			case 'h':
				usage();
				exit (0);
				break;
			case 'O':
				hbobject = strdup (optarg);
				break;
			case 'o':
				objectname = strdup (optarg);
				break;
			case 'w':
				waitseconds = atoi (optarg);
				break;
			case 'V':
				fprintf(stdout,
					"nidaemon - Version %s\n",
					VERSION);
				exit (0);
				break;
			default:
				usage();
				exit (EINVAL);
				break;
		}
	}
	if (optind < argc) {
		cmd = argv[optind];
		args = &argv[optind];
	} else {
		fprintf (stderr,
			"You must specify a command to execute.\n");
		usage ();
		exit (EINVAL);
	}
	if (objectname == NULL) {
		fprintf (stderr,
			"You must specify an object name.\n");
		usage ();
		exit (EINVAL);
	}
	if (hbobject == NULL) {
		rc = strlen (HEARTBEAT_NAME) + strlen (objectname) + 1;
		hbobject = calloc (1, rc);
		if (hbobject == NULL) {
			return rc;
		}
		rc = snprintf (hbobject, 
				rc,
				"%s%s",
				HEARTBEAT_NAME,
				objectname);
	}
	erc = eventdata_new (&ed);
	if (erc != NETINFO_RC_SUCCESS)
		return erc;
	pid = fork ();
	if (pid == -1) {
		rc = errno;
		perror ("fork");
		exit (rc);
	} else if (pid > 0) {
		/* parent */
		erc = eventdata_free (&ed);
		if (waitseconds == 0)
			exit (0);
		saction.sa_handler = signal_handler;
		saction.sa_sigaction = NULL;
		rc = sigemptyset (&saction.sa_mask);
		saction.sa_flags = 0;
		rc = sigaction (SIGCHLD,
				&saction,
				NULL);
		rc = sigaction (SIGALRM,
				&saction,
				NULL);
		rc = alarm (waitseconds);
		pid = waitpid (pid, &procstatus, 0);
		if (pid == -1) {
			rc = errno;
			perror ("waitpid");
			exit (rc);
		} else {
			if (WIFEXITED (procstatus)) {
				rc = WEXITSTATUS (procstatus);
				fprintf (stdout,
					"The monitor exited: rc=%d.\n",
					rc);
				exit (rc);
			} else if (WIFSIGNALED (procstatus)) {
				rc = WTERMSIG (procstatus);
				fprintf (stdout,
					"The monitor caught signal: %d.\n",
					rc);
				exit (rc);
			}
		}
	} else {
		/* child */
		chdir ("/");
		setsid ();
		close (2);
		close (1);
		close (0);
		fdin = -1;
		fdout = -1;
		fderr = -1;
		if (login != NULL) {
			fdin = open (login, O_RDONLY);
			if (fdin == -1)
				exit (errno);
		}
		fdflags = O_WRONLY | O_CREAT;
		if (flags & FLAG_APPEND)
			fdflags |= O_APPEND;
		if (logout != NULL) {
			fdout = open (logout, fdflags, 0600);
			if (fdout == -1)
				exit (errno);
			if (fdout == 0) {
				fdout = dup2 (fdout, 1);
				if (fdout == -1)
					exit (errno);
			}
		}
		if (logerr != NULL) {
			fderr = open (logerr, fdflags, 0600);
			if (fderr == -1)
				exit (errno);
			if (fderr == 1) {
				fderr = dup2 (fderr, 2);
				if (fderr == -1)
					exit (errno);
			}
		}
		erc = eventclient_init (&ec, 0);
		if (erc != NETINFO_RC_SUCCESS)
			exit (erc);
		erc = eventclient_configure_heartbeat (ec,
					objectname,
					hbinterval);	
		pid = fork ();
		if (pid == -1) {
			rc = errno;
			exit (rc);
		} else if (pid > 0) {
			/* parent to watch child */
			snprintf (msgbuffer,
				STRING_SIZE,
				"The daemon process is running (pid=%d).\n",
				pid);
			erc = eventdata_set (ed,
					0,
					RC_OK,
					objectname,
					msgbuffer,
					strlen (msgbuffer));
			erc = eventclient_emit_event (ec, ed);
			/* erc = eventclient_uninit (&ec); */
			saction.sa_handler = signal_handler;
			saction.sa_sigaction = NULL;
			rc = sigemptyset (&saction.sa_mask);
			saction.sa_flags = 0;
			rc = sigaction (SIGCHLD,
					&saction,
					NULL);
parent_wait_for_child:
			pid = waitpid (pid, &procstatus, 0);
			if (pid == -1) {
				rc = errno;
				if (rc == EAGAIN)
					goto parent_wait_for_child;
				erc = eventdata_free (&ed);
				exit (rc);
			} else {
				/*
				erc = eventchain_open (&ec);
				if (erc != NETINFO_RC_SUCCESS)
					exit (rc);
				*/
				if (WIFEXITED (procstatus)) {
					rc = WEXITSTATUS (procstatus);
					snprintf (msgbuffer,
						STRING_SIZE,
						"The daemon exited: rc=%d.\n",
						rc);
					if (rc != 0)
						rc = RC_CRITICAL;
				} else if (WIFSIGNALED (procstatus)) {
					rc = WTERMSIG (procstatus);
					snprintf (msgbuffer,
						STRING_SIZE,
						"The daemon caught signal: %d.\n",
						rc);
					rc = RC_CRITICAL;
				}
				erc = eventdata_set (ed,
						0,
						rc,
						objectname,
						msgbuffer,
						strlen (msgbuffer));
				erc = eventclient_emit_event (ec, ed);
				erc = eventdata_free (&ed);
				erc = eventclient_uninit (&ec);
				exit (rc);
			}
		} else {
			/* child to run command */
			erc = eventdata_free (&ed);
			rc = execv (cmd, args);
			if (rc == -1)
				exit (errno);
			exit (0);
		}
	}
	

	return rc;
}


void usage (void)
{
	fprintf (stderr,
		"Usage: nidaemon [option]... -- command [argument]...\n");
	fprintf (stderr,
		"  -0 filename    - Attach the file to the daemon's stdin.\n");
	fprintf (stderr,
		"  -1 filename    - Attach the file to the daemon's stdout.\n");
	fprintf (stderr,
		"  -2 filename    - Attach the file to the daemon's stderr.\n");
	fprintf (stderr,
		"  -H interval    - Heartbeat interval, in seconds(default 300).\n");
	fprintf (stderr,
		"  -h             - Display command usage text.\n");
	fprintf (stderr,
		"  -O objectname  - The object name to use in the heartbeat events.\n");
	fprintf (stderr,
		"  -o objectname  - The object name to use in the events.\n");
	fprintf (stderr,
		"  -w seconds     - Wait for the daemon process to exit.\n");
	fprintf (stderr,
		"                   If the daemon exits within the timeframe,\n");
	fprintf (stderr,
		"                   the parent will receive the exit code.\n");
	fprintf (stderr,
		"  -V             - Display version information and exit.\n");
	return;
}


void signal_handler (int signum)
{
	switch (signum) {
		case SIGTERM:
		case SIGINT:
		case SIGQUIT:
			fprintf (stdout, "DEBUG: caught SIGTERM\n");
			exit (0);
			break;
		case SIGALRM:
			fprintf (stdout, "DEBUG: caught SIGALRM\n");
			break;
	}
	return;
}


